! Module containing SLEPc matrixes and variables
module SLEPC

! Use
use GANLIB

! Include
#include <slepc/finclude/slepcepsdef.h>
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscvec.h90>

! Variables
integer, allocatable, dimension(:) ::  ipt_fill
integer, allocatable, dimension(:,:) ::  aij_matrixes, map_aij
integer ::  num_b1j_matrixes, calculate_flux, num_ener_grp, ups_ener_grp
integer :: flag_rav, flag_spn
integer, allocatable, dimension(:) ::  num_aij_matrixes, b1j_matrixes, map_b1j

! SLEPc varibles
Mat   matB(500,500), matA(500,500)
Vec   vecC(500), evec_slepc(500), xups, bups
KSP   ksp_matA(500)
Mat matCwx(500) , matCwy(500), matCxy(500) 
Mat matAw(500), matAx(500), matAy(500), matAz(500)
Mat matRw, matRx, matRy, matRz
Vec vecT(500)
Vec vecTl(500,8)
Mat matCwxl(500,8) , matCwyl(500,8), matCxyl(500,8) 
Mat matAwl(500,8), matAxl(500,8), matAyl(500,8), matAzl(500,8)
Vec vecQwl(8,8),vecQxl(8,8),vecQyl(8,8),vecQzl(8,8)


contains










! Subroutine which controls the process of solving the eigenvalue problem with
! SLEPc
subroutine SLEPC_driver(IPTRK,IPSYS,IMPX,NUN,NGRP,LMOD,FKEFFV,EV,AD)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! IMPX: print parameter: =0: no print ; =1: minimum printing;
!       =2: iteration history is printed; =3: solution is printed.
! NUN: total number of unknowns per group.
! NGRP: number of energy groups.
! LMOD: number of bi-orthogonal harmonics to compute.
! FKEFFV: effective multiplication factor of each harmonic.
! EV: converged direct harmonic vector.
! AD: converged adjoint harmonic vector.




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer :: IMPX, NUN, NGRP, LMOD
REAL :: FKEFFV(LMOD),EV(NUN,NGRP,LMOD), AD(NUN,NGRP,LMOD)



! PETSc variables
PetscErrorCode  ierr
Vec   v, chi1, chig 
PetscScalar zero, one

! Local variables
integer :: i, j, k, type_problem_slepc, gups, npre, num_ups, npre_thomas
PARAMETER(NSTATE=40)
INTEGER ISTATE(NSTATE), ILONG, ITYLCM, LL4, LL4TS, ITY, ICHX, LL4F, NLF
CHARACTER TEXT12*12
LOGICAL LMU,LMUW,LMUX,LMUY,LMUZ,DIAG
REAL TK1, TK2
integer npre_spn, size_matrix





! Get the dimension of the system matrixes LL4
CALL LCMGET(IPSYS,'STATE-VECTOR',ISTATE)
LL4=ISTATE(2)
ITY=ISTATE(4)
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ICHX=ISTATE(12)
IF(ICHX == 2) THEN ! Dual (Raviart-Thomas)
   flag_rav=1
   LL4F=ISTATE(25)
ELSE ! Primal or MCFD(Nodal Collocation)
   flag_rav=0
   LL4F=LL4
ENDIF
NLF=ISTATE(30)
flag_spn=NLF


! Determine the type of non-diagonal matrixes
CALL LCMLEN(IPTRK,'MU',ILONG,ITYLCM)
LMU=(ILONG.NE.0).AND.(ITYLCM.EQ.1)
CALL LCMLEN(IPTRK,'MUW',ILONG,ITYLCM)
LMUW=(ILONG.NE.0).AND.(ITYLCM.EQ.1)
CALL LCMLEN(IPTRK,'MUX',ILONG,ITYLCM)
LMUX=(ILONG.NE.0).AND.(ITYLCM.EQ.1)
CALL LCMLEN(IPTRK,'MUY',ILONG,ITYLCM)
LMUY=(ILONG.NE.0).AND.(ITYLCM.EQ.1)
CALL LCMLEN(IPTRK,'MUZ',ILONG,ITYLCM)
LMUZ=(ILONG.NE.0).AND.(ITYLCM.EQ.1)
DIAG=LMUY.AND.(.NOT.LMUX)


! Determine the type of problem: 1, 2, 3
! Problem 1: No upscattering 
! Problem 2: No upscattering to first group
! Problem 3: Upscattering to first group
type_problem_slepc=0
i=1
num_ups=0
do while (type_problem_slepc==0)
   if (i==NGRP) then
      type_problem_slepc=1 
   else
      do j=i+1,NGRP
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .and. i==1 ) then
            type_problem_slepc=3
         elseif (ILONG /= 0) then
            type_problem_slepc=2
            gups=i
            num_ups=NGRP-gups+1
         endif 
      enddo
      i=i+1
   endif
enddo




! Create matrixes Aij in SLEPC format
if (flag_spn>0) then
   npre_spn=LL4*NLF/2/40
   size_matrix=LL4*NLF/2
else
   if (ICHX==2) call prealloc_matrixes_raviart_thomas(IPTRK,IPSYS,npre_thomas) 
   size_matrix=LL4
endif
select case(type_problem_slepc)
case(1) ! Case with no upscattering
   allocate(num_aij_matrixes(NGRP),aij_matrixes(NGRP,NGRP))
   allocate(map_aij(NGRP,NGRP))
   map_aij=0
   do i=1,NGRP
      num_aij_matrixes(i)=0
      do j=1,i
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            num_aij_matrixes(i)=num_aij_matrixes(i)+1
            aij_matrixes(i,num_aij_matrixes(i))=j
            map_aij(i,j)=1
            ! Determine npre
            ! SPn
            if (flag_spn>0) then
               npre=npre_spn
            ! Diffusion
            else
               if (ICHX==2 .and. i==j) then
                  npre=npre_thomas
               else
                  call prealloc_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,0, &
                   & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,npre)
               endif
            endif
            call create_matrixes_slepc(i,j,size_matrix,0,npre) 
         endif 
      enddo
   enddo
case(2) ! Case with no upscattering to first group
   allocate(num_aij_matrixes(NGRP),aij_matrixes(NGRP,NGRP))
   allocate(map_aij(NGRP,NGRP))
   map_aij=0
   ! Matrixes with no upscattering
   do i=1,NGRP
      num_aij_matrixes(i)=0
      do j=1,min(i,gups-1)
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            num_aij_matrixes(i)=num_aij_matrixes(i)+1
            aij_matrixes(i,num_aij_matrixes(i))=j
            map_aij(i,j)=1
            ! Determine npre
            ! SPn
            if (flag_spn>0) then
               npre=npre_spn
            ! Diffusion
            else
               if (ICHX==2 .and. i==j) then
                  npre=npre_thomas
               else
                  call prealloc_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,0, &
                   & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,npre)
               endif
            endif
            ! Create matrix
            call create_matrixes_slepc(i,j,size_matrix,0,npre) 
         endif 
      enddo
   enddo
   ! Matrix containing upscattering 
   ! Determine preallocation
   npre=0
   do i=gups,NGRP
      do j=gups,NGRP
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Determine npre
            ! SPn
            if (flag_spn>0) then
               k=npre_spn
            ! Diffusion
            else
               if (ICHX==2 .and. i==j) then
                  k=npre_thomas
               else
                  call prealloc_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,0, &
                   & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,k)
               endif
            endif
         else
            k=0
         endif 
         npre=npre+k
      enddo
   enddo
   ! Create matrix 
   call create_matrixes_slepc(gups,gups,size_matrix*num_ups,0,npre) 
case(3) ! Case with upscattering to first group
   ! Determine preallocation
   npre=0
   do i=1,NGRP
      do j=1,NGRP
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Determine npre
            ! SPn
            if (flag_spn>0) then
               k=npre_spn
            ! Diffusion
            else
               if (ICHX==2 .and. i==j) then
                  k=npre_thomas
               else
                  call prealloc_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,0, &
                   & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,k)
               endif
            endif
         else
            k=0
         endif 
         npre=npre+k
      enddo
   enddo
   ! Create matrix 
   call create_matrixes_slepc(1,1,size_matrix*NGRP,0,npre) 
end select



! Create matrixes Bij in SLEPC format
select case(type_problem_slepc)
case(1,2) ! Cases with no upscattering to first group
   allocate(b1j_matrixes(NGRP))
   allocate(map_b1j(NGRP))
   map_b1j=0
   num_b1j_matrixes=0
   do i=1,NGRP
      do j=1,NGRP
         write(TEXT12,'(1HB,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0) then
            if (i==1) then
               num_b1j_matrixes=num_b1j_matrixes+1
               b1j_matrixes(num_b1j_matrixes)=j
               map_b1j(j)=1
            endif
            ! Determine npre
            call prealloc_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,1, &
             & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,npre)
            ! Create matrix
            call create_matrixes_slepc(i,j,size_matrix,1,npre) 
         endif 
      enddo
   enddo
case(3) ! Case with upscattering to first group
   ! Determine preallocation
   npre=0
   do i=1,NGRP
      do j=1,NGRP
         write(TEXT12,'(1HB,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0) then
            ! Determine npre
            call prealloc_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,1, &
             & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,k)
         else
            k=0
         endif 
         npre=npre+k
      enddo
   enddo
   ! Create matrix 
   call create_matrixes_slepc(1,1,size_matrix*NGRP,1,npre) 
end select





! Allocate variable "IPT", which will be used to filling matrixes Agg
allocate(ipt_fill(LL4F))





! Fill matrixes Aij in SLEPC format
select case(type_problem_slepc)
case(1) ! Case with no upscattering
   do i=1,NGRP
      do j=1,i
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Fill matrix
            ! SPn
            if (flag_spn>0) then
               if (i==j) then
                  call fill_matrix_Aii_raviart_thomas_spn(IPTRK,IPSYS,i,i)
               else
                  ! Cartesian geometry
                  call fill_matrix_Aij_cart_raviart_thomas_spn(IPTRK,IPSYS,i,j,i,j)
                  ! Hexagonal geometry
               endif
            ! Diffusion
            else
               if (ICHX==2 .and. i==j) then
                  call fill_matrixA_raviart_thomas(IPTRK,IPSYS,i) 
               else
                  call fill_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,0, &
                   & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG) 
               endif
            endif
         endif 
      enddo
   enddo
case(2) ! Case with no upscattering to first group
   ! Matrixes with no upscattering
   do i=1,NGRP
      do j=1,min(i,gups-1)
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Fill matrix
            ! SPn
            if (flag_spn>0) then
               if (i==j) then
                  call fill_matrix_Aii_raviart_thomas_spn(IPTRK,IPSYS,i,i)
               else
                  ! Cartesian geometry
                  call fill_matrix_Aij_cart_raviart_thomas_spn(IPTRK,IPSYS,i,j,i,j)
                  ! Hexagonal geometry
               endif
            ! Diffusion
            else
               if (ICHX==2 .and. i==j) then
                  call fill_matrixA_raviart_thomas(IPTRK,IPSYS,i) 
               else
                  call fill_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,0, &
                   & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG) 
               endif 
            endif 
         endif 
      enddo
   enddo
   do i=gups,NGRP
      do j=gups,NGRP
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Fill matrix
            ! SPn
            if (flag_spn>0) then
               if (i==j) then
                  call fill_matrix_Aii_raviart_thomas_spn(IPTRK,IPSYS,i,gups)
               else
                  ! Cartesian geometry
                  call fill_matrix_Aij_cart_raviart_thomas_spn(IPTRK,IPSYS,i,j,gups,gups)
                  ! Hexagonal geometry
               endif
            ! Diffusion
            else
               if (ICHX==2 .and. i==j) then
                  call fill_matrixA_raviart_thomas_several_groups(IPTRK,IPSYS,i,gups)
               else
                  call fill_matrix_several_groups_slepc(IPTRK,IPSYS,i,j,size_matrix,ITY,0, &
                   & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,gups) 
               endif
            endif
         endif 
      enddo
   enddo
case(3) ! Case with upscattering to first group
   do i=1,NGRP
      do j=1,NGRP
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Fill matrix
            ! SPn
            if (flag_spn>0) then
               if (i==j) then
                  call fill_matrix_Aii_raviart_thomas_spn(IPTRK,IPSYS,i,1)
               else
                  ! Cartesian geometry
                  call fill_matrix_Aij_cart_raviart_thomas_spn(IPTRK,IPSYS,i,j,1,1)
                  ! Hexagonal geometry
               endif
            ! Diffusion
            else
               if (ICHX==2 .and. i==j) then
                  call fill_matrixA_raviart_thomas_several_groups(IPTRK,IPSYS,i,1)
               else
                  call fill_matrix_several_groups_slepc(IPTRK,IPSYS,i,j,size_matrix,ITY,0, &
                   & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,1) 
               endif 
            endif 
         endif 
      enddo
   enddo
end select




! Fill matrixes Bij in SLEPC format
select case(type_problem_slepc)
case(1,2) ! Cases with no upscattering to first group
   do i=1,NGRP
      do j=1,NGRP
         write(TEXT12,'(1HB,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0) then
            ! Fill matrix
            call fill_matrixes_slepc(IPTRK,IPSYS,i,j,LL4F,ITY,1, &
             & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG) 
         endif 
      enddo
   enddo
case(3) ! Case with upscattering to first group
   do i=1,NGRP
      do j=1,NGRP
         write(TEXT12,'(1HB,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0) then
            ! Fill matrix
            call fill_matrix_several_groups_slepc(IPTRK,IPSYS,i,j,size_matrix,ITY,1, &
             & LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,1) 
         endif 
      enddo
   enddo
end select




! Deallocate variable "IPT", which was used to filling matrixes Agg
deallocate(ipt_fill)




! Assemble matrixes Aij in SLEPC format
select case(type_problem_slepc)
case(1) ! Case with no upscattering
   do i=1,NGRP
      do j=1,i
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Assemble matrix
            call assemble_matrixes_slepc(i,j,0) 
         endif 
      enddo
   enddo
case(2) ! Case with no upscattering to first group
   ! Matrixes with no upscattering
   do i=1,NGRP
      do j=1,min(i,gups-1)
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Assemble matrix
            call assemble_matrixes_slepc(i,j,0) 
         endif 
      enddo
   enddo
   ! Assemble matrix with upscattering terms
   call assemble_matrixes_slepc(gups,gups,0) 
case(3) ! Case with upscattering to first group
   ! Assemble matrix 
   call assemble_matrixes_slepc(1,1,0) 
end select



! Assemble matrixes Bij in SLEPC format
select case(type_problem_slepc)
case(1,2) ! Cases with no upscattering to first group
   do i=1,NGRP
      do j=1,NGRP
         write(TEXT12,'(1HB,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0) then
            ! Assemble matrix
            call assemble_matrixes_slepc(i,j,1) 
         endif 
      enddo
   enddo
case(3) ! Case with upscattering to first group
   ! Assemble matrix
   call assemble_matrixes_slepc(1,1,1) 
end select



! Create vecC, containing the ratio of the fission spectrum for each row of matB
if (type_problem_slepc<3 .and. NGRP>1) then
   ! Initialize PETSc variables
   zero=0.0
   one=1.0
   call MatCreateVecs(matA(1,1),chi1,PETSC_NULL_OBJECT,ierr)
   call VecDuplicate(chi1,v,ierr)
   call VecDuplicate(chi1,chig,ierr)
   ! Calculate the sum of the vectors containing the diagonal of matB(1,:)
   i=1
   call VecSet(chi1,zero,ierr)
   do j=1,NGRP
      write(TEXT12,'(1HB,2I3.3)') i,j
      CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
      if(ILONG /= 0) then
         call MatGetDiagonal(matB(i,j),v,ierr)
      else
         call VecSet(v,zero,ierr)
      endif 
      call VecAXPY(chi1,one,v,ierr)
   enddo
   call VecReciprocal(chi1,ierr)
   ! Calculate the sum of the vectors containing the diagonal of matB(1,:)
   do i=2,NGRP
      call VecDuplicate(chi1,vecC(i),ierr)
      call VecSet(chig,zero,ierr)
      do j=1,NGRP
         write(TEXT12,'(1HB,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0) then
            call MatGetDiagonal(matB(i,j),v,ierr)
         else
            call VecSet(v,zero,ierr)
         endif 
         call VecAXPY(chig,one,v,ierr)
      enddo
      ! Calculate vecC(i)
      call VecPointwiseMult(vecC(i),chig,chi1,ierr)
   enddo
   ! Destroy vectors: v, chi1, chig
   call VecDestroy(v,ierr)
   call VecDestroy(chi1,ierr)
   call VecDestroy(chig,ierr)
endif





! Solve eigenvalue problem with SLEPc
call cpu_time(TK1)
write(6,*)
write(6,*) '*************************************************'
select case(type_problem_slepc)
case(1) ! Case with no upscattering
   call solver_slepc_without_upscattering(NUN,size_matrix,NGRP,LMOD,FKEFFV,EV,AD)
case(2) ! Case with no upscattering to first group
   call solver_slepc_without_upscattering_to_1(NUN,size_matrix,NGRP,gups,LMOD,FKEFFV,EV,AD)
case(3) ! Case with upscattering to first group
   call solver_slepc_with_upscattering_to_1(NUN,size_matrix,NGRP,LMOD,FKEFFV,EV,AD)
end select
write(6,*) '*************************************************'
write(6,*)
call cpu_time(TK2)
! Print results
if (IMPX >= 1) then
   do j=1,LMOD
      WRITE(6,'(1H1//13H HARMONIC NB.,I3//)') j
      WRITE(6,'(/42H SLEPc : EFFECTIVE MULTIPLICATION FACTOR =,1P,D17.10/)') FKEFFV(j)
      do i=1,NGRP
         WRITE(6,'(//9H SLEPc : ,A7,37H EIGENVECTOR CORRESPONDING TO THE GRO,&
           &2HUP,I4//(5X,1P,8E14.5))') 'DIRECT',i,(EV(k,i,j),k=1,LL4)
      enddo
      do i=1,NGRP
         WRITE(6,'(//9H SLEPc : ,A7,37H EIGENVECTOR CORRESPONDING TO THE GRO,&
           &2HUP,I4//(5X,1P,8E14.5))') 'ADJOINT',i,(AD(k,i,j),k=1,LL4)
      enddo
   enddo
   WRITE(6,'(/53H SLEPc : CPU TIME USED TO SOLVE EIGENVALUE PROBLEM  =,F10.3/)') TK2-TK1
endif






! Destroy matrixes Aij in SLEPC format
select case(type_problem_slepc)
case(1) ! Case with no upscattering
   deallocate(num_aij_matrixes,aij_matrixes)
   deallocate(map_aij)
   do i=1,NGRP
      do j=1,i
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Destroy matrix
            call destroy_matrixes_slepc(i,j,0) 
         endif 
      enddo
   enddo
case(2) ! Case with no upscattering to first group
   deallocate(num_aij_matrixes,aij_matrixes)
   deallocate(map_aij)
   ! Matrixes with no upscattering
   do i=1,NGRP
      do j=1,min(i,gups-1)
         write(TEXT12,'(1HA,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0 .or. (i==j)) then
            ! Destroy matrix
            call destroy_matrixes_slepc(i,j,0) 
         endif 
      enddo
   enddo
   ! Destroy matrix
   call destroy_matrixes_slepc(gups,gups,0) 
case(3) ! Case with upscattering to first group
   ! Destroy matrix
   call destroy_matrixes_slepc(1,1,0) 
end select




! Destroy matrixes Bij in SLEPC format
select case(type_problem_slepc)
case(1,2) ! Cases with no upscattering to first group
   deallocate(b1j_matrixes)
   deallocate(map_b1j)
   do i=1,NGRP
      do j=1,NGRP
         write(TEXT12,'(1HB,2I3.3)') i,j
         CALL LCMLEN(IPSYS,TEXT12,ILONG,ITYLCM)
         if(ILONG /= 0) then
            ! Destroy matrix
            call destroy_matrixes_slepc(i,j,1) 
         endif 
      enddo
   enddo
case(3) ! Case with upscattering to first group
   ! Destroy matrix
   call destroy_matrixes_slepc(1,1,1) 
end select




! Destroy vectors C in SLEPC format
if (type_problem_slepc<3 .and. NGRP>1) then
   do i=2,NGRP
      call VecDestroy(vecC(i),ierr)
   enddo
endif

end subroutine SLEPC_driver












! This subroutine creates the system matrixes in SLEPc format 
subroutine create_matrixes_slepc(g1,g2,LL4,mat_a_or_b,npre)
! g1: energy group g1 of Matrix(g1,g2)
! g2: energy group g2 of Matrix(g1,g2)
! LL4: size of the matrix
! mat_a_or_b = 0 or 1;   if it is 0, the matrix is matA; if it is 1, the matrix is matB 
! npre: maximum number of not null terms in one row of the system matrixes


! Global variables
integer :: g1,g2, LL4, mat_a_or_b
! SLEPc variables
PetscInt npre





! Local variables
! SLEPc variables
PetscErrorCode ierr



! Create matrix
select case(mat_a_or_b)
case(0) ! matA
   call MatCreate(PETSC_COMM_WORLD,matA(g1,g2),ierr)
   call MatSetSizes(matA(g1,g2),PETSC_DECIDE,PETSC_DECIDE,LL4,LL4,ierr)
   call MatSetFromOptions(matA(g1,g2),ierr)
   !!!!npre=LL4*2/100+1
   call MatSeqAIJSetPreallocation(matA(g1,g2),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matA(g1,g2),ierr)
case(1) ! matB
   call MatCreate(PETSC_COMM_WORLD,matB(g1,g2),ierr)
   call MatSetSizes(matB(g1,g2),PETSC_DECIDE,PETSC_DECIDE,LL4,LL4,ierr)
   call MatSetFromOptions(matB(g1,g2),ierr)
   !!!!npre=LL4*2/100+1
   call MatSeqAIJSetPreallocation(matB(g1,g2),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matB(g1,g2),ierr)
end select


end subroutine create_matrixes_slepc













! This subroutine determines the maximum number of not null terms in one row of the system matrixes  
subroutine prealloc_matrixes_slepc(IPTRK,IPSYS,g1,g2,LL4,ITY,mat_a_or_b, &
& LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,npre)
! IPTRK: L_TRACK pointer to the TRIVAC tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 of Matrix(g1,g2)
! g2: energy group g2 of Matrix(g1,g2)
! LL4: size of the matrix
! ITY: number to determine if the matrix contains W,X,Y,Z matrixes
! mat_a_or_b = 0 or 1;   if it is 0, the matrix is matA; if it is 1, the matrix is matB 
! LMU: Logical variable defining system matrixes contains off-diagonal terms 
! LMUW: Logical variable defining system matrixes are W-oriented 
! LMUX: Logical variable defining system matrixes are X-oriented 
! LMUY: Logical variable defining system matrixes are Y-oriented 
! LMUY: Logical variable defining system matrixes are Y-oriented 
! DIAG: Logical variable defining geometry has been defined with diagonal
! boundary conditions  
! npre: maximum number of not null terms in one row of the system matrixes


! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer :: g1,g2, LL4, ITY, mat_a_or_b, npre
LOGICAL LMU,LMUW,LMUX,LMUY,LMUZ,DIAG





! Local variables
type(c_ptr) :: ptr_c
integer :: j, i, k
INTEGER ILONG, ITYLCM, LL4TS
CHARACTER NAMT*12, TEXT12*12
TYPE(C_PTR) MU_PTR
INTEGER, DIMENSION(:), POINTER :: MU



! Determine the name of the matrix: A_g1,g2 or B_g1,g2
select case(mat_a_or_b)
case(0) ! matA
   WRITE(NAMT,'(1HA,2I3.3)') g1,g2
case(1) ! matB
   WRITE(NAMT,'(1HB,2I3.3)') g1,g2
end select

! Matrix is non-diagonal
IF((ITY==1 .or. ITY==11).or.(((ITY.EQ.13).or.(g1==g2)) .and. mat_a_or_b==0))  THEN 
   IF(LMU) THEN
      CALL LCMLEN(IPTRK,'MU',LL4TS,ITYLCM)
      IF(LL4.NE.LL4TS) CALL XABORT('MTLDLM: INVALID LL4(1).')
      CALL LCMGPD(IPTRK,'MU',MU_PTR)
      CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
      npre=0
      do i=2,LL4
         j=MU(i)-MU(i-1)
         npre=max(npre,j) 
      enddo
   ELSE 
      npre=0
      ! W- OR X-ORIENTED MATRIX.
      IF(LMUW) THEN
         TEXT12='W_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUW',MU_PTR)
      ELSE IF(DIAG) THEN
         TEXT12='Y_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUY',MU_PTR)
      ELSE
         TEXT12='X_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUX',MU_PTR)
      ENDIF
      CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
      k=0
      do i=2,LL4
         j=MU(i)-MU(i-1)
         k=max(k,j) 
      enddo
      npre=npre+k
      ! Y-ORIENTED MATRIX.
      IF(LMUY) THEN
         CALL LCMGPD(IPTRK,'MUY',MU_PTR)
         CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
         k=0
         do i=2,LL4
            j=MU(i)-MU(i-1)
            k=max(k,j) 
         enddo
         npre=npre+k
      ENDIF
      ! Z-ORIENTED MATRIX.
      IF(LMUZ) THEN
         CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
         CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
         k=0
         do i=2,LL4
            j=MU(i)-MU(i-1)
            k=max(k,j) 
         enddo
         npre=npre+k
      ENDIF
   ENDIF

! Matrix is diagonal
ELSE
   npre=1
ENDIF

npre=2*(npre-1)+1



end subroutine prealloc_matrixes_slepc












! This subroutine fills the system matrixes in SLEPc format 
subroutine fill_matrixes_slepc(IPTRK,IPSYS,g1,g2,LL4,ITY,mat_a_or_b, &
& LMU,LMUW,LMUX,LMUY,LMUZ,DIAG)
! IPTRK: L_TRACK pointer to the TRIVAC tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 of Matrix(g1,g2)
! g2: energy group g2 of Matrix(g1,g2)
! LL4: size of the matrix
! ITY: number to determine if the matrix contains W,X,Y,Z matrixes
! mat_a_or_b = 0 or 1;   if it is 0, the matrix is matA; if it is 1, the matrix is matB 
! LMU: Logical variable defining system matrixes contains off-diagonal terms 
! LMUW: Logical variable defining system matrixes are W-oriented 
! LMUX: Logical variable defining system matrixes are X-oriented 
! LMUY: Logical variable defining system matrixes are Y-oriented 
! LMUY: Logical variable defining system matrixes are Y-oriented 
! DIAG: Logical variable defining geometry has been defined with diagonal
! boundary conditions  


! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer :: g1,g2, LL4, ITY, mat_a_or_b
LOGICAL LMU,LMUW,LMUX,LMUY,LMUZ,DIAG





! Local variables
type(c_ptr) :: ptr_c
integer :: j, i, k, i1, j1, kend
INTEGER ILONG, ITYLCM, LL4TS
CHARACTER NAMT*12, TEXT12*12
TYPE(C_PTR) MU_PTR,IP_PTR
TYPE(C_PTR) ASS_PTR
INTEGER, DIMENSION(:), POINTER :: MU,IP
REAL, DIMENSION(:), POINTER :: ASS
! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Determine the name of the matrix: A_g1,g2 or B_g1,g2
select case(mat_a_or_b)
case(0) ! matA
   WRITE(NAMT,'(1HA,2I3.3)') g1,g2
case(1) ! matB
   WRITE(NAMT,'(1HB,2I3.3)') g1,g2
end select

! Matrix is non-diagonal
IF(ITY==1 .or.((ITY==2).and.(g1==g2) .and. mat_a_or_b==0))  THEN 
   IF(LMU) THEN
      CALL LCMLEN(IPTRK,'MU',LL4TS,ITYLCM)
      IF(LL4.NE.LL4TS) CALL XABORT('MTLDLM: INVALID LL4(1).')
      CALL LCMGPD(IPTRK,'MU',MU_PTR)
      CALL LCMGPD(IPSYS,NAMT,ASS_PTR)
      CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
      CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4) /))
   ELSE 
      ! W- OR X-ORIENTED MATRIX.
      IF(LMUW) THEN
         TEXT12='W_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUW',MU_PTR)
         CALL LCMGPD(IPTRK,'IPW',IP_PTR)
         CALL LCMLEN(IPTRK,'IPW',LL4TS,ITYLCM)
      ELSE IF(DIAG) THEN
         TEXT12='Y_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUY',MU_PTR)
         CALL LCMGPD(IPTRK,'IPX',IP_PTR)
         CALL LCMLEN(IPTRK,'IPX',LL4TS,ITYLCM)
      ELSE
         TEXT12='X_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUX',MU_PTR)
         CALL LCMGPD(IPTRK,'IPX',IP_PTR)
         CALL LCMLEN(IPTRK,'IPX',LL4TS,ITYLCM)
      ENDIF
      IF(LL4.NE.LL4TS) CALL XABORT('MTLDLM: INVALID LL4(2).')
      CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
      CALL C_F_POINTER(IP_PTR,IP,(/ LL4 /))
      CALL LCMGPD(IPSYS,TEXT12,ASS_PTR)
      CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4) /))
      ! Transpose the index of permutation matrixes
      do i=1,LL4
         ipt_fill(IP(i))=i
      enddo
      ! Fill the matrix
      do i=1,LL4
         i1=ipt_fill(i)
         if (i>1) then
            kend=MU(i)-MU(i-1)
            do k=1,kend
               j=i-k+1
               j1=ipt_fill(j)
               val=ASS(MU(i)-k+1)
               call MatSetValue(matA(g1,g2),i1-1,j1-1,val,ADD_VALUES,ierr)
               if (i1/=j1) call MatSetValue(matA(g1,g2),j1-1,i1-1,val,ADD_VALUES,ierr)
            enddo
         else
            val=ASS(1)
            call MatSetValue(matA(g1,g2),i1-1,i1-1,val,ADD_VALUES,ierr)
         endif
      enddo
      !!!!write(100*g1+71,*) (IP(k),k=1,LL4) 
      !!!!write(100*g1+81,*) (MU(k),k=1,LL4) 
      ! X-ORIENTED MATRIX, IF W-ORIENTED MATRIX IS PRESENT.
      IF(LMUW) THEN
         IF(DIAG) THEN
            CALL LCMGPD(IPTRK,'MUY',MU_PTR)
            CALL LCMGPD(IPSYS,'Y_'//NAMT,ASS_PTR)
         ELSE
            CALL LCMGPD(IPTRK,'MUX',MU_PTR)
            CALL LCMGPD(IPSYS,'X_'//NAMT,ASS_PTR)
         ENDIF
         CALL LCMLEN(IPTRK,'IPX',LL4TS,ITYLCM)
         IF(LL4.NE.LL4TS) CALL XABORT('MTLDLM: INVALID LL4(6).')
         CALL LCMGPD(IPTRK,'IPX',IP_PTR)
         CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
         CALL C_F_POINTER(IP_PTR,IP,(/ LL4 /))
         CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4) /))
         ! Transpose the index of permutation matrixes
         do i=1,LL4
            ipt_fill(IP(i))=i
         enddo
         ! Fill the matrix
         do i=2,LL4
            i1=ipt_fill(i)
            kend=MU(i)-MU(i-1)
            do k=1,kend
               j=i-k+1
               j1=ipt_fill(j)
               if (i1/=j1) then
                  val=ASS(MU(i)-k+1)
                  call MatSetValue(matA(g1,g2),i1-1,j1-1,val,ADD_VALUES,ierr)
                  call MatSetValue(matA(g1,g2),j1-1,i1-1,val,ADD_VALUES,ierr)
               endif
            enddo
         enddo
      ENDIF
      ! Y-ORIENTED MATRIX.
      IF(LMUY) THEN
         CALL LCMGPD(IPTRK,'MUY',MU_PTR)
         CALL LCMLEN(IPTRK,'IPY',LL4TS,ITYLCM)
         IF(LL4.NE.LL4TS) CALL XABORT('MTLDLM: INVALID LL4(6).')
         CALL LCMGPD(IPTRK,'IPY',IP_PTR)
         CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
         CALL C_F_POINTER(IP_PTR,IP,(/ LL4 /))
         CALL LCMGPD(IPSYS,'Y_'//NAMT,ASS_PTR)
         CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4) /))
         ! Transpose the index of permutation matrixes
         do i=1,LL4
            ipt_fill(IP(i))=i
         enddo
         ! Fill the matrix
         do i=2,LL4
            i1=ipt_fill(i)
            kend=MU(i)-MU(i-1)
            do k=1,kend
               j=i-k+1
               j1=ipt_fill(j)
               if (i1/=j1) then
                  val=ASS(MU(i)-k+1)
                  call MatSetValue(matA(g1,g2),i1-1,j1-1,val,ADD_VALUES,ierr)
                  call MatSetValue(matA(g1,g2),j1-1,i1-1,val,ADD_VALUES,ierr)
               endif
            enddo
         enddo
         !!!!write(100*g1+72,*) (IP(k),k=1,LL4) 
         !!!!write(100*g1+82,*) (MU(k),k=1,LL4) 
      ENDIF
      ! Z-ORIENTED MATRIX.
      IF(LMUZ) THEN
         CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
         CALL LCMLEN(IPTRK,'IPZ',LL4TS,ITYLCM)
         IF(LL4.NE.LL4TS) CALL XABORT('MTLDLM: INVALID LL4(7).')
         CALL LCMGPD(IPTRK,'IPZ',IP_PTR)
         CALL C_F_POINTER(MU_PTR,MU,(/ LL4 /))
         CALL C_F_POINTER(IP_PTR,IP,(/ LL4 /))
         CALL LCMGPD(IPSYS,'Z_'//NAMT,ASS_PTR)
         CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4) /))
         ! Transpose the index of permutation matrixes
         do i=1,LL4
            ipt_fill(IP(i))=i
         enddo
         ! Fill the matrix
         do i=2,LL4
            i1=ipt_fill(i)
            kend=MU(i)-MU(i-1)
            do k=1,kend
               j=i-k+1
               j1=ipt_fill(j)
               if (i1/=j1) then
                  val=ASS(MU(i)-k+1)
                  call MatSetValue(matA(g1,g2),i1-1,j1-1,val,ADD_VALUES,ierr)
                  call MatSetValue(matA(g1,g2),j1-1,i1-1,val,ADD_VALUES,ierr)
               endif
            enddo
         enddo
         !!!!write(100*g1+73,*) (IP(k),k=1,LL4) 
         !!!!write(100*g1+83,*) (MU(k),k=1,LL4) 
      ENDIF
   ENDIF

! Matrix is diagonal
ELSEIF (g1/=g2 .or. mat_a_or_b==1) then
   ! Get the values of the matrix from IPSYS
   CALL LCMGPD(IPSYS,NAMT,ASS_PTR)
   CALL C_F_POINTER(ASS_PTR,ASS,(/ LL4 /))
   ! Set the values in SLEPc matrixes 
   select case(mat_a_or_b)
   case(0) ! matA
      do j=1,LL4
         val=ASS(j)
         call MatSetValue(matA(g1,g2),j-1,j-1,val,INSERT_VALUES,ierr)
      enddo
   case(1) ! matB
      do j=1,LL4
         val=ASS(j)
         call MatSetValue(matB(g1,g2),j-1,j-1,val,INSERT_VALUES,ierr)
      enddo
   end select
ENDIF



end subroutine fill_matrixes_slepc












! This subroutine fills the system matrixes in SLEPc format 
subroutine fill_matrix_several_groups_slepc(IPTRK,IPSYS,g1,g2,LL4,ITY,mat_a_or_b, &
& LMU,LMUW,LMUX,LMUY,LMUZ,DIAG,gups)
! IPTRK: L_TRACK pointer to the TRIVAC tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 of Matrix(g1,g2)
! g2: energy group g2 of Matrix(g1,g2)
! LL4: size of the matrix
! ITY: number to determine if the matrix contains W,X,Y,Z matrixes
! mat_a_or_b = 0 or 1;   if it is 0, the matrix is matA; if it is 1, the matrix is matB 
! LMU: Logical variable defining system matrixes contains off-diagonal terms 
! LMUW: Logical variable defining system matrixes are W-oriented 
! LMUX: Logical variable defining system matrixes are X-oriented 
! LMUY: Logical variable defining system matrixes are Y-oriented 
! LMUY: Logical variable defining system matrixes are Y-oriented 
! DIAG: Logical variable defining geometry has been defined with diagonal
! boundary conditions  
! gups: first energy group containing upscattring terms


! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer :: g1,g2, LL4, ITY, mat_a_or_b, gups
LOGICAL LMU,LMUW,LMUX,LMUY,LMUZ,DIAG





! Local variables
type(c_ptr) :: ptr_c
integer :: j, i, k, i1, j1, kend
INTEGER ILONG, ITYLCM, LL4TS
CHARACTER NAMT*12, TEXT12*12
TYPE(C_PTR) MU_PTR,IP_PTR
TYPE(C_PTR) ASS_PTR
INTEGER, DIMENSION(:), POINTER :: MU,IP
REAL, DIMENSION(:), POINTER :: ASS
! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Determine the name of the matrix: A_g1,g2 or B_g1,g2
select case(mat_a_or_b)
case(0) ! matA
   WRITE(NAMT,'(1HA,2I3.3)') g1,g2
case(1) ! matB
   WRITE(NAMT,'(1HB,2I3.3)') g1,g2
end select

! Matrix is non-diagonal
IF(ITY==1 .or.((ITY==2).and.(g1==g2) .and. mat_a_or_b==0))  THEN 
   IF(LMU) THEN
      CALL LCMLEN(IPTRK,'MU',LL4TS,ITYLCM)
      CALL LCMGPD(IPTRK,'MU',MU_PTR)
      CALL LCMGPD(IPSYS,NAMT,ASS_PTR)
      CALL C_F_POINTER(MU_PTR,MU,(/ LL4TS /))
      CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4TS) /))
   ELSE 
      ! W- OR X-ORIENTED MATRIX.
      IF(LMUW) THEN
         TEXT12='W_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUW',MU_PTR)
         CALL LCMGPD(IPTRK,'IPW',IP_PTR)
         CALL LCMLEN(IPTRK,'IPW',LL4TS,ITYLCM)
      ELSE IF(DIAG) THEN
         TEXT12='Y_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUY',MU_PTR)
         CALL LCMGPD(IPTRK,'IPX',IP_PTR)
         CALL LCMLEN(IPTRK,'IPX',LL4TS,ITYLCM)
      ELSE
         TEXT12='X_'//NAMT(:10)
         CALL LCMGPD(IPTRK,'MUX',MU_PTR)
         CALL LCMGPD(IPTRK,'IPX',IP_PTR)
         CALL LCMLEN(IPTRK,'IPX',LL4TS,ITYLCM)
      ENDIF
      CALL C_F_POINTER(MU_PTR,MU,(/ LL4TS /))
      CALL C_F_POINTER(IP_PTR,IP,(/ LL4TS /))
      CALL LCMGPD(IPSYS,TEXT12,ASS_PTR)
      CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4TS) /))
      ! Transpose the index of permutation matrixes
      do i=1,LL4TS
         ipt_fill(IP(i))=i
      enddo
      ! Fill the matrix
      do i=1,LL4TS
         i1=ipt_fill(i)+(g1-gups)*LL4
         if (i>1) then
            kend=MU(i)-MU(i-1)
            do k=1,kend
               j=i-k+1
               j1=ipt_fill(j)+(g2-gups)*LL4
               val=ASS(MU(i)-k+1)
               if (g1/=g2) val=-1.0*val
               call MatSetValue(matA(gups,gups),i1-1,j1-1,val,ADD_VALUES,ierr)
               if (i1-(g1-gups)*LL4 /= j1-(g2-gups)*LL4) call MatSetValue(matA(gups,gups),j1-1,i1-1,val,ADD_VALUES,ierr)
            enddo
         else
            val=ASS(1)
            if (g1/=g2) val=-1.0*val
            call MatSetValue(matA(gups,gups),i1-1,i1-1,val,ADD_VALUES,ierr)
         endif
      enddo
      ! X-ORIENTED MATRIX, IF W-ORIENTED MATRIX IS PRESENT.
      IF(LMUW) THEN
         IF(DIAG) THEN
            CALL LCMGPD(IPTRK,'MUY',MU_PTR)
            CALL LCMGPD(IPSYS,'Y_'//NAMT,ASS_PTR)
         ELSE
            CALL LCMGPD(IPTRK,'MUX',MU_PTR)
            CALL LCMGPD(IPSYS,'X_'//NAMT,ASS_PTR)
         ENDIF
         CALL LCMLEN(IPTRK,'IPX',LL4TS,ITYLCM)
         CALL LCMGPD(IPTRK,'IPX',IP_PTR)
         CALL C_F_POINTER(MU_PTR,MU,(/ LL4TS /))
         CALL C_F_POINTER(IP_PTR,IP,(/ LL4TS /))
         CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4TS) /))
         ! Transpose the index of permutation matrixes
         do i=1,LL4TS
            ipt_fill(IP(i))=i
         enddo
         ! Fill the matrix
         do i=2,LL4TS
            i1=ipt_fill(i)+(g1-gups)*LL4
            kend=MU(i)-MU(i-1)
            do k=1,kend
               j=i-k+1
               j1=ipt_fill(j)+(g2-gups)*LL4
               if (i1-(g1-gups)*LL4 /= j1-(g2-gups)*LL4) then
                  val=ASS(MU(i)-k+1)
                  if (g1/=g2) val=-1.0*val
                  call MatSetValue(matA(gups,gups),i1-1,j1-1,val,ADD_VALUES,ierr)
                  call MatSetValue(matA(gups,gups),j1-1,i1-1,val,ADD_VALUES,ierr)
               endif
            enddo
         enddo
      ENDIF
      ! Y-ORIENTED MATRIX.
      IF(LMUY) THEN
         CALL LCMGPD(IPTRK,'MUY',MU_PTR)
         CALL LCMLEN(IPTRK,'IPY',LL4TS,ITYLCM)
         CALL LCMGPD(IPTRK,'IPY',IP_PTR)
         CALL C_F_POINTER(MU_PTR,MU,(/ LL4TS /))
         CALL C_F_POINTER(IP_PTR,IP,(/ LL4TS /))
         CALL LCMGPD(IPSYS,'Y_'//NAMT,ASS_PTR)
         CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4TS) /))
         ! Transpose the index of permutation matrixes
         do i=1,LL4TS
            ipt_fill(IP(i))=i
         enddo
         ! Fill the matrix
         do i=2,LL4TS
            i1=ipt_fill(i)+(g1-gups)*LL4
            kend=MU(i)-MU(i-1)
            do k=1,kend
               j=i-k+1
               j1=ipt_fill(j)+(g2-gups)*LL4
               if (i1-(g1-gups)*LL4 /= j1-(g2-gups)*LL4) then
                  val=ASS(MU(i)-k+1)
                  if (g1/=g2) val=-1.0*val
                  call MatSetValue(matA(gups,gups),i1-1,j1-1,val,ADD_VALUES,ierr)
                  call MatSetValue(matA(gups,gups),j1-1,i1-1,val,ADD_VALUES,ierr)
               endif
            enddo
         enddo
      ENDIF
      ! Z-ORIENTED MATRIX.
      IF(LMUZ) THEN
         CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
         CALL LCMLEN(IPTRK,'IPZ',LL4TS,ITYLCM)
         CALL LCMGPD(IPTRK,'IPZ',IP_PTR)
         CALL C_F_POINTER(MU_PTR,MU,(/ LL4TS /))
         CALL C_F_POINTER(IP_PTR,IP,(/ LL4TS /))
         CALL LCMGPD(IPSYS,'Z_'//NAMT,ASS_PTR)
         CALL C_F_POINTER(ASS_PTR,ASS,(/ MU(LL4TS) /))
         ! Transpose the index of permutation matrixes
         do i=1,LL4TS
            ipt_fill(IP(i))=i
         enddo
         ! Fill the matrix
         do i=2,LL4TS
            i1=ipt_fill(i)+(g1-gups)*LL4
            kend=MU(i)-MU(i-1)
            do k=1,kend
               j=i-k+1
               j1=ipt_fill(j)+(g2-gups)*LL4
               if (i1-(g1-gups)*LL4 /= j1-(g2-gups)*LL4) then
                  val=ASS(MU(i)-k+1)
                  if (g1/=g2) val=-1.0*val
                  call MatSetValue(matA(gups,gups),i1-1,j1-1,val,ADD_VALUES,ierr)
                  call MatSetValue(matA(gups,gups),j1-1,i1-1,val,ADD_VALUES,ierr)
               endif
            enddo
         enddo
      ENDIF
   ENDIF

! Matrix is diagonal
ELSE
   ! Get the values of the matrix from IPSYS
   CALL LCMLEN(IPSYS,NAMT,LL4TS,ITYLCM)
   CALL LCMGPD(IPSYS,NAMT,ASS_PTR)
   CALL C_F_POINTER(ASS_PTR,ASS,(/ LL4TS /))
   ! Set the values in SLEPc matrixes 
   select case(mat_a_or_b)
   case(0) ! matA
      do j=1,LL4TS
         val=ASS(j)
         if (g1/=g2) val=-1.0*val
         i1=j+(g1-gups)*LL4
         j1=j+(g2-gups)*LL4
         call MatSetValue(matA(gups,gups),i1-1,j1-1,val,ADD_VALUES,ierr)
      enddo
   case(1) ! matB
      do j=1,LL4TS
         val=ASS(j)
         i1=j+(g1-gups)*LL4
         j1=j+(g2-gups)*LL4
         call MatSetValue(matB(gups,gups),i1-1,j1-1,val,ADD_VALUES,ierr)
      enddo
   end select
ENDIF



end subroutine fill_matrix_several_groups_slepc














! This subroutine assembles the system matrixes in SLEPc format 
subroutine assemble_matrixes_slepc(g1,g2,mat_a_or_b)
! g1: energy group g1 of Matrix(g1,g2)
! g2: energy group g2 of Matrix(g1,g2)
! mat_a_or_b = 0 or 1;   if it is 0, the matrix is matA; if it is 1, the matrix is matB 

! Global variables
integer :: g1,g2, mat_a_or_b





! Local variables
! SLEPc variables
PetscErrorCode ierr



! Destroy matrix
select case(mat_a_or_b)
case(0) ! matA
   call MatAssemblyBegin(matA(g1,g2),MAT_FINAL_ASSEMBLY,ierr)
   call MatAssemblyEnd(matA(g1,g2),MAT_FINAL_ASSEMBLY,ierr)
case(1) ! matB
   call MatAssemblyBegin(matB(g1,g2),MAT_FINAL_ASSEMBLY,ierr)
   call MatAssemblyEnd(matB(g1,g2),MAT_FINAL_ASSEMBLY,ierr)
end select


end subroutine assemble_matrixes_slepc














! This subroutine solves the eigenvalue problem with SLEPc for the case without
! upscattering 
subroutine solver_slepc_without_upscattering(NUN,LL4,NGRP,LMOD,FKEFFV,EV,AD)
! NUN: total number of unknowns per group.
! LL4: size of the matrix
! NGRP: number of energy groups 
! LMOD: number of bi-orthogonal harmonics to compute.
! FKEFFV: effective multiplication factor of each harmonic.
! EV: converged direct harmonic vector.
! AD: converged adjoint harmonic vector.

use slepceps

! Global variables
integer :: NUN, LL4, NGRP,LMOD
REAL :: FKEFFV(LMOD),EV(NUN,NGRP,LMOD),AD(NUN,NGRP,LMOD)




! Local variables
integer :: i,j,k
! SLEPc variables
PetscErrorCode ierr
Mat            MShell, Mshell_adj
EPS            eps, eps_adj 
EPSType        tname,method
EPSWhich       which
PetscReal      tol, error_eigenpairs
PetscInt       nev, ncv, mpd, its, nconv
PetscScalar    kr, ki, mone
Vec            xi, u, ui
PC             pc
PetscScalar, pointer :: array(:)





! Create KSP solvers, for solving linear systems
do i=1,NGRP
   call KSPCreate(PETSC_COMM_WORLD,ksp_matA(i),ierr)
   call KSPSetOperators(ksp_matA(i),matA(i,i),matA(i,i),ierr)
   if (flag_rav==1) then ! Raviart-thomas
      call KSPSetType(ksp_matA(i),KSPBCGS,ierr)
      call KSPGetPC(ksp_matA(i),pc,ierr)
      call PCSetType(pc,PCILU,ierr)
   else ! MCFD or finite elements
      call KSPSetType(ksp_matA(i),KSPGMRES,ierr)
      call KSPGetPC(ksp_matA(i),pc,ierr)
      call PCSetType(pc,PCASM,ierr)
   endif
   call KSPSetFromOptions(ksp_matA(i),ierr)
   call KSPSetUp(ksp_matA(i),ierr)
enddo



! Create eigenvectors in SLEPc format
call MatCreateVecs(matA(1,1),evec_slepc(1),PETSC_NULL_OBJECT,ierr)
do i=2,NGRP
   call VecDuplicate(evec_slepc(1),evec_slepc(i),ierr)
enddo

! Define Shell matrix
num_ener_grp=NGRP
call MatCreateShell(PETSC_COMM_WORLD,LL4,LL4,LL4,LL4,PETSC_NULL_INTEGER,Mshell,ierr)
call MatShellSetOperation(Mshell,MATOP_MULT,shell_matrix_without_upscattering,ierr) 





! Create eigensolver
call EPSCreate(PETSC_COMM_WORLD,eps,ierr)

! Set operators of the eigensolver
call EPSSetOperators(eps,Mshell,PETSC_NULL_OBJECT,ierr) 
call EPSSetProblemType(eps,EPS_NHEP,ierr)

! Set eigensolver parameters
nev=LMOD
ncv=max(4*LMOD,25)
mpd=max(4*LMOD,25)
call EPSSetDimensions(eps,nev,ncv,mpd,ierr)
which=EPS_LARGEST_MAGNITUDE
call EPSSetWhichEigenpairs(eps,which,ierr)
method=EPSKRYLOVSCHUR
call EPSSetType(eps,method,ierr)
tol=1e-6
maxit=999
call EPSSetTolerances(eps,tol,maxit,ierr)
call EPSSetFromOptions(eps,ierr)
call EPSSetUp(eps,ierr)

! Solve the eigenvalue problem 
calculate_flux=0
call EPSSolve(eps,ierr)


! Get some information from the solver
! Iterations
call EPSGetIterationNumber(eps,its,ierr)
! Method used
call EPSGetType(eps,tname,ierr)
! Tolerances
call EPSGetTolerances(eps,tol,maxit,ierr)
! Converged eigenvalues
call EPSGetConverged(eps,nconv,ierr)

! Print the information of the solver
write(6,*) 'Eigenvalue Problem solved by SLEPc'
write(6,'(/a,I4)') ' Number of iterations of the method:',its
write(6,'(a,a)') ' Solution method: ',tname
write(6,'(a,I2)') ' Number of requested eigenvalues:',nev
write(6,'(a,1P,E10.4,a,I4)') ' Stopping condition: tol=',tol,', maxit=', maxit
write(6,'(a,I2)') ' Number of converged eigenpairs:',nconv




! Define Shell matrix for the adjoint calculation
call MatCreateShell(PETSC_COMM_WORLD,LL4,LL4,LL4,LL4,PETSC_NULL_INTEGER,Mshell_adj,ierr)
call MatShellSetOperation(Mshell_adj,MATOP_MULT,shell_matrix_without_upscattering_adjoint,ierr) 



! Obtain the solution from the solver
call VecDuplicate(evec_slepc(1),xi,ierr)
do j=0,min(nconv,LMOD)-1 
   ! Get the eigenpair
   call EPSGetEigenpair(eps,j,kr,ki,evec_slepc(1),xi,ierr)
   ! Compute the error
   call EPSComputeError(eps,j,EPS_ERROR_RELATIVE,error_eigenpairs,ierr)
   ! Print eigenvalues and errors
   if (j==0) write(6,'(/a/a)') '           k          ||Ax-kx||/||kx||'&
             &,'   ----------------- ------------------'
   write(6,'(a,f12.6,a,e12.2e2)') '   ',kr,'       ',error_eigenpairs
   ! Store the eigenvalues
   FKEFFV(j+1)=kr 
   ! Determine the fluxes
   calculate_flux=1
   call MatMult(Mshell,evec_slepc(1),xi,ierr) 
   calculate_flux=0
   ! Store the fluxes
   do i=1,NGRP
      call VecGetArrayF90(evec_slepc(i),array,ierr)
      do k=1,LL4
         EV(k,i,j+1)=array(k) 
      enddo
      call VecRestoreArrayF90(evec_slepc(i),array,ierr)
   enddo
   ! Calculate the adjoint
   call MatMult(Mshell_adj,evec_slepc(1),xi,ierr) 
   ! Store the adjoint
   do i=1,NGRP
      call VecGetArrayF90(evec_slepc(i),array,ierr)
      do k=1,LL4
         AD(k,i,j+1)=array(k) 
      enddo
      call VecRestoreArrayF90(evec_slepc(i),array,ierr)
   enddo
enddo




! Destroy vec xi, u, ui
call VecDestroy(xi,ierr)


! Destroy Mshell, Mshell_adj
call MatDestroy(Mshell,ierr)
call MatDestroy(Mshell_adj,ierr)


! Destroy KSP solvers
do i=1,NGRP
   call KSPDestroy(ksp_matA(i),ierr)
enddo


! Destroy vectors evec_slepc()
do i=1,NGRP
   call VecDestroy(evec_slepc(i),ierr)
enddo


! Destroy Eigenvalue solver
call EPSDestroy(eps,ierr)



end subroutine solver_slepc_without_upscattering














! This subroutine calculates the shell matrix for the eigenvalue problem without
! upscattering
subroutine shell_matrix_without_upscattering(Mshell,initial,final,ierr)

! Global variables
! SLEPc variabes
Mat            Mshell
PetscErrorCode ierr
Vec            initial, final



! Local variables
integer :: g,i,j
! SLEPc variables
Vec            b, aux
PetscScalar    zero, one


! Define zero=0.0, one=1.0
zero=0.0
one=1.0


! Create vec b
call VecDuplicate(initial,b,ierr)
call VecDuplicate(initial,aux,ierr)

! Copy initial to evec_slepc(1). Flux of the first energy group is the initial
! estimation of the eigenvalue calculation
call VecCopy(initial,evec_slepc(1),ierr)


! Multiply matA(1,1)*evec_slepc(1) = final
call MatMult(matA(1,1),evec_slepc(1),final,ierr) 


! Determine the flux for the rest of energy groups
do g=2,num_ener_grp
   call VecPointwiseMult(b,final,vecC(g),ierr)
   do j=1,num_aij_matrixes(g)-1
      i=aij_matrixes(g,j)
      call MatMult(matA(g,i),evec_slepc(i),aux,ierr)  
      call VecAXPY(b,one,aux,ierr)
   enddo
   call KSPSolve(ksp_matA(g),b,evec_slepc(g),ierr)
enddo


! Solve eigenvalue problem
if (calculate_flux==0) then
   call VecSet(b,zero,ierr)
   do i=1,num_b1j_matrixes
      g=b1j_matrixes(i)
      call MatMult(matB(1,g),evec_slepc(g),aux,ierr) 
      call VecAXPY(b,one,aux,ierr)
   enddo
   call KSPSolve(ksp_matA(1),b,final,ierr)
endif


! Destroy vec b
call VecDestroy(b,ierr)
call VecDestroy(aux,ierr)


end subroutine shell_matrix_without_upscattering














! This subroutine calculates the shell matrix for the eigenvalue problem without
! upscattering
subroutine shell_matrix_without_upscattering_adjoint(Mshell,initial,final,ierr)

! Global variables
! SLEPc variabes
Mat            Mshell
PetscErrorCode ierr
Vec            initial, final



! Local variables
integer :: g,i,j,k
! SLEPc variables
Vec            b, aux
PetscScalar    zero, one 


! Define zero=0.0, one=1.0
zero=0.0
one=1.0


! Set evec_slepc(1)=initial
call VecCopy(initial,evec_slepc(1),ierr)


! Create vec b, aux
call VecDuplicate(evec_slepc(1),b,ierr)
call VecDuplicate(evec_slepc(1),aux,ierr)



! Determine the adjoint flux for the rest of energy groups
do k=2,num_ener_grp
   g=num_ener_grp+2-k
   if (map_b1j(g)==1) then
      call MatMultTranspose(matB(1,g),evec_slepc(1),b,ierr)   
   else
      call VecSet(b,zero,ierr)
   endif
   do i=g+1,num_ener_grp
      if (map_aij(i,g)==1) then
         call MatMultTranspose(matA(i,g),evec_slepc(i),aux,ierr)  
         call VecAXPY(b,one,aux,ierr)
      endif
   enddo
   call KSPSolveTranspose(ksp_matA(g),b,evec_slepc(g),ierr)
enddo


! Determine the adjoint flux for the first group
call VecSet(final,zero,ierr)
do i=2,num_ener_grp
   call VecPointwiseMult(aux,vecC(i),evec_slepc(i),ierr)
   call VecAXPY(final,one,aux,ierr)
enddo
call MatMultTranspose(matA(1,1),final,aux,ierr)   
if (map_b1j(1)==1) then
   call MatMultTranspose(matB(1,1),evec_slepc(1),final,ierr)   
else
   call VecSet(final,zero,ierr)
endif
call VecAXPY(final,one,aux,ierr)
do i=2,num_ener_grp
   if (map_aij(i,1)==1) then
      call MatMultTranspose(matA(i,1),evec_slepc(i),aux,ierr)  
      call VecAXPY(final,one,aux,ierr)
   endif
enddo
call KSPSolveTranspose(ksp_matA(1),final,evec_slepc(1),ierr)



! Destroy vec b
call VecDestroy(b,ierr)
call VecDestroy(aux,ierr)


end subroutine shell_matrix_without_upscattering_adjoint














! This subroutine solves the eigenvalue problem with SLEPc for the case without
! upscattering to first group 
subroutine solver_slepc_without_upscattering_to_1(NUN,LL4,NGRP,gups,LMOD,FKEFFV,EV,AD)
! NUN: total number of unknowns per group.
! LL4: size of the matrix
! NGRP: number of energy groups 
! gups: first energy group containing upscattering terms 
! LMOD: number of bi-orthogonal harmonics to compute.
! FKEFFV: effective multiplication factor of each harmonic.
! EV: converged direct harmonic vector.
! AD: converged adjoint harmonic vector.

use slepceps

! Global variables
integer :: NUN, LL4, NGRP, gups, LMOD
REAL :: FKEFFV(LMOD),EV(NUN,NGRP,LMOD),AD(NUN,NGRP,LMOD)




! Local variables
integer :: i,j,k
! SLEPc variables
PetscErrorCode ierr
Mat            MShell, Mshell_adj
EPS            eps, eps_adj 
EPSType        tname,method
EPSWhich       which
PetscReal      tol, error_eigenpairs
PetscInt       nev, ncv, mpd, its, nconv
PetscScalar    kr, ki, mone
Vec            xi, u, ui
PC             pc
PetscScalar, pointer :: array(:)





! Create KSP solvers, for solving linear systems
do i=1,gups
   call KSPCreate(PETSC_COMM_WORLD,ksp_matA(i),ierr)
   call KSPSetOperators(ksp_matA(i),matA(i,i),matA(i,i),ierr)
   if (flag_rav==1) then ! Raviart-thomas
      call KSPSetType(ksp_matA(i),KSPBCGS,ierr)
      call KSPGetPC(ksp_matA(i),pc,ierr)
      call PCSetType(pc,PCILU,ierr)
   else ! MCFD or finite elements
      call KSPSetType(ksp_matA(i),KSPGMRES,ierr)
      call KSPGetPC(ksp_matA(i),pc,ierr)
      call PCSetType(pc,PCASM,ierr)
   endif
   call KSPSetFromOptions(ksp_matA(i),ierr)
   call KSPSetUp(ksp_matA(i),ierr)
enddo



! Create eigenvectors in SLEPc format
call MatCreateVecs(matA(1,1),evec_slepc(1),PETSC_NULL_OBJECT,ierr)
do i=2,NGRP
   call VecDuplicate(evec_slepc(1),evec_slepc(i),ierr)
enddo

! Create xups and bups
call MatCreateVecs(matA(gups,gups),xups,PETSC_NULL_OBJECT,ierr)
call VecDuplicate(xups,bups,ierr)

! Define Shell matrix
num_ener_grp=NGRP
ups_ener_grp=gups
call MatCreateShell(PETSC_COMM_WORLD,LL4,LL4,LL4,LL4,PETSC_NULL_INTEGER,Mshell,ierr)
call MatShellSetOperation(Mshell,MATOP_MULT,shell_matrix_without_upscattering_to_1,ierr) 





! Create eigensolver
call EPSCreate(PETSC_COMM_WORLD,eps,ierr)

! Set operators of the eigensolver
call EPSSetOperators(eps,Mshell,PETSC_NULL_OBJECT,ierr) 
call EPSSetProblemType(eps,EPS_NHEP,ierr)

! Set eigensolver parameters
nev=LMOD
ncv=max(4*LMOD,25)
mpd=max(4*LMOD,25)
call EPSSetDimensions(eps,nev,ncv,mpd,ierr)
which=EPS_LARGEST_MAGNITUDE
call EPSSetWhichEigenpairs(eps,which,ierr)
method=EPSKRYLOVSCHUR
call EPSSetType(eps,method,ierr)
tol=1e-6
maxit=999
call EPSSetTolerances(eps,tol,maxit,ierr)
call EPSSetFromOptions(eps,ierr)
call EPSSetUp(eps,ierr)

! Solve the eigenvalue problem 
calculate_flux=0
call EPSSolve(eps,ierr)

! Get some information from the solver
! Iterations
call EPSGetIterationNumber(eps,its,ierr)
! Method used
call EPSGetType(eps,tname,ierr)
! Tolerances
call EPSGetTolerances(eps,tol,maxit,ierr)
! Converged eigenvalues
call EPSGetConverged(eps,nconv,ierr)

! Print the information of the solver
write(6,*) 'Eigenvalue Problem solved by SLEPc'
write(6,'(/a,I4)') ' Number of iterations of the method:',its
write(6,'(a,a)') ' Solution method: ',tname
write(6,'(a,I2)') ' Number of requested eigenvalues:',nev
write(6,'(a,1P,E10.4,a,I4)') ' Stopping condition: tol=',tol,', maxit=', maxit
write(6,'(a,I2)') ' Number of converged eigenpairs:',nconv



! Define Shell matrix for the adjoint calculation
call MatCreateShell(PETSC_COMM_WORLD,LL4,LL4,LL4,LL4,PETSC_NULL_INTEGER,Mshell_adj,ierr)
call MatShellSetOperation(Mshell_adj,MATOP_MULT,shell_matrix_without_upscattering_to_1_adjoint,ierr)


! Obtain the solution from the solver
call VecDuplicate(evec_slepc(1),xi,ierr)
do j=0,min(nconv,LMOD)-1 
   ! Get the eigenpair
   call EPSGetEigenpair(eps,j,kr,ki,evec_slepc(1),xi,ierr)
   ! Compute the error
   call EPSComputeError(eps,j,EPS_ERROR_RELATIVE,error_eigenpairs,ierr)
   ! Print eigenvalues and errors
   if (j==0) write(6,'(/a/a)') '           k          ||Ax-kx||/||kx||'&
             &,'   ----------------- ------------------'
   write(6,'(a,f12.6,a,e12.2e2)') '   ',kr,'       ',error_eigenpairs
   ! Store the eigenvalues
   FKEFFV(j+1)=kr 
   ! Determine the fluxes
   calculate_flux=1
   call MatMult(Mshell,evec_slepc(1),xi,ierr) 
   calculate_flux=0
   ! Store the fluxes
   do i=1,NGRP
      call VecGetArrayF90(evec_slepc(i),array,ierr)
      do k=1,LL4
         EV(k,i,j+1)=array(k) 
      enddo
      call VecRestoreArrayF90(evec_slepc(i),array,ierr)
   enddo
   ! Calculate the adjoint
   call MatMult(Mshell_adj,evec_slepc(1),xi,ierr) 
   ! Store the adjoint
   do i=1,NGRP
      call VecGetArrayF90(evec_slepc(i),array,ierr)
      do k=1,LL4
         AD(k,i,j+1)=array(k) 
      enddo
      call VecRestoreArrayF90(evec_slepc(i),array,ierr)
   enddo
enddo



! Destroy vec xi, u, ui
call VecDestroy(xi,ierr)


! Destroy Mshell, Mshell_adj
call MatDestroy(Mshell,ierr)
call MatDestroy(Mshell_adj,ierr)


! Destroy vecs xups and bups
call VecDestroy(xups,ierr)
call VecDestroy(bups,ierr)


! Destroy KSP solvers
do i=1,gups
   call KSPDestroy(ksp_matA(i),ierr)
enddo


! Destroy vectors evec_slepc()
do i=1,NGRP
   call VecDestroy(evec_slepc(i),ierr)
enddo


! Destroy Eigenvalue solver
call EPSDestroy(eps,ierr)



end subroutine solver_slepc_without_upscattering_to_1














! This subroutine calculates the shell matrix for the eigenvalue problem without
! upscattering to first group
subroutine shell_matrix_without_upscattering_to_1(Mshell,initial,final,ierr)

! Global variables
! SLEPc variabes
Mat            Mshell
PetscErrorCode ierr
Vec            initial, final



! Local variables
integer :: g,i,j
! SLEPc variables
Vec            b, aux
PetscScalar    zero, one
PetscScalar, pointer :: array(:), array_ups(:)
PetscInt       size_b


! Define zero=0.0, one=1.0
zero=0.0
one=1.0


! Create vec b
call VecDuplicate(initial,b,ierr)
call VecDuplicate(initial,aux,ierr)

! Copy initial to evec_slepc(1). Flux of the first energy group is the initial
! estimation of the eigenvalue calculation
call VecCopy(initial,evec_slepc(1),ierr)


! Multiply matA(1,1)*evec_slepc(1) = final
call MatMult(matA(1,1),evec_slepc(1),final,ierr) 


! Determine the flux for the energy groups without upscattering terms
do g=2,ups_ener_grp-1
   call VecPointwiseMult(b,final,vecC(g),ierr)
   do j=1,num_aij_matrixes(g)-1
      i=aij_matrixes(g,j)
      call MatMult(matA(g,i),evec_slepc(i),aux,ierr)  
      call VecAXPY(b,one,aux,ierr)
   enddo
   call KSPSolve(ksp_matA(g),b,evec_slepc(g),ierr)
enddo


! Get the size of b
call VecGetSize(b,size_b,ierr)


! Determine the flux for the energy groups with upscattering terms
! Get bup of linear system of the upscattering terms: Aups * xups = bups
call VecGetArrayF90(bups,array_ups,ierr)
do g=ups_ener_grp,num_ener_grp
   call VecPointwiseMult(b,final,vecC(g),ierr)
   j=1
   i=aij_matrixes(g,j)
   do while (i<ups_ener_grp .and. j<=num_aij_matrixes(g))
      call MatMult(matA(g,i),evec_slepc(i),aux,ierr)  
      call VecAXPY(b,one,aux,ierr)
      j=j+1
      i=aij_matrixes(g,j)
   enddo
   call VecGetArrayF90(b,array,ierr)
   do i=1,size_b
      j=i+size_b*(g-ups_ener_grp)
      array_ups(j)=array(i)
   enddo
   call VecRestoreArrayF90(b,array,ierr)
enddo
call VecRestoreArrayF90(bups,array_ups,ierr)
! Solve the system
call KSPSolve(ksp_matA(ups_ener_grp),bups,xups,ierr)
! Get the fluxes containing upscattering terms
call VecGetArrayF90(xups,array_ups,ierr)
do g=ups_ener_grp,num_ener_grp
   call VecGetArrayF90(evec_slepc(g),array,ierr)
   do i=1,size_b
      j=i+size_b*(g-ups_ener_grp)
      array(i)=array_ups(j)
   enddo
   call VecRestoreArrayF90(evec_slepc(g),array,ierr)
enddo
call VecRestoreArrayF90(xups,array_ups,ierr)


! Solve eigenvalue problem
if (calculate_flux==0) then
   call VecSet(b,zero,ierr)
   do i=1,num_b1j_matrixes
      g=b1j_matrixes(i)
      call MatMult(matB(1,g),evec_slepc(g),aux,ierr) 
      call VecAXPY(b,one,aux,ierr)
   enddo
   call KSPSolve(ksp_matA(1),b,final,ierr)
endif


! Destroy vec b
call VecDestroy(b,ierr)
call VecDestroy(aux,ierr)


end subroutine shell_matrix_without_upscattering_to_1














! This subroutine calculates the shell matrix for the eigenvalue problem without
! upscattering to first group
subroutine shell_matrix_without_upscattering_to_1_adjoint(Mshell,initial,final,ierr)

! Global variables
! SLEPc variabes
Mat            Mshell
PetscErrorCode ierr
Vec            initial, final



! Local variables
integer :: g,i,j,k
! SLEPc variables
Vec            b, aux
PetscScalar    zero, one 
PetscScalar, pointer :: array(:), array_ups(:)
PetscInt       size_b


! Define zero=0.0, one=1.0
zero=0.0
one=1.0


! Create vec b, aux
call VecDuplicate(evec_slepc(1),b,ierr)
call VecDuplicate(evec_slepc(1),aux,ierr)

! Set evec_slepc(1)=initial
call VecCopy(initial,evec_slepc(1),ierr)

! Get the size of b
call VecGetSize(b,size_b,ierr)


! Determine the flux for the energy groups with upscattering terms
! Get bup of linear system of the upscattering terms: Aups * xups = bups
call VecGetArrayF90(bups,array_ups,ierr)
do g=ups_ener_grp,num_ener_grp
   if (map_b1j(g)==1) then
      call MatMultTranspose(matB(1,g),evec_slepc(1),b,ierr)   
   else
      call VecSet(b,zero,ierr)
   endif
   call VecGetArrayF90(b,array,ierr)
   do i=1,size_b
      j=i+size_b*(g-ups_ener_grp)
      array_ups(j)=array(i)
   enddo
   call VecRestoreArrayF90(b,array,ierr)
enddo
call VecRestoreArrayF90(bups,array_ups,ierr)
! Solve the system
call KSPSolveTranspose(ksp_matA(ups_ener_grp),bups,xups,ierr)
! Get the fluxes containing upscattering terms
call VecGetArrayF90(xups,array_ups,ierr)
do g=ups_ener_grp,num_ener_grp
   call VecGetArrayF90(evec_slepc(g),array,ierr)
   do i=1,size_b
      j=i+size_b*(g-ups_ener_grp)
      array(i)=array_ups(j)
   enddo
   call VecRestoreArrayF90(evec_slepc(g),array,ierr)
enddo
call VecRestoreArrayF90(xups,array_ups,ierr)




! Determine the adjoint flux for the rest of energy groups
do k=2,ups_ener_grp-1
   g=ups_ener_grp+1-k
   if (map_b1j(g)==1) then
      call MatMultTranspose(matB(1,g),evec_slepc(1),b,ierr)   
   else
      call VecSet(b,zero,ierr)
   endif
   do i=g+1,num_ener_grp
      if (map_aij(i,g)==1) then
         call MatMultTranspose(matA(i,g),evec_slepc(i),aux,ierr)  
         call VecAXPY(b,one,aux,ierr)
      endif
   enddo
   call KSPSolveTranspose(ksp_matA(g),b,evec_slepc(g),ierr)
enddo


! Determine the adjoint flux for the first energy group
call VecSet(final,zero,ierr)
do i=2,num_ener_grp
   call VecPointwiseMult(aux,vecC(i),evec_slepc(i),ierr)
   call VecAXPY(final,one,aux,ierr)
enddo
call MatMultTranspose(matA(1,1),final,aux,ierr)   
if (map_b1j(1)==1) then
   call MatMultTranspose(matB(1,1),evec_slepc(1),final,ierr)
else
   call VecSet(final,zero,ierr)
endif
call VecAXPY(final,one,aux,ierr)
do i=2,num_ener_grp
   if (map_aij(i,1)==1) then
      call MatMultTranspose(matA(i,1),evec_slepc(i),aux,ierr)  
      call VecAXPY(final,one,aux,ierr)
   endif
enddo
call KSPSolveTranspose(ksp_matA(1),final,evec_slepc(1),ierr)



! Destroy vec b
call VecDestroy(b,ierr)
call VecDestroy(aux,ierr)


end subroutine shell_matrix_without_upscattering_to_1_adjoint
















! This subroutine solves the eigenvalue problem with SLEPc for the case with
! upscattering to first group 
subroutine solver_slepc_with_upscattering_to_1(NUN,LL4,NGRP,LMOD,FKEFFV,EV,AD)
! NUN: total number of unknowns per group.
! LL4: size of the matrix
! NGRP: number of energy groups 
! LMOD: number of bi-orthogonal harmonics to compute.
! FKEFFV: effective multiplication factor of each harmonic.
! EV: converged direct harmonic vector.
! AD: converged adjoint harmonic vector.

use slepceps

! Global variables
integer :: NUN, LL4, NGRP, LMOD
REAL :: FKEFFV(LMOD),EV(NUN,NGRP,LMOD),AD(NUN,NGRP,LMOD)




! Local variables
integer :: i,j,k
! SLEPc variables
PetscErrorCode ierr
EPS            eps 
EPSType        tname,method
EPSWhich       which
PetscReal      tol, error_eigenpairs
PetscInt       nev, ncv, mpd, its, nconv
PetscScalar    kr, ki, val 
Vec            xi, u, ui
PC             pc
KSP            ksp
PetscScalar, pointer :: array(:)



! Create KSP solver for matA(1,1)
i=1
call KSPCreate(PETSC_COMM_WORLD,ksp_matA(i),ierr)
call KSPSetOperators(ksp_matA(i),matA(i,i),matA(i,i),ierr)
if (flag_rav==1) then ! Raviart-thomas
   call KSPSetType(ksp_matA(i),KSPBCGS,ierr)
   call KSPGetPC(ksp_matA(i),pc,ierr)
   call PCSetType(pc,PCILU,ierr)
else ! MCFD or finite elements
   call KSPSetType(ksp_matA(i),KSPGMRES,ierr)
   call KSPGetPC(ksp_matA(i),pc,ierr)
   call PCSetType(pc,PCASM,ierr)
endif
call KSPSetFromOptions(ksp_matA(i),ierr)
call KSPSetUp(ksp_matA(i),ierr)



! Create eigenvectors in SLEPc format
call MatCreateVecs(matA(1,1),evec_slepc(1),PETSC_NULL_OBJECT,ierr)



! Create eigensolver
call EPSCreate(PETSC_COMM_WORLD,eps,ierr)

! Set operators of the eigensolver
call EPSSetOperators(eps,matB(1,1),matA(1,1),ierr) 
call EPSSetProblemType(eps,EPS_GNHEP,ierr)

! Set eigensolver parameters
nev=LMOD
ncv=max(4*LMOD,25)
mpd=max(4*LMOD,25)
call EPSSetDimensions(eps,nev,ncv,mpd,ierr)
which=EPS_LARGEST_MAGNITUDE
call EPSSetWhichEigenpairs(eps,which,ierr)
method=EPSKRYLOVSCHUR
call EPSSetType(eps,method,ierr)
! Additional options for accelerating
val=1.0
call EPSSetTarget (eps,val,ierr)
call EPSGetST (eps,st,ierr)
call STGetKSP (st,ksp,ierr)
if (flag_rav==1) then ! Raviart-thomas
   call KSPSetType(ksp,KSPBCGS,ierr)
   call KSPGetPC(ksp,pc,ierr)
   call PCSetType(pc,PCILU,ierr)
else ! MCFD or finite elements
   call KSPSetType(ksp,KSPGMRES,ierr)
   call KSPGetPC(ksp,pc,ierr)
   call PCSetType(pc,PCASM,ierr)
endif
tol=1e-6
maxit=999
call EPSSetTolerances(eps,tol,maxit,ierr)
call EPSSetFromOptions(eps,ierr)
call EPSSetUp(eps,ierr)

! Solve the eigenvalue problem 
call EPSSolve(eps,ierr)

! Get some information from the solver
! Iterations
call EPSGetIterationNumber(eps,its,ierr)
! Method used
call EPSGetType(eps,tname,ierr)
! Tolerances
call EPSGetTolerances(eps,tol,maxit,ierr)
! Converged eigenvalues
call EPSGetConverged(eps,nconv,ierr)

! Print the information of the solver
write(6,*) 'Generalized Eigenvalue Problem solved by SLEPc'
write(6,'(/a,I4)') ' Number of iterations of the method:',its
write(6,'(a,a)') ' Solution method: ',tname
write(6,'(a,I2)') ' Number of requested eigenvalues:',nev
write(6,'(a,1P,E10.4,a,I4)') ' Stopping condition: tol=',tol,', maxit=', maxit
write(6,'(a,I2)') ' Number of converged eigenpairs:',nconv




! Obtain the solution from the solver
call VecDuplicate(evec_slepc(1),xi,ierr)
do j=0,min(nconv,LMOD)-1 
   ! Get the eigenpair
   call EPSGetEigenpair(eps,j,kr,ki,evec_slepc(1),xi,ierr)
   ! Compute the error
   call EPSComputeError(eps,j,EPS_ERROR_RELATIVE,error_eigenpairs,ierr)
   ! Print eigenvalues and errors
   if (j==0) write(6,'(/a/a)') '           k          ||Ax-kx||/||kx||'&
             &,'   ----------------- ------------------'
   write(6,'(a,f12.6,a,e12.2e2)') '   ',kr,'       ',error_eigenpairs
   ! Store the eigenvalues
   FKEFFV(j+1)=kr 
   ! Store the fluxes
   call VecGetArrayF90(evec_slepc(1),array,ierr)
   do i=1,NGRP
      do k=1,LL4
         EV(k,i,j+1)=array(k+LL4*(i-1)) 
      enddo
   enddo
   call VecRestoreArrayF90(evec_slepc(1),array,ierr)
   ! Calculation of the adjoint
   call MatMultTranspose(matB(1,1),evec_slepc(1),xi,ierr)
   call KSPSolveTranspose(ksp_matA(1),xi,evec_slepc(1),ierr)
   ! Store the adjoint
   call VecGetArrayF90(evec_slepc(1),array,ierr)
   do i=1,NGRP
      do k=1,LL4
         AD(k,i,j+1)=array(k+LL4*(i-1)) 
      enddo
   enddo
   call VecRestoreArrayF90(evec_slepc(1),array,ierr)
enddo






! Destroy vec xi
call VecDestroy(xi,ierr)



! Destroy vectors evec_slepc()
call VecDestroy(evec_slepc(1),ierr)


! Destroy Eigenvalue solver
call EPSDestroy(eps,ierr)



end subroutine solver_slepc_with_upscattering_to_1

























! This subroutine destroys the system matrixes in SLEPc format 
subroutine destroy_matrixes_slepc(g1,g2,mat_a_or_b)
! g1: energy group g1 of Matrix(g1,g2)
! g2: energy group g2 of Matrix(g1,g2)
! mat_a_or_b = 0 or 1;   if it is 0, the matrix is matA; if it is 1, the matrix is matB 

! Global variables
integer :: g1,g2, mat_a_or_b





! Local variables
! SLEPc variables
PetscErrorCode ierr



! Destroy matrix
select case(mat_a_or_b)
case(0) ! matA
   call MatDestroy(matA(g1,g2),ierr)
case(1) ! matB
   call MatDestroy(matB(g1,g2),ierr)
end select


end subroutine destroy_matrixes_slepc
























! This subroutine fills matrixes used in Raviart-Thomas, in SLEPc format 
subroutine fill_matrixes_raviart_thomas(IPTRK,IPSYS,g1)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CYLIND,CHEX,DIAG,LOGY,LOGZ
CHARACTER TEXT10*10
INTEGER ISTATE(NSTATE), LL4, LL4W, LL4X, LL4Y, LL4Z, ITY, ITYPE
INTEGER IHEX, IELEM, ICOL, ICHX, ISPLH, ISEG, LX, LY, LZ, NBLOS
INTEGER MAXKN,ITYLCM
INTEGER, DIMENSION(:), POINTER :: MU
REAL, DIMENSION(:), ALLOCATABLE :: T
INTEGER, DIMENSION(:), POINTER :: IPW,IPX,IPY,IPZ, IZGLOB
INTEGER, DIMENSION(:,:), POINTER :: IPBW,IPBX,IPBY,IPBZ, KN
REAL, DIMENSION(:), POINTER :: A, DIFF
REAL, DIMENSION(:,:), POINTER :: BW,BX,BY,BZ
DOUBLE PRECISION, DIMENSION(:,:), POINTER :: CTRAN
TYPE(C_PTR) IPW_PTR,IPX_PTR,IPY_PTR,IPZ_PTR
TYPE(C_PTR) IPBW_PTR,IPBX_PTR,IPBY_PTR,IPBZ_PTR
TYPE(C_PTR) A_PTR, MU_PTR
TYPE(C_PTR) BW_PTR,BX_PTR,BY_PTR,BZ_PTR
TYPE(C_PTR) KN_PTR,IZGLOB_PTR,DIFF_PTR,CTRAN_PTR
integer i,i0,i1, j0, j, k, g, row, col, NUM, NELEM, NELEH, KN_col, KN_row 
real    DINV, SG

! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CYLIND=(ITYPE.EQ.3).OR.(ITYPE.EQ.6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
IDIM=1
IF((ITYPE.EQ.5).OR.(ITYPE.EQ.6).OR.(ITYPE.EQ.8)) IDIM=2
IF((ITYPE.EQ.7).OR.(ITYPE.EQ.9)) IDIM=3
IHEX=ISTATE(7)
DIAG=(ISTATE(8).EQ.1)
IELEM=ISTATE(9)
ICOL=ISTATE(10)
LL4=ISTATE(11)
ICHX=ISTATE(12)
ISPLH=ISTATE(13)
LX=ISTATE(14)
LY=ISTATE(15)
LZ=ISTATE(16)
ISEG=ISTATE(17)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
   LOGY=LL4Y.GT.0
   LOGZ=LL4Z.GT.0
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
   LOGY=IDIM.GT.1
   LOGZ=IDIM.GT.2
ENDIF




! Fill matrix matRw 
if (g1==1) then
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'IPBBW',IPBW_PTR)
   CALL LCMGPD(IPTRK,'WB',BW_PTR)
   CALL C_F_POINTER(IPBW_PTR,IPBW,(/ 2*IELEM,LL4W /))
   CALL C_F_POINTER(BW_PTR,BW,(/ 2*IELEM,LL4W /))
   do i=1,LL4W
      do j=1,2*IELEM
         k=IPBW(j,i)
         if(k/=0) then
            val= BW(j,i)
            call MatSetValue(matRw,i-1,k-1,val,INSERT_VALUES,ierr)
         endif
      enddo
   enddo
endif
endif

if (g1==1) then
! Fill matrix matRx 
CALL LCMGPD(IPTRK,'IPBBX',IPBX_PTR)
CALL LCMGPD(IPTRK,'XB',BX_PTR)
CALL C_F_POINTER(IPBX_PTR,IPBX,(/ 2*IELEM,LL4X /))
CALL C_F_POINTER(BX_PTR,BX,(/ 2*IELEM,LL4X /))
do i=1,LL4X
   do j=1,2*IELEM
      k=IPBX(j,i)
      if(k/=0) then
         val= BX(j,i)
         call MatSetValue(matRx,i-1,k-1,val,INSERT_VALUES,ierr)
      endif
   enddo
enddo
! Fill matrix matRy 
CALL LCMGPD(IPTRK,'IPBBY',IPBY_PTR)
CALL LCMGPD(IPTRK,'YB',BY_PTR)
CALL C_F_POINTER(IPBY_PTR,IPBY,(/ 2*IELEM,LL4Y /))
CALL C_F_POINTER(BY_PTR,BY,(/ 2*IELEM,LL4Y /))
do i=1,LL4Y
   do j=1,2*IELEM
      k=IPBY(j,i)
      if(k/=0) then
         val= BY(j,i)
         call MatSetValue(matRy,i-1,k-1,val,INSERT_VALUES,ierr)
      endif
   enddo
enddo
! Fill matrix matRz 
CALL LCMGPD(IPTRK,'IPBBZ',IPBZ_PTR)
CALL LCMGPD(IPTRK,'ZB',BZ_PTR)
CALL C_F_POINTER(IPBZ_PTR,IPBZ,(/ 2*IELEM,LL4Z /))
CALL C_F_POINTER(BZ_PTR,BZ,(/ 2*IELEM,LL4Z /))
do i=1,LL4Z
   do j=1,2*IELEM
      k=IPBZ(j,i)
      if(k/=0) then
         val= BZ(j,i)
         call MatSetValue(matRz,i-1,k-1,val,INSERT_VALUES,ierr)
      endif
   enddo
enddo
endif

! Fill matrix matAw 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'MUW',MU_PTR)
   CALL C_F_POINTER(MU_PTR,MU,(/ LL4W /))
   WRITE(TEXT10,'(3HWAA,2I3.3)') g1,g1
   CALL LCMGPD(IPSYS,TEXT10,A_PTR)
   CALL C_F_POINTER(A_PTR,A,(/ MU(LL4W)/))
   do i=1,LL4W
      if (i>1) then
         kend=MU(i)-MU(i-1)
         do k=1,kend
            j=i-k+1
            val=A(MU(i)-k+1)
            call MatSetValue(matAw(g1),i-1,j-1,val,INSERT_VALUES,ierr)
            call MatSetValue(matAw(g1),j-1,i-1,val,INSERT_VALUES,ierr)
         enddo
      else
         val=A(1)
         call MatSetValue(matAw(g1),i-1,i-1,val,INSERT_VALUES,ierr)
      endif
   enddo
endif

! Fill matrix matAx 
IF(DIAG) THEN
   CALL LCMGPD(IPTRK,'MUY',MU_PTR)
   WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
ELSE
   CALL LCMGPD(IPTRK,'MUX',MU_PTR)
   WRITE(TEXT10,'(3HXAA,2I3.3)') g1,g1
ENDIF
CALL C_F_POINTER(MU_PTR,MU,(/ LL4X /))
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4X)/))
do i=1,LL4X
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matAx(g1),i-1,j-1,val,INSERT_VALUES,ierr)
         call MatSetValue(matAx(g1),j-1,i-1,val,INSERT_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matAx(g1),i-1,i-1,val,INSERT_VALUES,ierr)
   endif
enddo

! Fill matrix matAy 
CALL LCMGPD(IPTRK,'MUY',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Y /))
WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Y)/))
do i=1,LL4Y
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matAy(g1),i-1,j-1,val,INSERT_VALUES,ierr)
         call MatSetValue(matAy(g1),j-1,i-1,val,INSERT_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matAy(g1),i-1,i-1,val,INSERT_VALUES,ierr)
   endif
enddo

! Fill matrix matAz 
CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Z /))
WRITE(TEXT10,'(3HZAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Z)/))
do i=1,LL4Z
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matAz(g1),i-1,j-1,val,INSERT_VALUES,ierr)
         call MatSetValue(matAz(g1),j-1,i-1,val,INSERT_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matAz(g1),i-1,i-1,val,INSERT_VALUES,ierr)
   endif
enddo

! Fill vecT
WRITE(TEXT10,'(3HTFA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ LL4F /))
do i=1,LL4F
   val=A(i)
   call VecSetValue(vecT(g1),i-1,val,INSERT_VALUES,ierr)
enddo





! Fill matrices matCwx, matCwy, matCxy
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'CTRAN',CTRAN_PTR)
   CALL LCMGPD(IPTRK,'KN',KN_PTR)
   CALL LCMGPD(IPTRK,'ZGLOB',IZGLOB_PTR)
   NBLOS=LX*LZ*ISPLH**2
   CALL C_F_POINTER(CTRAN_PTR,CTRAN,(/ (IELEM+1)*IELEM,(IELEM+1)*IELEM /))
   CALL C_F_POINTER(KN_PTR,KN,(/ NBLOS,3+6*(IELEM+2)*IELEM**2 /))
   CALL C_F_POINTER(IZGLOB_PTR,IZGLOB,(/ NBLOS /))
   NELEM=(IELEM+1)*IELEM
   NELEH=NELEM*IELEM
   write(TEXT10,'(1HA,2I3.3)') g1,g1
   CALL LCMGPD(IPSYS,'DIFF'//TEXT10,DIFF_PTR)
   CALL C_F_POINTER(DIFF_PTR,DIFF,(/ NBLOS /))
   ! matrix matCwx 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+2*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matCwx(g1),row-1,col-1,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCwy 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_row=KN(NUM,3+i)
               if(KN_row /= 0) then
                  row=ABS(KN_row)
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_col=KN(NUM,3+5*NELEH+j)
                     if (KN_col /= 0) then
                        col=ABS(KN_col)-LL4W-LL4X
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matCwy(g1),row-1,col-1,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCxy 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+4*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W-LL4X
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+3*NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)-LL4W
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matCxy(g1),row-1,col-1,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
endif



end subroutine fill_matrixes_raviart_thomas
























! This subroutine determines the preallocation number ofr matrixes used in Raviart-Thomas, in SLEPc format 
subroutine prealloc_matrixes_raviart_thomas(IPTRK,IPSYS,npre)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! npre: maximum number of non-null colums 




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer npre




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CYLIND,CHEX,DIAG,LOGY,LOGZ
CHARACTER TEXT10*10
INTEGER ISTATE(NSTATE), LL4, LL4W, LL4X, LL4Y, LL4Z, ITY, ITYPE
INTEGER IHEX, IELEM, ICOL, ICHX, ISPLH, ISEG, LX, LY, LZ, NBLOS
INTEGER MAXKN,ITYLCM
INTEGER, DIMENSION(:), POINTER :: MU
REAL, DIMENSION(:), ALLOCATABLE :: T
INTEGER, DIMENSION(:), POINTER :: IPW,IPX,IPY,IPZ, IZGLOB
INTEGER, DIMENSION(:,:), POINTER :: IPBW,IPBX,IPBY,IPBZ, KN
REAL, DIMENSION(:), POINTER :: A, DIFF
REAL, DIMENSION(:,:), POINTER :: BW,BX,BY,BZ
DOUBLE PRECISION, DIMENSION(:,:), POINTER :: CTRAN
TYPE(C_PTR) IPW_PTR,IPX_PTR,IPY_PTR,IPZ_PTR
TYPE(C_PTR) IPBW_PTR,IPBX_PTR,IPBY_PTR,IPBZ_PTR
TYPE(C_PTR) A_PTR, MU_PTR
TYPE(C_PTR) BW_PTR,BX_PTR,BY_PTR,BZ_PTR
TYPE(C_PTR) KN_PTR,IZGLOB_PTR,DIFF_PTR,CTRAN_PTR
integer i,i0,i1, j0, j, k, g, row, col, NUM, NELEM, NELEH, KN_col, KN_row 
real    DINV, SG
integer, dimension(:), allocatable :: npre_rows




! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CYLIND=(ITYPE.EQ.3).OR.(ITYPE.EQ.6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
IDIM=1
IF((ITYPE.EQ.5).OR.(ITYPE.EQ.6).OR.(ITYPE.EQ.8)) IDIM=2
IF((ITYPE.EQ.7).OR.(ITYPE.EQ.9)) IDIM=3
IHEX=ISTATE(7)
DIAG=(ISTATE(8).EQ.1)
IELEM=ISTATE(9)
ICOL=ISTATE(10)
LL4=ISTATE(11)
ICHX=ISTATE(12)
ISPLH=ISTATE(13)
LX=ISTATE(14)
LY=ISTATE(15)
LZ=ISTATE(16)
ISEG=ISTATE(17)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
   LOGY=LL4Y.GT.0
   LOGZ=LL4Z.GT.0
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
   LOGY=IDIM.GT.1
   LOGZ=IDIM.GT.2
ENDIF



! Allocate npre_rows
allocate(npre_rows(LL4))
! Initialuze npre_rows to 0
npre_rows=0



!  matrix matRw 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'IPBBW',IPBW_PTR)
   CALL C_F_POINTER(IPBW_PTR,IPBW,(/ 2*IELEM,LL4W /))
   do i=1,LL4W
      do j=1,2*IELEM
         k=IPBW(j,i)
         if(k/=0) then
            npre_rows(k)=npre_rows(k)+1
            npre_rows(i+LL4F)=npre_rows(i+LL4F)+1
         endif
      enddo
   enddo
endif

!  matrix matRx 
CALL LCMGPD(IPTRK,'IPBBX',IPBX_PTR)
CALL C_F_POINTER(IPBX_PTR,IPBX,(/ 2*IELEM,LL4X /))
do i=1,LL4X
   do j=1,2*IELEM
      k=IPBX(j,i)
      if(k/=0) then
         npre_rows(k)=npre_rows(k)+1
         npre_rows(i+LL4F+LL4W)=npre_rows(i+LL4F+LL4W)+1
      endif
   enddo
enddo
!  matrix matRy 
CALL LCMGPD(IPTRK,'IPBBY',IPBY_PTR)
CALL C_F_POINTER(IPBY_PTR,IPBY,(/ 2*IELEM,LL4Y /))
do i=1,LL4Y
   do j=1,2*IELEM
      k=IPBY(j,i)
      if(k/=0) then
         npre_rows(k)=npre_rows(k)+1
         npre_rows(i+LL4F+LL4W+LL4X)=npre_rows(i+LL4F+LL4W+LL4X)+1
      endif
   enddo
enddo
!  matrix matRz 
CALL LCMGPD(IPTRK,'IPBBZ',IPBZ_PTR)
CALL C_F_POINTER(IPBZ_PTR,IPBZ,(/ 2*IELEM,LL4Z /))
do i=1,LL4Z
   do j=1,2*IELEM
      k=IPBZ(j,i)
      if(k/=0) then
         npre_rows(k)=npre_rows(k)+1
         npre_rows(i+LL4F+LL4W+LL4X+LL4Y)=npre_rows(i+LL4F+LL4W+LL4X+LL4Y)+1
      endif
   enddo
enddo

!  matrix matAw 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'MUW',MU_PTR)
   CALL C_F_POINTER(MU_PTR,MU,(/ LL4W /))
   do i=1,LL4W
      if (i>1) then
         kend=MU(i)-MU(i-1)
      else
         kend=1
      endif
      npre_rows(i+LL4F)=npre_rows(i+LL4F)+2*(kend-1)+1
   enddo
endif

!  matrix matAx 
IF(DIAG) THEN
   CALL LCMGPD(IPTRK,'MUY',MU_PTR)
ELSE
   CALL LCMGPD(IPTRK,'MUX',MU_PTR)
ENDIF
CALL C_F_POINTER(MU_PTR,MU,(/ LL4X /))
do i=1,LL4X
   if (i>1) then
      kend=MU(i)-MU(i-1)
   else
      kend=1
   endif
   npre_rows(i+LL4F+LL4W)=npre_rows(i+LL4F+LL4W)+2*(kend-1)+1
enddo

!  matrix matAy 
CALL LCMGPD(IPTRK,'MUY',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Y /))
do i=1,LL4Y
   if (i>1) then
      kend=MU(i)-MU(i-1)
   else
      kend=1
   endif
   npre_rows(i+LL4F+LL4W+LL4X)=npre_rows(i+LL4F+LL4W+LL4X)+2*(kend-1)+1
enddo

!  matrix matAz 
CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Z /))
do i=1,LL4Z
   if (i>1) then
      kend=MU(i)-MU(i-1)
   else
      kend=1
   endif
   npre_rows(i+LL4F+LL4W+LL4X+LL4Y)=npre_rows(i+LL4F+LL4W+LL4X+LL4Y)+2*(kend-1)+1
enddo

!  vecT
do i=1,LL4F
   npre_rows(i)=npre_rows(i)+1
enddo





!  matrices matCwx, matCwy, matCxy
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'CTRAN',CTRAN_PTR)
   CALL LCMGPD(IPTRK,'KN',KN_PTR)
   CALL LCMGPD(IPTRK,'ZGLOB',IZGLOB_PTR)
   NBLOS=LX*LZ*ISPLH**2
   CALL C_F_POINTER(CTRAN_PTR,CTRAN,(/ (IELEM+1)*IELEM,(IELEM+1)*IELEM /))
   CALL C_F_POINTER(KN_PTR,KN,(/ NBLOS,3+6*(IELEM+2)*IELEM**2 /))
   CALL C_F_POINTER(IZGLOB_PTR,IZGLOB,(/ NBLOS /))
   NELEM=(IELEM+1)*IELEM
   NELEH=NELEM*IELEM
   ! matrix matCwx 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+2*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)
                        npre_rows(row+LL4F)=npre_rows(row+LL4F)+1
                        npre_rows(col+LL4F+LL4W)=npre_rows(col+LL4F+LL4W)+1
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCwy 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_row=KN(NUM,3+i)
               if(KN_row /= 0) then
                  row=ABS(KN_row)
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_col=KN(NUM,3+5*NELEH+j)
                     if (KN_col /= 0) then
                        col=ABS(KN_col)-LL4W-LL4X
                        npre_rows(row+LL4F)=npre_rows(row+LL4F)+1
                        npre_rows(col+LL4F+LL4W+LL4X)=npre_rows(col+LL4F+LL4W+LL4X)+1
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCxy 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+4*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W-LL4X
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+3*NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)-LL4W
                        npre_rows(row+LL4F+LL4W)=npre_rows(row+LL4F+LL4W)+1
                        npre_rows(col+LL4F+LL4W+LL4X)=npre_rows(col+LL4F+LL4W+LL4X)+1
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
endif


! Calculate npre
npre=maxval(npre_rows)


! Deallocate npre_rows
deallocate(npre_rows)


end subroutine prealloc_matrixes_raviart_thomas
























! This subroutine creates matrixes used in Raviart-Thomas, in SLEPc format 
subroutine create_matrixes_raviart_thomas(IPTRK,IPSYS,g1,npre)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 
! npre: maximum number of non-null colums 




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1, npre




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CHEX
INTEGER ISTATE(NSTATE) , ITYPE

! SLEPc variables
PetscErrorCode ierr




! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
LL4=ISTATE(11)
ICHX=ISTATE(12)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
ENDIF




! Create matrixes
IF(CHEX) THEN
   ! matCwx
   call MatCreate(PETSC_COMM_WORLD,matCwx(g1),ierr)
   call MatSetSizes(matCwx(g1),PETSC_DECIDE,PETSC_DECIDE,LL4W,LL4X,ierr)
   call MatSetFromOptions(matCwx(g1),ierr)
   call MatSeqAIJSetPreallocation(matCwx(g1),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matCwx(g1),ierr)
   ! matCwy
   call MatCreate(PETSC_COMM_WORLD,matCwy(g1),ierr)
   call MatSetSizes(matCwy(g1),PETSC_DECIDE,PETSC_DECIDE,LL4W,LL4Y,ierr)
   call MatSetFromOptions(matCwy(g1),ierr)
   call MatSeqAIJSetPreallocation(matCwy(g1),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matCwy(g1),ierr)
   ! matCxy
   call MatCreate(PETSC_COMM_WORLD,matCxy(g1),ierr)
   call MatSetSizes(matCxy(g1),PETSC_DECIDE,PETSC_DECIDE,LL4X,LL4Y,ierr)
   call MatSetFromOptions(matCxy(g1),ierr)
   call MatSeqAIJSetPreallocation(matCxy(g1),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matCxy(g1),ierr)
   ! matAw
   call MatCreate(PETSC_COMM_WORLD,matAw(g1),ierr)
   call MatSetSizes(matAw(g1),PETSC_DECIDE,PETSC_DECIDE,LL4W,LL4W,ierr)
   call MatSetFromOptions(matAw(g1),ierr)
   call MatSeqAIJSetPreallocation(matAw(g1),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matAw(g1),ierr)
   if (g1==1) then
      ! matRw
      call MatCreate(PETSC_COMM_WORLD,matRw,ierr)
      call MatSetSizes(matRw,PETSC_DECIDE,PETSC_DECIDE,LL4W,LL4F,ierr)
      call MatSetFromOptions(matRw,ierr)
      call MatSeqAIJSetPreallocation(matRw,npre,PETSC_NULL_INTEGER,ierr)
      call MatSetUp(matRw,ierr)
   endif
endif
! matAx
call MatCreate(PETSC_COMM_WORLD,matAx(g1),ierr)
call MatSetSizes(matAx(g1),PETSC_DECIDE,PETSC_DECIDE,LL4X,LL4X,ierr)
call MatSetFromOptions(matAx(g1),ierr)
call MatSeqAIJSetPreallocation(matAx(g1),npre,PETSC_NULL_INTEGER,ierr)
call MatSetUp(matAx(g1),ierr)
! matAy
call MatCreate(PETSC_COMM_WORLD,matAy(g1),ierr)
call MatSetSizes(matAy(g1),PETSC_DECIDE,PETSC_DECIDE,LL4Y,LL4Y,ierr)
call MatSetFromOptions(matAy(g1),ierr)
call MatSeqAIJSetPreallocation(matAy(g1),npre,PETSC_NULL_INTEGER,ierr)
call MatSetUp(matAy(g1),ierr)
! matAz
call MatCreate(PETSC_COMM_WORLD,matAz(g1),ierr)
call MatSetSizes(matAz(g1),PETSC_DECIDE,PETSC_DECIDE,LL4Z,LL4Z,ierr)
call MatSetFromOptions(matAz(g1),ierr)
call MatSeqAIJSetPreallocation(matAz(g1),npre,PETSC_NULL_INTEGER,ierr)
call MatSetUp(matAz(g1),ierr)
! VecT
call VecCreate(PETSC_COMM_WORLD,vecT(g1),ierr)
call VecSetSizes(vecT(g1),PETSC_DECIDE,LL4F,ierr)
call VecSetFromOptions(vecT(g1),ierr)

if (g1==1) then
   ! matRx
   call MatCreate(PETSC_COMM_WORLD,matRx,ierr)
   call MatSetSizes(matRx,PETSC_DECIDE,PETSC_DECIDE,LL4X,LL4F,ierr)
   call MatSetFromOptions(matRx,ierr)
   call MatSeqAIJSetPreallocation(matRx,npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matRx,ierr)
   ! matRy
   call MatCreate(PETSC_COMM_WORLD,matRy,ierr)
   call MatSetSizes(matRy,PETSC_DECIDE,PETSC_DECIDE,LL4Y,LL4F,ierr)
   call MatSetFromOptions(matRy,ierr)
   call MatSeqAIJSetPreallocation(matRy,npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matRy,ierr)
   ! matRz
   call MatCreate(PETSC_COMM_WORLD,matRz,ierr)
   call MatSetSizes(matRz,PETSC_DECIDE,PETSC_DECIDE,LL4Z,LL4F,ierr)
   call MatSetFromOptions(matRz,ierr)
   call MatSeqAIJSetPreallocation(matRz,npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matRz,ierr)
endif

end subroutine create_matrixes_raviart_thomas

























! This subroutine assembles matrixes used in Raviart-Thomas, in SLEPc format 
subroutine assemble_matrixes_raviart_thomas(IPTRK,IPSYS,g1)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1 




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CHEX
INTEGER ISTATE(NSTATE) , ITYPE

! SLEPc variables
PetscErrorCode ierr



! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)



! Assemble matrixes

! Begin assembly
IF(CHEX) THEN
   ! matCwx
   call MatAssemblyBegin(matCwx(g1),MAT_FINAL_ASSEMBLY,ierr)
   ! matCwy
   call MatAssemblyBegin(matCwy(g1),MAT_FINAL_ASSEMBLY,ierr)
   ! matCxy
   call MatAssemblyBegin(matCxy(g1),MAT_FINAL_ASSEMBLY,ierr)
   ! matAw
   call MatAssemblyBegin(matAw(g1),MAT_FINAL_ASSEMBLY,ierr)
   ! matRw
   if (g1==1) call MatAssemblyBegin(matRw,MAT_FINAL_ASSEMBLY,ierr)
endif
! matAx
call MatAssemblyBegin(matAx(g1),MAT_FINAL_ASSEMBLY,ierr)
! matAy
call MatAssemblyBegin(matAy(g1),MAT_FINAL_ASSEMBLY,ierr)
! matAz
call MatAssemblyBegin(matAz(g1),MAT_FINAL_ASSEMBLY,ierr)
! VecT
call VecAssemblyBegin(vecT(g1),ierr)
! matRx
if (g1==1) call MatAssemblyBegin(matRx,MAT_FINAL_ASSEMBLY,ierr)
! matRy
if (g1==1) call MatAssemblyBegin(matRy,MAT_FINAL_ASSEMBLY,ierr)
! matRz
if (g1==1) call MatAssemblyBegin(matRz,MAT_FINAL_ASSEMBLY,ierr)

! End assembly
IF(CHEX) THEN
   ! matCwx
   call MatAssemblyEnd(matCwx(g1),MAT_FINAL_ASSEMBLY,ierr)
   ! matCwy
   call MatAssemblyEnd(matCwy(g1),MAT_FINAL_ASSEMBLY,ierr)
   ! matCxy
   call MatAssemblyEnd(matCxy(g1),MAT_FINAL_ASSEMBLY,ierr)
   ! matAw
   call MatAssemblyEnd(matAw(g1),MAT_FINAL_ASSEMBLY,ierr)
   ! matRw
   if (g1==1) call MatAssemblyEnd(matRw,MAT_FINAL_ASSEMBLY,ierr)
endif
! matAx
call MatAssemblyEnd(matAx(g1),MAT_FINAL_ASSEMBLY,ierr)
! matAy
call MatAssemblyEnd(matAy(g1),MAT_FINAL_ASSEMBLY,ierr)
! matAz
call MatAssemblyEnd(matAz(g1),MAT_FINAL_ASSEMBLY,ierr)
! VecT
call VecAssemblyEnd(vecT(g1),ierr)
! matRx
if (g1==1) call MatAssemblyEnd(matRx,MAT_FINAL_ASSEMBLY,ierr)
! matRy
if (g1==1) call MatAssemblyEnd(matRy,MAT_FINAL_ASSEMBLY,ierr)
! matRz
if (g1==1) call MatAssemblyEnd(matRz,MAT_FINAL_ASSEMBLY,ierr)



end subroutine assemble_matrixes_raviart_thomas
























! This subroutine destroys matrixes used in Raviart-Thomas, in SLEPc format 
subroutine destroy_matrixes_raviart_thomas(IPTRK,IPSYS,g1)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1 




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CHEX
INTEGER ISTATE(NSTATE) , ITYPE

! SLEPc variables
PetscErrorCode ierr



! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)




! Destroy matrixes
IF(CHEX) THEN
   ! matCwx
   call MatDestroy(matCwx(g1),ierr)
   ! matCwy
   call MatDestroy(matCwy(g1),ierr)
   ! matCxy
   call MatDestroy(matCxy(g1),ierr)
   ! matAw
   call MatDestroy(matAw(g1),ierr)
   ! matRw
   if (g1==1) call MatDestroy(matRw,ierr)
endif
! matAx
call MatDestroy(matAx(g1),ierr)
! matAy
call MatDestroy(matAy(g1),ierr)
! matAz
call MatDestroy(matAz(g1),ierr)
! VecT
call VecDestroy(vecT(g1),ierr)
! matRx
if (g1==1) call MatDestroy(matRx,ierr)
! matRy
if (g1==1) call MatDestroy(matRy,ierr)
! matRz
if (g1==1) call MatDestroy(matRz,ierr)


end subroutine destroy_matrixes_raviart_thomas
























! This subroutine fills matrix A used in Raviart-Thomas, in SLEPc format 
subroutine fill_matrixA_raviart_thomas(IPTRK,IPSYS,g1)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CYLIND,CHEX,DIAG,LOGY,LOGZ
CHARACTER TEXT10*10
INTEGER ISTATE(NSTATE), LL4, LL4W, LL4X, LL4Y, LL4Z, ITY, ITYPE
INTEGER IHEX, IELEM, ICOL, ICHX, ISPLH, ISEG, LX, LY, LZ, NBLOS
INTEGER MAXKN,ITYLCM
INTEGER, DIMENSION(:), POINTER :: MU
REAL, DIMENSION(:), ALLOCATABLE :: T
INTEGER, DIMENSION(:), POINTER :: IPW,IPX,IPY,IPZ, IZGLOB
INTEGER, DIMENSION(:,:), POINTER :: IPBW,IPBX,IPBY,IPBZ, KN
REAL, DIMENSION(:), POINTER :: A, DIFF
REAL, DIMENSION(:,:), POINTER :: BW,BX,BY,BZ
DOUBLE PRECISION, DIMENSION(:,:), POINTER :: CTRAN
TYPE(C_PTR) IPW_PTR,IPX_PTR,IPY_PTR,IPZ_PTR
TYPE(C_PTR) IPBW_PTR,IPBX_PTR,IPBY_PTR,IPBZ_PTR
TYPE(C_PTR) A_PTR, MU_PTR
TYPE(C_PTR) BW_PTR,BX_PTR,BY_PTR,BZ_PTR
TYPE(C_PTR) KN_PTR,IZGLOB_PTR,DIFF_PTR,CTRAN_PTR
integer i,i0,i1, j0, j, k, g, row, col, NUM, NELEM, NELEH, KN_col, KN_row 
real    DINV, SG

! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CYLIND=(ITYPE.EQ.3).OR.(ITYPE.EQ.6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
IDIM=1
IF((ITYPE.EQ.5).OR.(ITYPE.EQ.6).OR.(ITYPE.EQ.8)) IDIM=2
IF((ITYPE.EQ.7).OR.(ITYPE.EQ.9)) IDIM=3
IHEX=ISTATE(7)
DIAG=(ISTATE(8).EQ.1)
IELEM=ISTATE(9)
ICOL=ISTATE(10)
LL4=ISTATE(11)
ICHX=ISTATE(12)
ISPLH=ISTATE(13)
LX=ISTATE(14)
LY=ISTATE(15)
LZ=ISTATE(16)
ISEG=ISTATE(17)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
   LOGY=LL4Y.GT.0
   LOGZ=LL4Z.GT.0
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
   LOGY=IDIM.GT.1
   LOGZ=IDIM.GT.2
ENDIF




! Fill matrix matRw 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'IPBBW',IPBW_PTR)
   CALL LCMGPD(IPTRK,'WB',BW_PTR)
   CALL C_F_POINTER(IPBW_PTR,IPBW,(/ 2*IELEM,LL4W /))
   CALL C_F_POINTER(BW_PTR,BW,(/ 2*IELEM,LL4W /))
   do i=1,LL4W
      do j=1,2*IELEM
         k=IPBW(j,i)
         if(k/=0) then
            val= BW(j,i)
            call MatSetValue(matA(g1,g1),i+LL4F-1,k-1,val,ADD_VALUES,ierr)
            call MatSetValue(matA(g1,g1),k-1,i+LL4F-1,val,ADD_VALUES,ierr)
         endif
      enddo
   enddo
endif

! Fill matrix matRx 
CALL LCMGPD(IPTRK,'IPBBX',IPBX_PTR)
CALL LCMGPD(IPTRK,'XB',BX_PTR)
CALL C_F_POINTER(IPBX_PTR,IPBX,(/ 2*IELEM,LL4X /))
CALL C_F_POINTER(BX_PTR,BX,(/ 2*IELEM,LL4X /))
do i=1,LL4X
   do j=1,2*IELEM
      k=IPBX(j,i)
      if(k/=0) then
         val= BX(j,i)
         call MatSetValue(matA(g1,g1),i+LL4F+LL4W-1,k-1,val,ADD_VALUES,ierr)
         call MatSetValue(matA(g1,g1),k-1,i+LL4F+LL4W-1,val,ADD_VALUES,ierr)
      endif
   enddo
enddo
! Fill matrix matRy 
CALL LCMGPD(IPTRK,'IPBBY',IPBY_PTR)
CALL LCMGPD(IPTRK,'YB',BY_PTR)
CALL C_F_POINTER(IPBY_PTR,IPBY,(/ 2*IELEM,LL4Y /))
CALL C_F_POINTER(BY_PTR,BY,(/ 2*IELEM,LL4Y /))
do i=1,LL4Y
   do j=1,2*IELEM
      k=IPBY(j,i)
      if(k/=0) then
         val= BY(j,i)
         call MatSetValue(matA(g1,g1),i+LL4F+LL4W+LL4X-1,k-1,val,ADD_VALUES,ierr)
         call MatSetValue(matA(g1,g1),k-1,i+LL4F+LL4W+LL4X-1,val,ADD_VALUES,ierr)
      endif
   enddo
enddo
! Fill matrix matRz 
CALL LCMGPD(IPTRK,'IPBBZ',IPBZ_PTR)
CALL LCMGPD(IPTRK,'ZB',BZ_PTR)
CALL C_F_POINTER(IPBZ_PTR,IPBZ,(/ 2*IELEM,LL4Z /))
CALL C_F_POINTER(BZ_PTR,BZ,(/ 2*IELEM,LL4Z /))
do i=1,LL4Z
   do j=1,2*IELEM
      k=IPBZ(j,i)
      if(k/=0) then
         val= BZ(j,i)
         call MatSetValue(matA(g1,g1),i+LL4F+LL4W+LL4X+LL4Y-1,k-1,val,ADD_VALUES,ierr)
         call MatSetValue(matA(g1,g1),k-1,i+LL4F+LL4W+LL4X+LL4Y-1,val,ADD_VALUES,ierr)
      endif
   enddo
enddo

! Fill matrix matAw 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'MUW',MU_PTR)
   CALL C_F_POINTER(MU_PTR,MU,(/ LL4W /))
   WRITE(TEXT10,'(3HWAA,2I3.3)') g1,g1
   CALL LCMGPD(IPSYS,TEXT10,A_PTR)
   CALL C_F_POINTER(A_PTR,A,(/ MU(LL4W)/))
   do i=1,LL4W
      if (i>1) then
         kend=MU(i)-MU(i-1)
         do k=1,kend
            j=i-k+1
            val=A(MU(i)-k+1)
            call MatSetValue(matA(g1,g1),i+LL4F-1,j+LL4F-1,val,ADD_VALUES,ierr)
            if (i/=j) call MatSetValue(matA(g1,g1),j+LL4F-1,i+LL4F-1,val,ADD_VALUES,ierr)
         enddo
      else
         val=A(1)
         call MatSetValue(matA(g1,g1),i+LL4F-1,i+LL4F-1,val,ADD_VALUES,ierr)
      endif
   enddo
endif

! Fill matrix matAx 
IF(DIAG) THEN
   CALL LCMGPD(IPTRK,'MUY',MU_PTR)
   WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
ELSE
   CALL LCMGPD(IPTRK,'MUX',MU_PTR)
   WRITE(TEXT10,'(3HXAA,2I3.3)') g1,g1
ENDIF
CALL C_F_POINTER(MU_PTR,MU,(/ LL4X /))
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4X)/))
do i=1,LL4X
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matA(g1,g1),i+LL4F+LL4W-1,j+LL4F+LL4W-1,val,ADD_VALUES,ierr)
         if (i/=j) call MatSetValue(matA(g1,g1),j+LL4F+LL4W-1,i+LL4F+LL4W-1,val,ADD_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matA(g1,g1),i+LL4F+LL4W-1,i+LL4F+LL4W-1,val,ADD_VALUES,ierr)
   endif
enddo

! Fill matrix matAy 
CALL LCMGPD(IPTRK,'MUY',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Y /))
WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Y)/))
do i=1,LL4Y
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matA(g1,g1),i+LL4F+LL4W+LL4X-1,j+LL4F+LL4W+LL4X-1,val,ADD_VALUES,ierr)
         if (i/=j) call MatSetValue(matA(g1,g1),j+LL4F+LL4W+LL4X-1,i+LL4F+LL4W+LL4X-1,val,ADD_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matA(g1,g1),i+LL4F+LL4W+LL4X-1,i+LL4F+LL4W+LL4X-1,val,ADD_VALUES,ierr)
   endif
enddo

! Fill matrix matAz 
CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Z /))
WRITE(TEXT10,'(3HZAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Z)/))
do i=1,LL4Z
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matA(g1,g1),i+LL4F+LL4W+LL4X+LL4Y-1,j+LL4F+LL4W+LL4X+LL4Y-1,val,ADD_VALUES,ierr)
         if (i/=j) call MatSetValue(matA(g1,g1),j+LL4F+LL4W+LL4X+LL4Y-1,i+LL4F+LL4W+LL4X+LL4Y-1,val,ADD_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matA(g1,g1),i+LL4F+LL4W+LL4X+LL4Y-1,i+LL4F+LL4W+LL4X+LL4Y-1,val,ADD_VALUES,ierr)
   endif
enddo

! Fill vecT
WRITE(TEXT10,'(3HTFA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ LL4F /))
do i=1,LL4F
   val=A(i)
   call MatSetValue(matA(g1,g1),i-1,i-1,val,ADD_VALUES,ierr)
enddo





! Fill matrices matCwx, matCwy, matCxy
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'CTRAN',CTRAN_PTR)
   CALL LCMGPD(IPTRK,'KN',KN_PTR)
   CALL LCMGPD(IPTRK,'ZGLOB',IZGLOB_PTR)
   NBLOS=LX*LZ*ISPLH**2
   CALL C_F_POINTER(CTRAN_PTR,CTRAN,(/ (IELEM+1)*IELEM,(IELEM+1)*IELEM /))
   CALL C_F_POINTER(KN_PTR,KN,(/ NBLOS,3+6*(IELEM+2)*IELEM**2 /))
   CALL C_F_POINTER(IZGLOB_PTR,IZGLOB,(/ NBLOS /))
   NELEM=(IELEM+1)*IELEM
   NELEH=NELEM*IELEM
   write(TEXT10,'(1HA,2I3.3)') g1,g1
   CALL LCMGPD(IPSYS,'DIFF'//TEXT10,DIFF_PTR)
   CALL C_F_POINTER(DIFF_PTR,DIFF,(/ NBLOS /))
   ! matrix matCwx 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+2*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matA(g1,g1),row+LL4F-1,col+LL4F+LL4W-1,val,ADD_VALUES,ierr)
                        call MatSetValue(matA(g1,g1),col+LL4F+LL4W-1,row+LL4F-1,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCwy 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_row=KN(NUM,3+i)
               if(KN_row /= 0) then
                  row=ABS(KN_row)
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_col=KN(NUM,3+5*NELEH+j)
                     if (KN_col /= 0) then
                        col=ABS(KN_col)-LL4W-LL4X
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matA(g1,g1),row+LL4F-1,col+LL4F+LL4W+LL4X-1,val,ADD_VALUES,ierr)
                        call MatSetValue(matA(g1,g1),col+LL4F+LL4W+LL4X-1,row+LL4F-1,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCxy 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+4*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W-LL4X
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+3*NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)-LL4W
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matA(g1,g1),row+LL4F+LL4W-1,col+LL4F+LL4W+LL4X-1,val,ADD_VALUES,ierr)
                        call MatSetValue(matA(g1,g1),col+LL4F+LL4W+LL4X-1,row+LL4F+LL4W-1,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
endif



end subroutine fill_matrixA_raviart_thomas
























! This subroutine fills matrix A used in Raviart-Thomas, in SLEPc format, for
! including upscattering 
subroutine fill_matrixA_raviart_thomas_several_groups(IPTRK,IPSYS,g1,gups)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 
! gups: first energy group containing upscattring terms




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1, gups




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CYLIND,CHEX,DIAG,LOGY,LOGZ
CHARACTER TEXT10*10
INTEGER ISTATE(NSTATE), LL4, LL4W, LL4X, LL4Y, LL4Z, ITY, ITYPE
INTEGER IHEX, IELEM, ICOL, ICHX, ISPLH, ISEG, LX, LY, LZ, NBLOS
INTEGER MAXKN,ITYLCM
INTEGER, DIMENSION(:), POINTER :: MU
REAL, DIMENSION(:), ALLOCATABLE :: T
INTEGER, DIMENSION(:), POINTER :: IPW,IPX,IPY,IPZ, IZGLOB
INTEGER, DIMENSION(:,:), POINTER :: IPBW,IPBX,IPBY,IPBZ, KN
REAL, DIMENSION(:), POINTER :: A, DIFF
REAL, DIMENSION(:,:), POINTER :: BW,BX,BY,BZ
DOUBLE PRECISION, DIMENSION(:,:), POINTER :: CTRAN
TYPE(C_PTR) IPW_PTR,IPX_PTR,IPY_PTR,IPZ_PTR
TYPE(C_PTR) IPBW_PTR,IPBX_PTR,IPBY_PTR,IPBZ_PTR
TYPE(C_PTR) A_PTR, MU_PTR
TYPE(C_PTR) BW_PTR,BX_PTR,BY_PTR,BZ_PTR
TYPE(C_PTR) KN_PTR,IZGLOB_PTR,DIFF_PTR,CTRAN_PTR
integer i,i0,i1, j0, j, k, g, row, col, NUM, NELEM, NELEH, KN_col, KN_row 
real    DINV, SG

! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CYLIND=(ITYPE.EQ.3).OR.(ITYPE.EQ.6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
IDIM=1
IF((ITYPE.EQ.5).OR.(ITYPE.EQ.6).OR.(ITYPE.EQ.8)) IDIM=2
IF((ITYPE.EQ.7).OR.(ITYPE.EQ.9)) IDIM=3
IHEX=ISTATE(7)
DIAG=(ISTATE(8).EQ.1)
IELEM=ISTATE(9)
ICOL=ISTATE(10)
LL4=ISTATE(11)
ICHX=ISTATE(12)
ISPLH=ISTATE(13)
LX=ISTATE(14)
LY=ISTATE(15)
LZ=ISTATE(16)
ISEG=ISTATE(17)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
   LOGY=LL4Y.GT.0
   LOGZ=LL4Z.GT.0
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
   LOGY=IDIM.GT.1
   LOGZ=IDIM.GT.2
ENDIF




! Fill matrix matRw 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'IPBBW',IPBW_PTR)
   CALL LCMGPD(IPTRK,'WB',BW_PTR)
   CALL C_F_POINTER(IPBW_PTR,IPBW,(/ 2*IELEM,LL4W /))
   CALL C_F_POINTER(BW_PTR,BW,(/ 2*IELEM,LL4W /))
   do i=1,LL4W
      do j=1,2*IELEM
         k=IPBW(j,i)
         if(k/=0) then
            val= BW(j,i)
            call MatSetValue(matA(gups,gups),i+LL4F-1+(g1-gups)*LL4,k-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
            call MatSetValue(matA(gups,gups),k-1+(g1-gups)*LL4,i+LL4F-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
         endif
      enddo
   enddo
endif

! Fill matrix matRx 
CALL LCMGPD(IPTRK,'IPBBX',IPBX_PTR)
CALL LCMGPD(IPTRK,'XB',BX_PTR)
CALL C_F_POINTER(IPBX_PTR,IPBX,(/ 2*IELEM,LL4X /))
CALL C_F_POINTER(BX_PTR,BX,(/ 2*IELEM,LL4X /))
do i=1,LL4X
   do j=1,2*IELEM
      k=IPBX(j,i)
      if(k/=0) then
         val= BX(j,i)
         call MatSetValue(matA(gups,gups),i+LL4F+LL4W-1+(g1-gups)*LL4,k-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
         call MatSetValue(matA(gups,gups),k-1+(g1-gups)*LL4,i+LL4F+LL4W-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
      endif
   enddo
enddo
! Fill matrix matRy 
CALL LCMGPD(IPTRK,'IPBBY',IPBY_PTR)
CALL LCMGPD(IPTRK,'YB',BY_PTR)
CALL C_F_POINTER(IPBY_PTR,IPBY,(/ 2*IELEM,LL4Y /))
CALL C_F_POINTER(BY_PTR,BY,(/ 2*IELEM,LL4Y /))
do i=1,LL4Y
   do j=1,2*IELEM
      k=IPBY(j,i)
      if(k/=0) then
         val= BY(j,i)
         call MatSetValue(matA(gups,gups),i+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,k-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
         call MatSetValue(matA(gups,gups),k-1+(g1-gups)*LL4,i+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
      endif
   enddo
enddo
! Fill matrix matRz 
CALL LCMGPD(IPTRK,'IPBBZ',IPBZ_PTR)
CALL LCMGPD(IPTRK,'ZB',BZ_PTR)
CALL C_F_POINTER(IPBZ_PTR,IPBZ,(/ 2*IELEM,LL4Z /))
CALL C_F_POINTER(BZ_PTR,BZ,(/ 2*IELEM,LL4Z /))
do i=1,LL4Z
   do j=1,2*IELEM
      k=IPBZ(j,i)
      if(k/=0) then
         val= BZ(j,i)
         call MatSetValue(matA(gups,gups),i+LL4F+LL4W+LL4X+LL4Y-1+(g1-gups)*LL4,k-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
         call MatSetValue(matA(gups,gups),k-1+(g1-gups)*LL4,i+LL4F+LL4W+LL4X+LL4Y-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
      endif
   enddo
enddo

! Fill matrix matAw 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'MUW',MU_PTR)
   CALL C_F_POINTER(MU_PTR,MU,(/ LL4W /))
   WRITE(TEXT10,'(3HWAA,2I3.3)') g1,g1
   CALL LCMGPD(IPSYS,TEXT10,A_PTR)
   CALL C_F_POINTER(A_PTR,A,(/ MU(LL4W)/))
   do i=1,LL4W
      if (i>1) then
         kend=MU(i)-MU(i-1)
         do k=1,kend
            j=i-k+1
            val=A(MU(i)-k+1)
            call MatSetValue(matA(gups,gups),i+LL4F-1+(g1-gups)*LL4,j+LL4F-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
            if (i/=j) call MatSetValue(matA(gups,gups),j+LL4F-1+(g1-gups)*LL4,i+LL4F-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
         enddo
      else
         val=A(1)
         call MatSetValue(matA(gups,gups),i+LL4F-1+(g1-gups)*LL4,i+LL4F-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
      endif
   enddo
endif

! Fill matrix matAx 
IF(DIAG) THEN
   CALL LCMGPD(IPTRK,'MUY',MU_PTR)
   WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
ELSE
   CALL LCMGPD(IPTRK,'MUX',MU_PTR)
   WRITE(TEXT10,'(3HXAA,2I3.3)') g1,g1
ENDIF
CALL C_F_POINTER(MU_PTR,MU,(/ LL4X /))
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4X)/))
do i=1,LL4X
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matA(gups,gups),i+LL4F+LL4W-1+(g1-gups)*LL4,j+LL4F+LL4W-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
         if (i/=j) call MatSetValue(matA(gups,gups),j+LL4F+LL4W-1+(g1-gups)*LL4,i+LL4F+LL4W-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matA(gups,gups),i+LL4F+LL4W-1+(g1-gups)*LL4,i+LL4F+LL4W-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
   endif
enddo

! Fill matrix matAy 
CALL LCMGPD(IPTRK,'MUY',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Y /))
WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Y)/))
do i=1,LL4Y
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matA(gups,gups),i+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,j+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
         if (i/=j) call MatSetValue(matA(gups,gups),j+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,i+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matA(gups,gups),i+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,i+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
   endif
enddo

! Fill matrix matAz 
CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Z /))
WRITE(TEXT10,'(3HZAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Z)/))
do i=1,LL4Z
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         val=A(MU(i)-k+1)
         call MatSetValue(matA(gups,gups),i+LL4F+LL4W+LL4X+LL4Y-1+(g1-gups)*LL4,j+LL4F+LL4W+LL4X+LL4Y-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
         if (i/=j) call MatSetValue(matA(gups,gups),j+LL4F+LL4W+LL4X+LL4Y-1+(g1-gups)*LL4,i+LL4F+LL4W+LL4X+LL4Y-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
      enddo
   else
      val=A(1)
      call MatSetValue(matA(gups,gups),i+LL4F+LL4W+LL4X+LL4Y-1+(g1-gups)*LL4,i+LL4F+LL4W+LL4X+LL4Y-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
   endif
enddo

! Fill vecT
WRITE(TEXT10,'(3HTFA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ LL4F /))
do i=1,LL4F
   val=A(i)
   call MatSetValue(matA(gups,gups),i-1+(g1-gups)*LL4,i-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
enddo





! Fill matrices matCwx, matCwy, matCxy
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'CTRAN',CTRAN_PTR)
   CALL LCMGPD(IPTRK,'KN',KN_PTR)
   CALL LCMGPD(IPTRK,'ZGLOB',IZGLOB_PTR)
   NBLOS=LX*LZ*ISPLH**2
   CALL C_F_POINTER(CTRAN_PTR,CTRAN,(/ (IELEM+1)*IELEM,(IELEM+1)*IELEM /))
   CALL C_F_POINTER(KN_PTR,KN,(/ NBLOS,3+6*(IELEM+2)*IELEM**2 /))
   CALL C_F_POINTER(IZGLOB_PTR,IZGLOB,(/ NBLOS /))
   NELEM=(IELEM+1)*IELEM
   NELEH=NELEM*IELEM
   write(TEXT10,'(1HA,2I3.3)') g1,g1
   CALL LCMGPD(IPSYS,'DIFF'//TEXT10,DIFF_PTR)
   CALL C_F_POINTER(DIFF_PTR,DIFF,(/ NBLOS /))
   ! matrix matCwx 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+2*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matA(gups,gups),row+LL4F-1+(g1-gups)*LL4,col+LL4F+LL4W-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
                        call MatSetValue(matA(gups,gups),col+LL4F+LL4W-1+(g1-gups)*LL4,row+LL4F-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCwy 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_row=KN(NUM,3+i)
               if(KN_row /= 0) then
                  row=ABS(KN_row)
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_col=KN(NUM,3+5*NELEH+j)
                     if (KN_col /= 0) then
                        col=ABS(KN_col)-LL4W-LL4X
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matA(gups,gups),row+LL4F-1+(g1-gups)*LL4,col+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
                        call MatSetValue(matA(gups,gups),col+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,row+LL4F-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCxy 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         DINV=DIFF(k)
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+4*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W-LL4X
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+3*NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)-LL4W
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        val=-SG*DINV*REAL(CTRAN(i0,j0))
                        call MatSetValue(matA(gups,gups),row+LL4F+LL4W-1+(g1-gups)*LL4,col+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
                        call MatSetValue(matA(gups,gups),col+LL4F+LL4W+LL4X-1+(g1-gups)*LL4,row+LL4F+LL4W-1+(g1-gups)*LL4,val,ADD_VALUES,ierr)
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
endif



end subroutine fill_matrixA_raviart_thomas_several_groups 
























! This subroutine creates matrixes used in Raviart-Thomas for SPn, in SLEPc format 
subroutine create_matrixes_raviart_thomas_spn(IPTRK,IPSYS,g1,npre)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 
! npre: maximum number of non-null colums 




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1, npre




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CHEX
INTEGER ISTATE(NSTATE) , ITYPE, LL4, ICHX, LL4F, NLF, &
& ITY, LL4W, LL4X, LL4Y, LL4Z
integer l, i

! SLEPc variables
PetscErrorCode ierr




! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
LL4=ISTATE(11)
ICHX=ISTATE(12)
LL4F=ISTATE(25)
NLF=ISTATE(30)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
ENDIF




! Create matrixes
IF(CHEX) THEN
   do l=1,NLF/2
      ! matCwxl
      call MatCreate(PETSC_COMM_WORLD,matCwxl(g1,l),ierr)
      call MatSetSizes(matCwxl(g1,l),PETSC_DECIDE,PETSC_DECIDE,LL4W,LL4X,ierr)
      call MatSetFromOptions(matCwxl(g1,l),ierr)
      call MatSeqAIJSetPreallocation(matCwxl(g1,l),npre,PETSC_NULL_INTEGER,ierr)
      call MatSetUp(matCwxl(g1,l),ierr)
      ! matCwyl
      call MatCreate(PETSC_COMM_WORLD,matCwyl(g1,l),ierr)
      call MatSetSizes(matCwyl(g1,l),PETSC_DECIDE,PETSC_DECIDE,LL4W,LL4Y,ierr)
      call MatSetFromOptions(matCwyl(g1,l),ierr)
      call MatSeqAIJSetPreallocation(matCwyl(g1,l),npre,PETSC_NULL_INTEGER,ierr)
      call MatSetUp(matCwyl(g1,l),ierr)
      ! matCxyl
      call MatCreate(PETSC_COMM_WORLD,matCxyl(g1,l),ierr)
      call MatSetSizes(matCxyl(g1,l),PETSC_DECIDE,PETSC_DECIDE,LL4X,LL4Y,ierr)
      call MatSetFromOptions(matCxyl(g1,l),ierr)
      call MatSeqAIJSetPreallocation(matCxyl(g1,l),npre,PETSC_NULL_INTEGER,ierr)
      call MatSetUp(matCxyl(g1,l),ierr)
      ! matAwl
      call MatCreate(PETSC_COMM_WORLD,matAwl(g1,l),ierr)
      call MatSetSizes(matAwl(g1,l),PETSC_DECIDE,PETSC_DECIDE,LL4W,LL4W,ierr)
      call MatSetFromOptions(matAwl(g1,l),ierr)
      call MatSeqAIJSetPreallocation(matAwl(g1,l),npre,PETSC_NULL_INTEGER,ierr)
      call MatSetUp(matAwl(g1,l),ierr)
   enddo
   if (g1==1) then
      ! matRw
      call MatCreate(PETSC_COMM_WORLD,matRw,ierr)
      call MatSetSizes(matRw,PETSC_DECIDE,PETSC_DECIDE,LL4W,LL4F,ierr)
      call MatSetFromOptions(matRw,ierr)
      call MatSeqAIJSetPreallocation(matRw,npre,PETSC_NULL_INTEGER,ierr)
      call MatSetUp(matRw,ierr)
   endif
endif
do l=1,NLF/2
   ! matAxl
   call MatCreate(PETSC_COMM_WORLD,matAxl(g1,l),ierr)
   call MatSetSizes(matAxl(g1,l),PETSC_DECIDE,PETSC_DECIDE,LL4X,LL4X,ierr)
   call MatSetFromOptions(matAxl(g1,l),ierr)
   call MatSeqAIJSetPreallocation(matAxl(g1,l),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matAxl(g1,l),ierr)
   ! matAyl
   call MatCreate(PETSC_COMM_WORLD,matAyl(g1,l),ierr)
   call MatSetSizes(matAyl(g1,l),PETSC_DECIDE,PETSC_DECIDE,LL4Y,LL4Y,ierr)
   call MatSetFromOptions(matAyl(g1,l),ierr)
   call MatSeqAIJSetPreallocation(matAyl(g1,l),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matAyl(g1,l),ierr)
   ! matAzl
   call MatCreate(PETSC_COMM_WORLD,matAzl(g1,l),ierr)
   call MatSetSizes(matAzl(g1,l),PETSC_DECIDE,PETSC_DECIDE,LL4Z,LL4Z,ierr)
   call MatSetFromOptions(matAzl(g1,l),ierr)
   call MatSeqAIJSetPreallocation(matAzl(g1,l),npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matAzl(g1,l),ierr)
   ! vecTl
   call VecCreate(PETSC_COMM_WORLD,vecTl(g1,l),ierr)
   call VecSetSizes(vecTl(g1,l),PETSC_DECIDE,LL4F,ierr)
   call VecSetFromOptions(vecTl(g1,l),ierr)
enddo

if (g1==1) then
   ! matRx
   call MatCreate(PETSC_COMM_WORLD,matRx,ierr)
   call MatSetSizes(matRx,PETSC_DECIDE,PETSC_DECIDE,LL4X,LL4F,ierr)
   call MatSetFromOptions(matRx,ierr)
   call MatSeqAIJSetPreallocation(matRx,npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matRx,ierr)
   ! matRy
   call MatCreate(PETSC_COMM_WORLD,matRy,ierr)
   call MatSetSizes(matRy,PETSC_DECIDE,PETSC_DECIDE,LL4Y,LL4F,ierr)
   call MatSetFromOptions(matRy,ierr)
   call MatSeqAIJSetPreallocation(matRy,npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matRy,ierr)
   ! matRz
   call MatCreate(PETSC_COMM_WORLD,matRz,ierr)
   call MatSetSizes(matRz,PETSC_DECIDE,PETSC_DECIDE,LL4Z,LL4F,ierr)
   call MatSetFromOptions(matRz,ierr)
   call MatSeqAIJSetPreallocation(matRz,npre,PETSC_NULL_INTEGER,ierr)
   call MatSetUp(matRz,ierr)
endif

if (g1==1) then
   IF(CHEX) THEN
      do l=1,NLF/2
         do i=1,NLF/2
            if (l/=i) then
               ! vecQwl
               call VecCreate(PETSC_COMM_WORLD,vecQwl(l,i),ierr)
               call VecSetSizes(vecQwl(l,i),PETSC_DECIDE,LL4W,ierr)
               call VecSetFromOptions(vecQwl(l,i),ierr)
            endif
         enddo
      enddo
   endif
   do l=1,NLF/2
      do i=1,NLF/2
         if (l/=i) then
            ! vecQxl
            call VecCreate(PETSC_COMM_WORLD,vecQxl(l,i),ierr)
            call VecSetSizes(vecQxl(l,i),PETSC_DECIDE,LL4X,ierr)
            call VecSetFromOptions(vecQxl(l,i),ierr)
            ! vecQyl
            call VecCreate(PETSC_COMM_WORLD,vecQyl(l,i),ierr)
            call VecSetSizes(vecQyl(l,i),PETSC_DECIDE,LL4Y,ierr)
            call VecSetFromOptions(vecQyl(l,i),ierr)
            ! vecQzl
            call VecCreate(PETSC_COMM_WORLD,vecQzl(l,i),ierr)
            call VecSetSizes(vecQzl(l,i),PETSC_DECIDE,LL4Z,ierr)
            call VecSetFromOptions(vecQzl(l,i),ierr)
         endif
      enddo
   enddo
endif


end subroutine create_matrixes_raviart_thomas_spn
























! This subroutine assembles the matrixes used in Raviart-Thomas for SPn, in SLEPc format 
subroutine assemble_matrixes_raviart_thomas_spn(IPTRK,g1)
! IPTRK: L_TRACK pointer to the tracking information.
! g1: energy group g1 




! Global variables
TYPE(C_PTR) IPTRK 
integer g1 




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CHEX
INTEGER ISTATE(NSTATE) , ITYPE,  NLF
integer l,i

! SLEPc variables
PetscErrorCode ierr




! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
NLF=ISTATE(30)




! Assembly begin
IF(CHEX) THEN
   do l=1,NLF/2
      ! matCwxl
      call MatAssemblyBegin(matCwxl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
      ! matCwyl
      call MatAssemblyBegin(matCwyl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
      ! matCxyl
      call MatAssemblyBegin(matCxyl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
      ! matAwl
      call MatAssemblyBegin(matAwl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
   enddo
   if (g1==1) then
      ! matRw
      call MatAssemblyBegin(matRw,MAT_FINAL_ASSEMBLY,ierr)
   endif
endif
do l=1,NLF/2
   ! matAxl
   call MatAssemblyBegin(matAxl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
   ! matAyl
   call MatAssemblyBegin(matAyl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
   ! matAzl
   call MatAssemblyBegin(matAzl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
   ! vecTl
   call VecAssemblyBegin(vecTl(g1,l),ierr)
enddo

if (g1==1) then
   ! matRx
   call MatAssemblyBegin(matRx,MAT_FINAL_ASSEMBLY,ierr)
   ! matRy
   call MatAssemblyBegin(matRy,MAT_FINAL_ASSEMBLY,ierr)
   ! matRz
   call MatAssemblyBegin(matRz,MAT_FINAL_ASSEMBLY,ierr)
endif

if (g1==1) then
   IF(CHEX) THEN
      do l=1,NLF/2
         do i=1,NLF/2
            if (l/=i) then
               ! vecQwl
               call VecAssemblyBegin(vecQwl(l,i),ierr)
            endif
         enddo
      enddo
   endif
   do l=1,NLF/2
      do i=1,NLF/2
         if (l/=i) then
            ! vecQxl
            call VecAssemblyBegin(vecQxl(l,i),ierr)
            ! vecQyl
            call VecAssemblyBegin(vecQyl(l,i),ierr)
            ! vecQzl
            call VecAssemblyBegin(vecQzl(l,i),ierr)
         endif
      enddo
   enddo
endif



! Assembly end
IF(CHEX) THEN
   do l=1,NLF/2
      ! matCwxl
      call MatAssemblyEnd(matCwxl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
      ! matCwyl
      call MatAssemblyEnd(matCwyl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
      ! matCxyl
      call MatAssemblyEnd(matCxyl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
      ! matAwl
      call MatAssemblyEnd(matAwl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
   enddo
   if (g1==1) then
      ! matRw
      call MatAssemblyEnd(matRw,MAT_FINAL_ASSEMBLY,ierr)
   endif
endif
do l=1,NLF/2
   ! matAxl
   call MatAssemblyEnd(matAxl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
   ! matAyl
   call MatAssemblyEnd(matAyl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
   ! matAzl
   call MatAssemblyEnd(matAzl(g1,l),MAT_FINAL_ASSEMBLY,ierr)
   ! vecTl
   call VecAssemblyEnd(vecTl(g1,l),ierr)
enddo

if (g1==1) then
   ! matRx
   call MatAssemblyEnd(matRx,MAT_FINAL_ASSEMBLY,ierr)
   ! matRy
   call MatAssemblyEnd(matRy,MAT_FINAL_ASSEMBLY,ierr)
   ! matRz
   call MatAssemblyEnd(matRz,MAT_FINAL_ASSEMBLY,ierr)
endif

if (g1==1) then
   IF(CHEX) THEN
      do l=1,NLF/2
         do i=1,NLF/2
            if (l/=i) then
               ! vecQwl
               call VecAssemblyEnd(vecQwl(l,i),ierr)
            endif
         enddo
      enddo
   endif
   do l=1,NLF/2
      do i=1,NLF/2
         if (l/=i) then
            ! vecQxl
            call VecAssemblyEnd(vecQxl(l,i),ierr)
            ! vecQyl
            call VecAssemblyEnd(vecQyl(l,i),ierr)
            ! vecQzl
            call VecAssemblyEnd(vecQzl(l,i),ierr)
         endif
      enddo
   enddo
endif

end subroutine assemble_matrixes_raviart_thomas_spn
























! This subroutine destroys the matrixes used in Raviart-Thomas for SPn, in SLEPc format 
subroutine destroy_matrixes_raviart_thomas_spn(IPTRK,g1)
! IPTRK: L_TRACK pointer to the tracking information.
! g1: energy group g1 




! Global variables
TYPE(C_PTR) IPTRK 
integer g1 




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CHEX
INTEGER ISTATE(NSTATE) , ITYPE,  NLF
integer l

! SLEPc variables
PetscErrorCode ierr




! Recover information from TRIVAC
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
NLF=ISTATE(30)




! Destroy
IF(CHEX) THEN
   do l=1,NLF/2
      ! matCwxl
      call MatDestroy(matCwxl(g1,l),ierr)
      ! matCwyl
      call MatDestroy(matCwyl(g1,l),ierr)
      ! matCxyl
      call MatDestroy(matCxyl(g1,l),ierr)
      ! matAwl
      call MatDestroy(matAwl(g1,l),ierr)
   enddo
   if (g1==1) then
      ! matRw
      call MatDestroy(matRw,ierr)
   endif
endif
do l=1,NLF/2
   ! matAxl
   call MatDestroy(matAxl(g1,l),ierr)
   ! matAyl
   call MatDestroy(matAyl(g1,l),ierr)
   ! matAzl
   call MatDestroy(matAzl(g1,l),ierr)
   ! vecTl
   call VecDestroy(vecTl(g1,l),ierr)
enddo

if (g1==1) then
   ! matRx
   call MatDestroy(matRx,ierr)
   ! matRy
   call MatDestroy(matRy,ierr)
   ! matRz
   call MatDestroy(matRz,ierr)
endif

if (g1==1) then
   IF(CHEX) THEN
      do l=1,NLF/2
         do i=1,NLF/2
            if (l/=i) then
               ! vecQwl
               call VecDestroy(vecQwl(g1,l),ierr)
            endif
         enddo
      enddo
   endif
   do l=1,NLF/2
      do i=1,NLF/2
         if (l/=i) then
            ! vecQxl
            call VecDestroy(vecQxl(g1,l),ierr)
            ! vecQyl
            call VecDestroy(vecQyl(g1,l),ierr)
            ! vecQzl
            call VecDestroy(vecQzl(g1,l),ierr)
         endif
      enddo
   enddo
endif

end subroutine destroy_matrixes_raviart_thomas_spn
























! This subroutine fills matrixes used in Raviart-Thomas for SPn, in SLEPc format 
subroutine fill_matrixes_raviart_thomas_spn(IPTRK,IPSYS,g1)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CYLIND,CHEX,DIAG,LOGY,LOGZ
CHARACTER TEXT10*10, TEXT12*12
INTEGER ISTATE(NSTATE), LL4, LL4W, LL4X, LL4Y, LL4Z, ITY, ITYPE
INTEGER IHEX, IELEM, ICOL, ICHX, ISPLH, ISEG, LX, LY, LZ, NBLOS
INTEGER MAXKN,ITYLCM, NFL, NREG, NVD, MAXQF , NBMIX, NAN
INTEGER, DIMENSION(:), POINTER :: MU
REAL, DIMENSION(:), ALLOCATABLE :: T, QFR_c, KN_c, ZZ
INTEGER, DIMENSION(:), ALLOCATABLE :: MAT
INTEGER, DIMENSION(:), POINTER :: IPW,IPX,IPY,IPZ, IZGLOB
INTEGER, DIMENSION(:,:), POINTER :: IPBW,IPBX,IPBY,IPBZ, KN
REAL, DIMENSION(:), POINTER :: A, FRZ
REAL, DIMENSION(:,:), POINTER :: BW,BX,BY,BZ
REAL, DIMENSION(:,:), ALLOCATABLE :: SIGT,R,V, DIFF, QFR
DOUBLE PRECISION, DIMENSION(:,:), POINTER :: CTRAN
TYPE(C_PTR) IPW_PTR,IPX_PTR,IPY_PTR,IPZ_PTR
TYPE(C_PTR) IPBW_PTR,IPBX_PTR,IPBY_PTR,IPBZ_PTR
TYPE(C_PTR) A_PTR, MU_PTR
TYPE(C_PTR) BW_PTR,BX_PTR,BY_PTR,BZ_PTR
TYPE(C_PTR) KN_PTR,IZGLOB_PTR,CTRAN_PTR, FRZ_PTR
integer i,i0,i1, j0, j, k, g, row, col, NUM, NELEM, NELEH, KN_col, KN_row
integer l,il,pos_mom_l , IND, IBM, NZMAR, KEL, K2, K3, K4, K5, KNW1, KNX1, &
& KNY1, INW1, INX1, INY1, IL2, KNZ2, KNZ1, INZ1, INZ2, NUM1, NUM2, KN1, &
& KN3, KN4, KN5, KN6, IND1, IND2, IND3, IND4, IND5, IND6
real    DINV, SG, DZ, FACT, ZMARS

! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Recover information from TRIVAC
CALL LCMGET(IPSYS,'STATE-VECTOR',ISTATE)
NBMIX=ISTATE(7)
NAN=ISTATE(8)
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CYLIND=(ITYPE.EQ.3).OR.(ITYPE.EQ.6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
IDIM=1
IF((ITYPE.EQ.5).OR.(ITYPE.EQ.6).OR.(ITYPE.EQ.8)) IDIM=2
IF((ITYPE.EQ.7).OR.(ITYPE.EQ.9)) IDIM=3
IHEX=ISTATE(7)
DIAG=(ISTATE(8).EQ.1)
IELEM=ISTATE(9)
ICOL=ISTATE(10)
LL4=ISTATE(11)
ICHX=ISTATE(12)
ISPLH=ISTATE(13)
LX=ISTATE(14)
LY=ISTATE(15)
LZ=ISTATE(16)
ISEG=ISTATE(17)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
   LOGY=LL4Y.GT.0
   LOGZ=LL4Z.GT.0
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
   LOGY=IDIM.GT.1
   LOGZ=IDIM.GT.2
ENDIF
NLF=ISTATE(30)
NREG=ISTATE(1)
NVD=ISTATE(34)




! Fill matrix matRw 
if (g1==1) then
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'IPBBW',IPBW_PTR)
   CALL LCMGPD(IPTRK,'WB',BW_PTR)
   CALL C_F_POINTER(IPBW_PTR,IPBW,(/ 2*IELEM,LL4W /))
   CALL C_F_POINTER(BW_PTR,BW,(/ 2*IELEM,LL4W /))
   do i=1,LL4W
      do j=1,2*IELEM
         k=IPBW(j,i)
         if(k/=0) then
            val= BW(j,i)
            call MatSetValue(matRw,i-1,k-1,val,INSERT_VALUES,ierr)
         endif
      enddo
   enddo
endif
endif

if (g1==1) then
! Fill matrix matRx 
CALL LCMGPD(IPTRK,'IPBBX',IPBX_PTR)
CALL LCMGPD(IPTRK,'XB',BX_PTR)
CALL C_F_POINTER(IPBX_PTR,IPBX,(/ 2*IELEM,LL4X /))
CALL C_F_POINTER(BX_PTR,BX,(/ 2*IELEM,LL4X /))
do i=1,LL4X
   do j=1,2*IELEM
      k=IPBX(j,i)
      if(k/=0) then
         val= BX(j,i)
         call MatSetValue(matRx,i-1,k-1,val,INSERT_VALUES,ierr)
      endif
   enddo
enddo
! Fill matrix matRy 
CALL LCMGPD(IPTRK,'IPBBY',IPBY_PTR)
CALL LCMGPD(IPTRK,'YB',BY_PTR)
CALL C_F_POINTER(IPBY_PTR,IPBY,(/ 2*IELEM,LL4Y /))
CALL C_F_POINTER(BY_PTR,BY,(/ 2*IELEM,LL4Y /))
do i=1,LL4Y
   do j=1,2*IELEM
      k=IPBY(j,i)
      if(k/=0) then
         val= BY(j,i)
         call MatSetValue(matRy,i-1,k-1,val,INSERT_VALUES,ierr)
      endif
   enddo
enddo
! Fill matrix matRz 
CALL LCMGPD(IPTRK,'IPBBZ',IPBZ_PTR)
CALL LCMGPD(IPTRK,'ZB',BZ_PTR)
CALL C_F_POINTER(IPBZ_PTR,IPBZ,(/ 2*IELEM,LL4Z /))
CALL C_F_POINTER(BZ_PTR,BZ,(/ 2*IELEM,LL4Z /))
do i=1,LL4Z
   do j=1,2*IELEM
      k=IPBZ(j,i)
      if(k/=0) then
         val= BZ(j,i)
         call MatSetValue(matRz,i-1,k-1,val,INSERT_VALUES,ierr)
      endif
   enddo
enddo
endif

! Fill matrix matAwl 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'MUW',MU_PTR)
   CALL C_F_POINTER(MU_PTR,MU,(/ LL4W /))
   WRITE(TEXT10,'(3HWAA,2I3.3)') g1,g1
   CALL LCMGPD(IPSYS,TEXT10,A_PTR)
   CALL C_F_POINTER(A_PTR,A,(/ MU(LL4W)*NLF/2 /))
   do i=1,LL4W
      if (i>1) then
         kend=MU(i)-MU(i-1)
         do k=1,kend
            j=i-k+1
            do l=1,NLF/2
               pos_mom_l=MU(LL4W)*(l-1)
               val=A(MU(i)-k+1+pos_mom_l)
               call MatSetValue(matAwl(g1,l),i-1,j-1,val,INSERT_VALUES,ierr)
               call MatSetValue(matAwl(g1,l),j-1,i-1,val,INSERT_VALUES,ierr)
            enddo
         enddo
      else
         do l=1,NLF/2
            pos_mom_l=MU(LL4W)*(l-1)
            val=A(1+pos_mom_l)
            call MatSetValue(matAwl(g1,l),i-1,i-1,val,INSERT_VALUES,ierr)
         enddo
      endif
   enddo
endif

! Fill matrix matAxl 
IF(DIAG) THEN
   CALL LCMGPD(IPTRK,'MUY',MU_PTR)
   WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
ELSE
   CALL LCMGPD(IPTRK,'MUX',MU_PTR)
   WRITE(TEXT10,'(3HXAA,2I3.3)') g1,g1
ENDIF
CALL C_F_POINTER(MU_PTR,MU,(/ LL4X /))
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4X)*NLF/2 /))
do i=1,LL4X
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         do l=1,NLF/2
            pos_mom_l=MU(LL4X)*(l-1)
            val=A(MU(i)-k+1+pos_mom_l)
            call MatSetValue(matAxl(g1,l),i-1,j-1,val,INSERT_VALUES,ierr)
            call MatSetValue(matAxl(g1,l),j-1,i-1,val,INSERT_VALUES,ierr)
         enddo
      enddo
   else
      do l=1,NLF/2
         pos_mom_l=MU(LL4X)*(l-1)
         val=A(1+pos_mom_l)
         call MatSetValue(matAxl(g1,l),i-1,i-1,val,INSERT_VALUES,ierr)
      enddo
   endif
enddo

! Fill matrix matAyl 
CALL LCMGPD(IPTRK,'MUY',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Y /))
WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Y)*NLF/2 /))
do i=1,LL4Y
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         do l=1,NLF/2
            pos_mom_l=MU(LL4Y)*(l-1)
            val=A(MU(i)-k+1+pos_mom_l)
            call MatSetValue(matAyl(g1,l),i-1,j-1,val,INSERT_VALUES,ierr)
            call MatSetValue(matAyl(g1,l),j-1,i-1,val,INSERT_VALUES,ierr)
         enddo
      enddo
   else
      do l=1,NLF/2
         pos_mom_l=MU(LL4Y)*(l-1)
         val=A(1+pos_mom_l)
         call MatSetValue(matAyl(g1,l),i-1,i-1,val,INSERT_VALUES,ierr)
      enddo
   endif
enddo

! Fill matrix matAzl 
CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Z /))
WRITE(TEXT10,'(3HZAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Z)*NLF/2 /))
do i=1,LL4Z
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         do l=1,NLF/2
            pos_mom_l=MU(LL4Z)*(l-1)
            val=A(MU(i)-k+1+pos_mom_l)
            call MatSetValue(matAzl(g1,l),i-1,j-1,val,INSERT_VALUES,ierr)
            call MatSetValue(matAzl(g1,l),j-1,i-1,val,INSERT_VALUES,ierr)
         enddo
      enddo
   else
      do l=1,NLF/2
         pos_mom_l=MU(LL4Z)*(l-1)
         val=A(1+pos_mom_l)
         call MatSetValue(matAzl(g1,l),i-1,i-1,val,INSERT_VALUES,ierr)
      enddo
   endif
enddo

! Fill vecTl
WRITE(TEXT10,'(3HTFA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ LL4F*NLF/2 /))
do i=1,LL4F
   do l=1,NLF/2
      pos_mom_l=LL4F*(l-1)
      val=A(i+pos_mom_l)
      call VecSetValue(vecTl(g1,l),i-1,val,INSERT_VALUES,ierr)
   enddo
enddo





! Get variables for filling matC and vecQ
ALLOCATE(MAT(NREG))
CALL LCMGET(IPTRK,'MATCOD',MAT)






! Fill matrices matCwxl, matCwyl, matCxyl
IF(CHEX) THEN
   ALLOCATE(ZZ(NREG),SIGT(NBMIX,NAN))
   do il=1,NAN
      WRITE(TEXT12,'(4HSCAR,I2.2,2I3.3)') il-1,g1,g1
      CALL LCMGET(IPSYS,TEXT12,SIGT(1,il))
   enddo
   CALL LCMGET(IPTRK,'ZZ',ZZ)
   CALL LCMGPD(IPTRK,'CTRAN',CTRAN_PTR)
   CALL LCMGPD(IPTRK,'KN',KN_PTR)
   CALL LCMGPD(IPTRK,'ZGLOB',IZGLOB_PTR)
   CALL LCMGPD(IPTRK,'FRZ',FRZ_PTR)
   NBLOS=LX*LZ*ISPLH**2
   CALL C_F_POINTER(CTRAN_PTR,CTRAN,(/ (IELEM+1)*IELEM,(IELEM+1)*IELEM /))
   CALL C_F_POINTER(KN_PTR,KN,(/ NBLOS,3+6*(IELEM+2)*IELEM**2 /))
   CALL C_F_POINTER(IZGLOB_PTR,IZGLOB,(/ NBLOS /))
   CALL C_F_POINTER(FRZ_PTR,FRZ,(/ LX*LZ /))
   NELEM=(IELEM+1)*IELEM
   NELEH=NELEM*IELEM
   ALLOCATE(DIFF(NBLOS,NLF/2))
   do i=1,NBLOS
      IND=IZGLOB(i)
      if (IND >0) then
         IBM=MAT(IND)
         DZ=ZZ(IND)*FRZ(IND)
         do l=1,NLF/2
            il=2*l-1
            FACT=REAL(2*il+1)
            DIFF(i,l)=FACT*DZ*SIGT(IBM,MIN(il+1,NAN))
         enddo
      else
         do l=1,NLF/2
            DIFF(i,l)=0.0
         enddo
      endif
   enddo
   ! matrix matCwxl 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+2*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        do l=1,NLF/2
                           DINV=DIFF(k,l)
                           val=-SG*DINV*REAL(CTRAN(i0,j0))
                           call MatSetValue(matCwxl(g1,l),row-1,col-1,val,ADD_VALUES,ierr)
                        enddo
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCwyl 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_row=KN(NUM,3+i)
               if(KN_row /= 0) then
                  row=ABS(KN_row)
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_col=KN(NUM,3+5*NELEH+j)
                     if (KN_col /= 0) then
                        col=ABS(KN_col)-LL4W-LL4X
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        do l=1,NLF/2
                           DINV=DIFF(k,l)
                           val=-SG*DINV*REAL(CTRAN(i0,j0))
                           call MatSetValue(matCwyl(g1,l),row-1,col-1,val,ADD_VALUES,ierr)
                        enddo
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCxyl 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+4*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W-LL4X
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+3*NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)-LL4W
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        do l=1,NLF/2
                           DINV=DIFF(k,l)
                           val=-SG*DINV*REAL(CTRAN(i0,j0))
                           call MatSetValue(matCxyl(g1,l),row-1,col-1,val,ADD_VALUES,ierr)
                        enddo
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   DEALLOCATE(SIGT,DIFF,ZZ)
endif





! Determine NZMAR used for obtaining vecQ
IF(ICOL==3) THEN
   IF(NVD==0) THEN
      NZMAR=NLF+1
   ELSE IF(NVD==1) THEN
      NZMAR=NLF
   ELSE IF(NVD==2) THEN
      NZMAR=65
   ENDIF
ELSE
   NZMAR=65
ENDIF



! Fill vecQ
if (g1==1) then
! vecQ for HEXAGONAL GEOMETRY
IF(CHEX) THEN ! HEXAGONAL GEOMETRY
   allocate(QFR(NBLOS,8))
   CALL LCMGET(IPTRK,'QFR',QFR)
   NELEH=(IELEM+1)*IELEM**2
   NUM=0
   do KEL=1,NBLOS
      IND=IZGLOB(KEL)
      IF(IND /= 0) then
         NUM=NUM+1
         do K5=0,1 ! TWO LOZENGES PER HEXAGON
         do K4=0,IELEM-1
         do K3=0,IELEM-1
         do K2=1,IELEM+1
            KNW1=KN(NUM,3+K5*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
            KNX1=KN(NUM,3+(K5+2)*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
            KNY1=KN(NUM,3+(K5+4)*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
            INW1=ABS(KNW1)
            INX1=ABS(KNX1)-LL4W
            INY1=ABS(KNY1)-LL4W-LL4X
            IF(KNW1 /= 0) THEN
               do IL=1,NLF-1,2
                  do IL2=1,NLF-1,2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+&
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        IF((K2==1).AND.(K5==0)) THEN
                           row=INW1
                           val=0.5*FACT*QFR(NUM,1)*ZMARS
                           call VecSetValue(vecQwl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        ELSE IF((K2==IELEM+1).AND.(K5==1)) THEN
                           row=INW1
                           val=0.5*FACT*QFR(NUM,2)*ZMARS
                           call VecSetValue(vecQwl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        ENDIF
                     endif
                  enddo
               enddo
            ENDIF
            IF(KNX1 /= 0) THEN
               do IL=1,NLF-1,2
                  do IL2=1,NLF-1,2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+&
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        IF((K2==1).AND.(K5==0)) THEN
                           row=INX1
                           val=0.5*FACT*QFR(NUM,3)*ZMARS
                           call VecSetValue(vecQxl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        ELSE IF((K2==IELEM+1).AND.(K5==1)) THEN
                           row=INX1
                           val=0.5*FACT*QFR(NUM,4)*ZMARS
                           call VecSetValue(vecQxl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        ENDIF
                     endif
                  enddo
               enddo
            ENDIF
            IF(KNY1 /= 0) THEN
               do IL=1,NLF-1,2
                  do IL2=1,NLF-1,2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+&
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        IF((K2==1).AND.(K5==0)) THEN
                           row=INY1
                           val=0.5*FACT*QFR(NUM,5)*ZMARS
                           call VecSetValue(vecQyl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        ELSE IF((K2==IELEM+1).AND.(K5==1)) THEN
                           row=INY1
                           val=0.5*FACT*QFR(NUM,6)*ZMARS
                           call VecSetValue(vecQyl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        ENDIF
                     endif
                  enddo
               enddo
            ENDIF
         enddo
         enddo
         enddo
         enddo
         do K5=0,2 ! THREE LOZENGES PER HEXAGON
         do K2=0,IELEM-1
         do K1=0,IELEM-1
            KNZ1=KN(NUM,3+6*NELEH+2*K5*IELEM**2+K2*IELEM+K1+1)
            KNZ2=KN(NUM,3+6*NELEH+(2*K5+1)*IELEM**2+K2*IELEM+K1+1)
            INZ1=ABS(KNZ1)-LL4W-LL4X-LL4Y
            INZ2=ABS(KNZ2)-LL4W-LL4X-LL4Y
            IF((QFR(NUM,7) /= 0.0).AND.(KNZ1 /= 0)) THEN
               ! ZINF SIDE.
               do IL=1,NLF-1,2
                  do IL2=1,NLF-1,2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+&
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        row=INZ1
                        val=0.5*FACT*QFR(NUM,7)*ZMARS
                        call VecSetValue(vecQzl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                     endif
                  enddo
               enddo
            ENDIF
            IF((QFR(NUM,8) /= 0.0).AND.(KNZ2 /= 0)) THEN
               ! ZSUP SIDE.
               do IL=1,NLF-1,2
                  do IL2=1,NLF-1,2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        row=INZ2
                        val=0.5*FACT*QFR(NUM,8)*ZMARS
                        call VecSetValue(vecQzl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                     endif
                  enddo
               enddo
            ENDIF
         enddo
         enddo
         enddo
      endif
   enddo
   deallocate(QFR)
   
! vecQ for CARTESIAN GEOMETRY
else ! CARTESIAN GEOMETRY
   CALL LCMLEN(IPTRK,'KN',MAXKN,ITYLCM)
   CALL LCMLEN(IPTRK,'QFR',MAXQF,ITYLCM)
   ALLOCATE(KN_c(MAXKN),QFR_c(MAXQF))
   CALL LCMGET(IPTRK,'KN',KN_c)
   CALL LCMGET(IPTRK,'QFR',QFR_c)
   NUM1=0
   NUM2=0
   do K=1,NREG
      IBM=MAT(K)
      IF(IBM /= 0) then
         do K3=0,IELEM-1
            do K2=0,IELEM-1
               KN1=KN_c(NUM1+2+K3*IELEM+K2)
               KN2=KN_c(NUM1+2+IELEM**2+K3*IELEM+K2)
               KN3=KN_c(NUM1+2+2*IELEM**2+K3*IELEM+K2)
               KN4=KN_c(NUM1+2+3*IELEM**2+K3*IELEM+K2)
               KN5=KN_c(NUM1+2+4*IELEM**2+K3*IELEM+K2)
               KN6=KN_c(NUM1+2+5*IELEM**2+K3*IELEM+K2)
               IND1=ABS(KN1)-LL4F
               IND2=ABS(KN2)-LL4F
               IND3=ABS(KN3)-LL4F-LL4X
               IND4=ABS(KN4)-LL4F-LL4X
               IND5=ABS(KN5)-LL4F-LL4X-LL4Y
               IND6=ABS(KN6)-LL4F-LL4X-LL4Y
               IF((QFR_c(NUM2+1) /= 0.0).AND.(KN1 /= 0)) THEN
                  ! XINF SIDE.
                  do IL=1,NLF-1,2
                     do IL2=1,NLF-1,2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                               & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND1
                           val=0.5*FACT*QFR_c(NUM2+1)*ZMARS
                           call VecSetValue(vecQxl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+2) /= 0.0).AND.(KN2 /= 0)) THEN
                  ! XSUP SIDE.
                  do IL=1,NLF-1,2
                     do IL2=1,NLF-1,2
                        if (IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                              & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND2
                           val=0.5*FACT*QFR_c(NUM2+2)*ZMARS
                           call VecSetValue(vecQxl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+3) /= 0.0).AND.(KN3 /= 0)) THEN
                  ! YINF SIDE.
                  do IL=1,NLF-1,2
                     do IL2=1,NLF-1,2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                                & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND3
                           val=0.5*FACT*QFR_c(NUM2+3)*ZMARS
                           call VecSetValue(vecQyl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+4) /= 0.0).AND.(KN4 /= 0)) THEN
                  ! YSUP SIDE.
                  do IL=1,NLF-1,2
                     do IL2=1,NLF-1,2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                               & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND4
                           val=0.5*FACT*QFR_c(NUM2+4)*ZMARS
                           call VecSetValue(vecQyl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+5) /= 0.0).AND.(KN5 /= 0)) THEN
                  ! ZINF SIDE.
                  do IL=1,NLF-1,2
                     do IL2=1,NLF-1,2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                               & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND5
                           val=0.5*FACT*QFR_c(NUM2+5)*ZMARS
                           call VecSetValue(vecQzl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+6) /= 0.0).AND.(KN6 /= 0)) THEN
                  ! ZSUP SIDE.
                  do IL=1,NLF-1,2
                     do IL2=1,NLF-1,2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                               & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND6
                           val=0.5*FACT*QFR_c(NUM2+6)*ZMARS
                           call VecSetValue(vecQzl((IL+1)/2,(IL2+1)/2),row-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
            enddo
         enddo
         NUM1=NUM1+1+6*IELEM**2
         NUM2=NUM2+6
      endif
   enddo
   DEALLOCATE(QFR_c,KN_c)

endif
endif




DEALLOCATE(MAT)

end subroutine fill_matrixes_raviart_thomas_spn
























! This subroutine fills matrix Aii used in Raviart-Thomas for SPn, in SLEPc format 
subroutine fill_matrix_Aii_raviart_thomas_spn(IPTRK,IPSYS,g1,gmat)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 
! gmat: index of the energy group of the matrix matA: matA(gmat,gmat)




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1, gmat




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CYLIND,CHEX,DIAG,LOGY,LOGZ
CHARACTER TEXT10*10, TEXT12*12
INTEGER ISTATE(NSTATE), LL4, LL4W, LL4X, LL4Y, LL4Z, ITY, ITYPE
INTEGER IHEX, IELEM, ICOL, ICHX, ISPLH, ISEG, LX, LY, LZ, NBLOS
INTEGER MAXKN,ITYLCM, NFL, NREG, NVD, MAXQF , NBMIX, NAN
INTEGER, DIMENSION(:), POINTER :: MU, KN_c
REAL, DIMENSION(:), ALLOCATABLE :: T, QFR_c, ZZ
INTEGER, DIMENSION(:), ALLOCATABLE :: MAT
INTEGER, DIMENSION(:), POINTER :: IPW,IPX,IPY,IPZ, IZGLOB
INTEGER, DIMENSION(:,:), POINTER :: IPBW,IPBX,IPBY,IPBZ, KN
REAL, DIMENSION(:), POINTER :: A, FRZ
REAL, DIMENSION(:,:), POINTER :: BW,BX,BY,BZ
REAL, DIMENSION(:,:), ALLOCATABLE :: SIGT,R,V, DIFF, QFR
DOUBLE PRECISION, DIMENSION(:,:), POINTER :: CTRAN
TYPE(C_PTR) IPW_PTR,IPX_PTR,IPY_PTR,IPZ_PTR
TYPE(C_PTR) IPBW_PTR,IPBX_PTR,IPBY_PTR,IPBZ_PTR
TYPE(C_PTR) A_PTR, MU_PTR
TYPE(C_PTR) BW_PTR,BX_PTR,BY_PTR,BZ_PTR
TYPE(C_PTR) KN_PTR,IZGLOB_PTR,CTRAN_PTR, FRZ_PTR
integer i,i0,i1, j0, j, k, g, row, col, NUM, NELEM, NELEH, KN_col, KN_row
integer l,il,pos_mom_l , IND, IBM, NZMAR, KEL, K2, K3, K4, K5, KNW1, KNX1, &
& KNY1, INW1, INX1, INY1, IL2, KNZ2, KNZ1, INZ1, INZ2, NUM1, NUM2, KN1, &
& KN3, KN4, KN5, KN6, IND1, IND2, IND3, IND4, IND5, IND6
real    DINV, SG, DZ, FACT, ZMARS
integer l2, lini,lend, row2, col2 

! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Recover information from TRIVAC
CALL LCMGET(IPSYS,'STATE-VECTOR',ISTATE)
NBMIX=ISTATE(7)
NAN=ISTATE(8)
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CYLIND=(ITYPE.EQ.3).OR.(ITYPE.EQ.6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
IDIM=1
IF((ITYPE.EQ.5).OR.(ITYPE.EQ.6).OR.(ITYPE.EQ.8)) IDIM=2
IF((ITYPE.EQ.7).OR.(ITYPE.EQ.9)) IDIM=3
IHEX=ISTATE(7)
DIAG=(ISTATE(8).EQ.1)
IELEM=ISTATE(9)
ICOL=ISTATE(10)
LL4=ISTATE(11)
ICHX=ISTATE(12)
ISPLH=ISTATE(13)
LX=ISTATE(14)
LY=ISTATE(15)
LZ=ISTATE(16)
ISEG=ISTATE(17)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
   LOGY=LL4Y.GT.0
   LOGZ=LL4Z.GT.0
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
   LOGY=IDIM.GT.1
   LOGZ=IDIM.GT.2
ENDIF
NLF=ISTATE(30)
NREG=ISTATE(1)
NVD=ISTATE(34)




! Fill terms corresponding to matrix matRw 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'IPBBW',IPBW_PTR)
   CALL LCMGPD(IPTRK,'WB',BW_PTR)
   CALL C_F_POINTER(IPBW_PTR,IPBW,(/ 2*IELEM,LL4W /))
   CALL C_F_POINTER(BW_PTR,BW,(/ 2*IELEM,LL4W /))
   do i=1,LL4W
      do j=1,2*IELEM
         k=IPBW(j,i)
         if(k/=0) then
            do l=1,NLF/2
               ! Odd equations
               lini=l
               if (l<NLF/2) then
                  lend=l+1
               else
                  lend=l
               endif
               row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+i 
               do l2=lini,lend
                  col=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+k
                  FACT=(2*l-1)+(l2-l)
                  val= FACT*BW(j,i)
                  call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
               enddo
               ! Even equations
               lini=l
               if (l>1) then
                  lend=l-1
               else
                  lend=l
               endif
               row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+k 
               do l2=lini,lend,-1
                  col=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+i
                  FACT=2*(l-1)+(l2-lini+1)
                  val= FACT*BW(j,i)
                  call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
               enddo
            enddo
         endif
      enddo
   enddo
endif

! Fill matrix matRx 
CALL LCMGPD(IPTRK,'IPBBX',IPBX_PTR)
CALL LCMGPD(IPTRK,'XB',BX_PTR)
CALL C_F_POINTER(IPBX_PTR,IPBX,(/ 2*IELEM,LL4X /))
CALL C_F_POINTER(BX_PTR,BX,(/ 2*IELEM,LL4X /))
do i=1,LL4X
   do j=1,2*IELEM
      k=IPBX(j,i)
      if(k/=0) then
         do l=1,NLF/2
            ! Odd equations
            lini=l
            if (l<NLF/2) then
               lend=l+1
            else
               lend=l
            endif
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+i 
            do l2=lini,lend
               col=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+k
               FACT=(2*l-1)+(l2-l)
               val= FACT*BX(j,i)
               call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            enddo
            ! Even equations
            lini=l
            if (l>1) then
               lend=l-1
            else
               lend=l
            endif
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+k 
            do l2=lini,lend,-1
               col=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+i
               FACT=2*(l-1)+(l2-lini+1)
               val= FACT*BX(j,i)
               call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            enddo
         enddo
      endif
   enddo
enddo

! Fill matrix matRy 
CALL LCMGPD(IPTRK,'IPBBY',IPBY_PTR)
CALL LCMGPD(IPTRK,'YB',BY_PTR)
CALL C_F_POINTER(IPBY_PTR,IPBY,(/ 2*IELEM,LL4Y /))
CALL C_F_POINTER(BY_PTR,BY,(/ 2*IELEM,LL4Y /))
do i=1,LL4Y
   do j=1,2*IELEM
      k=IPBY(j,i)
      if(k/=0) then
         do l=1,NLF/2
            ! Odd equations
            lini=l
            if (l<NLF/2) then
               lend=l+1
            else
               lend=l
            endif
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+i 
            do l2=lini,lend
               col=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+k
               FACT=(2*l-1)+(l2-l)
               val= FACT*BY(j,i)
               call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            enddo
            ! Even equations
            lini=l
            if (l>1) then
               lend=l-1
            else
               lend=l
            endif
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+k 
            do l2=lini,lend,-1
               col=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+i
               FACT=2*(l-1)+(l2-lini+1)
               val= FACT*BX(j,i)
               val= FACT*BY(j,i)
               call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            enddo
         enddo
      endif
   enddo
enddo

! Fill matrix matRz 
CALL LCMGPD(IPTRK,'IPBBZ',IPBZ_PTR)
CALL LCMGPD(IPTRK,'ZB',BZ_PTR)
CALL C_F_POINTER(IPBZ_PTR,IPBZ,(/ 2*IELEM,LL4Z /))
CALL C_F_POINTER(BZ_PTR,BZ,(/ 2*IELEM,LL4Z /))
do i=1,LL4Z
   do j=1,2*IELEM
      k=IPBZ(j,i)
      if(k/=0) then
         do l=1,NLF/2
            ! Odd equations
            lini=l
            if (l<NLF/2) then
               lend=l+1
            else
               lend=l
            endif
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+LL4Y+i 
            do l2=lini,lend
               col=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+k
               FACT=(2*l-1)+(l2-l)
               val= FACT*BZ(j,i)
               call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            enddo
            ! Even equations
            lini=l
            if (l>1) then
               lend=l-1
            else
               lend=l
            endif
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+k 
            do l2=lini,lend,-1
               col=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+LL4Y+i
               FACT=2*(l-1)+(l2-lini+1)
               val= FACT*BX(j,i)
               val= FACT*BZ(j,i)
               call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            enddo
         enddo
      endif
   enddo
enddo

! Fill matrix matAwl 
IF(CHEX) THEN
   CALL LCMGPD(IPTRK,'MUW',MU_PTR)
   CALL C_F_POINTER(MU_PTR,MU,(/ LL4W /))
   WRITE(TEXT10,'(3HWAA,2I3.3)') g1,g1
   CALL LCMGPD(IPSYS,TEXT10,A_PTR)
   CALL C_F_POINTER(A_PTR,A,(/ MU(LL4W)*NLF/2 /))
   do i=1,LL4W
      if (i>1) then
         kend=MU(i)-MU(i-1)
         do k=1,kend
            j=i-k+1
            do l=1,NLF/2
               pos_mom_l=MU(LL4W)*(l-1)
               val=A(MU(i)-k+1+pos_mom_l)
               row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+i 
               col=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+j 
               call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
               if (row/=col) call MatSetValue(matA(gmat,gmat),col-1,row-1,val,ADD_VALUES,ierr)
            enddo
         enddo
      else
         do l=1,NLF/2
            pos_mom_l=MU(LL4W)*(l-1)
            val=A(1+pos_mom_l)
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+i 
            call MatSetValue(matA(gmat,gmat),row-1,row-1,val,ADD_VALUES,ierr)
         enddo
      endif
   enddo
endif

! Fill matrix matAxl 
IF(DIAG) THEN
   CALL LCMGPD(IPTRK,'MUY',MU_PTR)
   WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
ELSE
   CALL LCMGPD(IPTRK,'MUX',MU_PTR)
   WRITE(TEXT10,'(3HXAA,2I3.3)') g1,g1
ENDIF
CALL C_F_POINTER(MU_PTR,MU,(/ LL4X /))
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4X)*NLF/2 /))
do i=1,LL4X
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         do l=1,NLF/2
            pos_mom_l=MU(LL4X)*(l-1)
            val=A(MU(i)-k+1+pos_mom_l)
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+i 
            col=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+j 
            call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            if (row/=col) call MatSetValue(matA(gmat,gmat),col-1,row-1,val,ADD_VALUES,ierr)
         enddo
      enddo
   else
      do l=1,NLF/2
         pos_mom_l=MU(LL4X)*(l-1)
         val=A(1+pos_mom_l)
         row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+i 
         call MatSetValue(matA(gmat,gmat),row-1,row-1,val,ADD_VALUES,ierr)
      enddo
   endif
enddo

! Fill matrix matAyl 
CALL LCMGPD(IPTRK,'MUY',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Y /))
WRITE(TEXT10,'(3HYAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Y)*NLF/2 /))
do i=1,LL4Y
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         do l=1,NLF/2
            pos_mom_l=MU(LL4Y)*(l-1)
            val=A(MU(i)-k+1+pos_mom_l)
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+i 
            col=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+j 
            call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            if (row/=col) call MatSetValue(matA(gmat,gmat),col-1,row-1,val,ADD_VALUES,ierr)
         enddo
      enddo
   else
      do l=1,NLF/2
         pos_mom_l=MU(LL4Y)*(l-1)
         val=A(1+pos_mom_l)
         row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+i 
         call MatSetValue(matA(gmat,gmat),row-1,row-1,val,ADD_VALUES,ierr)
      enddo
   endif
enddo

! Fill matrix matAzl 
CALL LCMGPD(IPTRK,'MUZ',MU_PTR)
CALL C_F_POINTER(MU_PTR,MU,(/ LL4Z /))
WRITE(TEXT10,'(3HZAA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ MU(LL4Z)*NLF/2 /))
do i=1,LL4Z
   if (i>1) then
      kend=MU(i)-MU(i-1)
      do k=1,kend
         j=i-k+1
         do l=1,NLF/2
            pos_mom_l=MU(LL4Z)*(l-1)
            val=A(MU(i)-k+1+pos_mom_l)
            row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+LL4Y+i 
            col=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+LL4Y+j 
            call MatSetValue(matA(gmat,gmat),row-1,col-1,val,ADD_VALUES,ierr)
            if (row/=col) call MatSetValue(matA(gmat,gmat),col-1,row-1,val,ADD_VALUES,ierr)
         enddo
      enddo
   else
      do l=1,NLF/2
         pos_mom_l=MU(LL4Z)*(l-1)
         val=A(1+pos_mom_l)
         row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+LL4Y+i 
         call MatSetValue(matA(gmat,gmat),row-1,row-1,val,ADD_VALUES,ierr)
      enddo
   endif
enddo

! Fill vecTl
WRITE(TEXT10,'(3HTFA,2I3.3)') g1,g1
CALL LCMGPD(IPSYS,TEXT10,A_PTR)
CALL C_F_POINTER(A_PTR,A,(/ LL4F*NLF/2 /))
do i=1,LL4F
   do l=1,NLF/2
      pos_mom_l=LL4F*(l-1)
      val=A(i+pos_mom_l)
      row=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+i 
      call MatSetValue(matA(gmat,gmat),row-1,row-1,val,ADD_VALUES,ierr)
   enddo
enddo





! Get variables for filling matC and vecQ
ALLOCATE(MAT(NREG))
CALL LCMGET(IPTRK,'MATCOD',MAT)






! Fill matrices matCwxl, matCwyl, matCxyl
IF(CHEX) THEN
   ALLOCATE(ZZ(NREG),SIGT(NBMIX,NAN))
   do il=1,NAN
      WRITE(TEXT12,'(4HSCAR,I2.2,2I3.3)') il-1,g1,g1
      CALL LCMGET(IPSYS,TEXT12,SIGT(1,il))
   enddo
   CALL LCMGET(IPTRK,'ZZ',ZZ)
   CALL LCMGPD(IPTRK,'CTRAN',CTRAN_PTR)
   CALL LCMGPD(IPTRK,'KN',KN_PTR)
   CALL LCMGPD(IPTRK,'ZGLOB',IZGLOB_PTR)
   CALL LCMGPD(IPTRK,'FRZ',FRZ_PTR)
   NBLOS=LX*LZ*ISPLH**2
   CALL C_F_POINTER(CTRAN_PTR,CTRAN,(/ (IELEM+1)*IELEM,(IELEM+1)*IELEM /))
   CALL C_F_POINTER(KN_PTR,KN,(/ NBLOS,3+6*(IELEM+2)*IELEM**2 /))
   CALL C_F_POINTER(IZGLOB_PTR,IZGLOB,(/ NBLOS /))
   CALL C_F_POINTER(FRZ_PTR,FRZ,(/ LX*LZ /))
   NELEM=(IELEM+1)*IELEM
   NELEH=NELEM*IELEM
   ALLOCATE(DIFF(NBLOS,NLF/2))
   do i=1,NBLOS
      IND=IZGLOB(i)
      if (IND >0) then
         IBM=MAT(IND)
         DZ=ZZ(IND)*FRZ(IND)
         do l=1,NLF/2
            il=2*l-1
            FACT=REAL(2*il+1)
            DIFF(i,l)=FACT*DZ*SIGT(IBM,MIN(il+1,NAN))
         enddo
      else
         do l=1,NLF/2
            DIFF(i,l)=0.0
         enddo
      endif
   enddo
   ! matrix matCwxl 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+2*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        do l=1,NLF/2
                           DINV=DIFF(k,l)
                           val=-SG*DINV*REAL(CTRAN(i0,j0))
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+col 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                           call MatSetValue(matA(gmat,gmat),col2-1,row2-1,val,ADD_VALUES,ierr)
                        enddo
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCwyl 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_row=KN(NUM,3+i)
               if(KN_row /= 0) then
                  row=ABS(KN_row)
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_col=KN(NUM,3+5*NELEH+j)
                     if (KN_col /= 0) then
                        col=ABS(KN_col)-LL4W-LL4X
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        do l=1,NLF/2
                           DINV=DIFF(k,l)
                           val=-SG*DINV*REAL(CTRAN(i0,j0))
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+col 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                           call MatSetValue(matA(gmat,gmat),col2-1,row2-1,val,ADD_VALUES,ierr)
                        enddo
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   ! matrix matCxyl 
   NUM=0
   do k=1,NBLOS
      if (IZGLOB(k) /= 0) then
         NUM=NUM+1
         do i1=0,IELEM-1
            do i0=1,NELEM
               i=i1*NELEM+i0
               KN_col=KN(NUM,3+4*NELEH+i)
               if(KN_col /= 0) then
                  col=ABS(KN_col)-LL4W-LL4X
                  do j0=1,NELEM
                     j=i1*NELEM+j0
                     KN_row=KN(NUM,3+3*NELEH+j)
                     if (KN_row /= 0) then
                        row=ABS(KN_row)-LL4W
                        SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                        do l=1,NLF/2
                           DINV=DIFF(k,l)
                           val=-SG*DINV*REAL(CTRAN(i0,j0))
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+col 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                           call MatSetValue(matA(gmat,gmat),col2-1,row2-1,val,ADD_VALUES,ierr)
                        enddo
                     endif
                  enddo 
               endif
            enddo
         enddo 
      endif
   enddo 
   DEALLOCATE(SIGT,DIFF,ZZ)
endif





! Determine NZMAR used for obtaining vecQ
IF(ICOL==3) THEN
   IF(NVD==0) THEN
      NZMAR=NLF+1
   ELSE IF(NVD==1) THEN
      NZMAR=NLF
   ELSE IF(NVD==2) THEN
      NZMAR=65
   ENDIF
ELSE
   NZMAR=65
ENDIF



! Fill vecQ
! vecQ for HEXAGONAL GEOMETRY
IF(CHEX) THEN ! HEXAGONAL GEOMETRY
   allocate(QFR(NBLOS,8))
   CALL LCMGET(IPTRK,'QFR',QFR)
   NELEH=(IELEM+1)*IELEM**2
   NUM=0
   do KEL=1,NBLOS
      IND=IZGLOB(KEL)
      IF(IND /= 0) then
         NUM=NUM+1
         do K5=0,1 ! TWO LOZENGES PER HEXAGON
         do K4=0,IELEM-1
         do K3=0,IELEM-1
         do K2=1,IELEM+1
            KNW1=KN(NUM,3+K5*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
            KNX1=KN(NUM,3+(K5+2)*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
            KNY1=KN(NUM,3+(K5+4)*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
            INW1=ABS(KNW1)
            INX1=ABS(KNX1)-LL4W
            INY1=ABS(KNY1)-LL4W-LL4X
            IF(KNW1 /= 0) THEN
               do IL=1,NLF-1,2
                  l=(IL+1)/2
                  do IL2=1,NLF-1,2
                     l2=(IL2+1)/2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+&
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        IF((K2==1).AND.(K5==0)) THEN
                           row=INW1
                           val=0.5*FACT*QFR(NUM,1)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ELSE IF((K2==IELEM+1).AND.(K5==1)) THEN
                           row=INW1
                           val=0.5*FACT*QFR(NUM,2)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ENDIF
                     endif
                  enddo
               enddo
            ENDIF
            IF(KNX1 /= 0) THEN
               do IL=1,NLF-1,2
                  l=(IL+1)/2
                  do IL2=1,NLF-1,2
                     l2=(IL2+1)/2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+&
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        IF((K2==1).AND.(K5==0)) THEN
                           row=INX1
                           val=0.5*FACT*QFR(NUM,3)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ELSE IF((K2==IELEM+1).AND.(K5==1)) THEN
                           row=INX1
                           val=0.5*FACT*QFR(NUM,4)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ENDIF
                     endif
                  enddo
               enddo
            ENDIF
            IF(KNY1 /= 0) THEN
               do IL=1,NLF-1,2
                  l=(IL+1)/2
                  do IL2=1,NLF-1,2
                     l2=(IL2+1)/2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+&
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        IF((K2==1).AND.(K5==0)) THEN
                           row=INY1
                           val=0.5*FACT*QFR(NUM,5)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ELSE IF((K2==IELEM+1).AND.(K5==1)) THEN
                           row=INY1
                           val=0.5*FACT*QFR(NUM,6)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ENDIF
                     endif
                  enddo
               enddo
            ENDIF
         enddo
         enddo
         enddo
         enddo
         do K5=0,2 ! THREE LOZENGES PER HEXAGON
         do K2=0,IELEM-1
         do K1=0,IELEM-1
            KNZ1=KN(NUM,3+6*NELEH+2*K5*IELEM**2+K2*IELEM+K1+1)
            KNZ2=KN(NUM,3+6*NELEH+(2*K5+1)*IELEM**2+K2*IELEM+K1+1)
            INZ1=ABS(KNZ1)-LL4W-LL4X-LL4Y
            INZ2=ABS(KNZ2)-LL4W-LL4X-LL4Y
            IF((QFR(NUM,7) /= 0.0).AND.(KNZ1 /= 0)) THEN
               ! ZINF SIDE.
               do IL=1,NLF-1,2
                  l=(IL+1)/2
                  do IL2=1,NLF-1,2
                     l2=(IL2+1)/2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+&
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        row=INZ1
                        val=0.5*FACT*QFR(NUM,7)*ZMARS
                        row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+LL4Y+row 
                        col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+LL4Y+row 
                        call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                     endif
                  enddo
               enddo
            ENDIF
            IF((QFR(NUM,8) /= 0.0).AND.(KNZ2 /= 0)) THEN
               ! ZSUP SIDE.
               do IL=1,NLF-1,2
                  l=(IL+1)/2
                  do IL2=1,NLF-1,2
                     l2=(IL2+1)/2
                     if(IL2 /= IL) then
                        ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                            & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                        FACT=REAL(2*IL+1)
                        row=INZ2
                        val=0.5*FACT*QFR(NUM,8)*ZMARS
                        row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+LL4Y+row 
                        col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+LL4Y+row 
                        call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                     endif
                  enddo
               enddo
            ENDIF
         enddo
         enddo
         enddo
      endif
   enddo
   deallocate(QFR)
   
! vecQ for CARTESIAN GEOMETRY
else ! CARTESIAN GEOMETRY
   CALL LCMLEN(IPTRK,'KN',MAXKN,ITYLCM)
   CALL LCMLEN(IPTRK,'QFR',MAXQF,ITYLCM)
   ALLOCATE(KN_c(MAXKN),QFR_c(MAXQF))
   CALL LCMGET(IPTRK,'KN',KN_c)
   CALL LCMGET(IPTRK,'QFR',QFR_c)
   NUM1=0
   NUM2=0
   do K=1,NREG
      IBM=MAT(K)
      IF(IBM /= 0) then
         do K3=0,IELEM-1
            do K2=0,IELEM-1
               KN1=KN_c(NUM1+2+K3*IELEM+K2)
               KN2=KN_c(NUM1+2+IELEM**2+K3*IELEM+K2)
               KN3=KN_c(NUM1+2+2*IELEM**2+K3*IELEM+K2)
               KN4=KN_c(NUM1+2+3*IELEM**2+K3*IELEM+K2)
               KN5=KN_c(NUM1+2+4*IELEM**2+K3*IELEM+K2)
               KN6=KN_c(NUM1+2+5*IELEM**2+K3*IELEM+K2)
               IND1=ABS(KN1)-LL4F
               IND2=ABS(KN2)-LL4F
               IND3=ABS(KN3)-LL4F-LL4X
               IND4=ABS(KN4)-LL4F-LL4X
               IND5=ABS(KN5)-LL4F-LL4X-LL4Y
               IND6=ABS(KN6)-LL4F-LL4X-LL4Y
               IF((QFR_c(NUM2+1) /= 0.0).AND.(KN1 /= 0)) THEN
                  ! XINF SIDE.
                  do IL=1,NLF-1,2
                     l=(IL+1)/2
                     do IL2=1,NLF-1,2
                        l2=(IL2+1)/2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                               & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND1
                           val=0.5*FACT*QFR_c(NUM2+1)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+2) /= 0.0).AND.(KN2 /= 0)) THEN
                  ! XSUP SIDE.
                  do IL=1,NLF-1,2
                     l=(IL+1)/2
                     do IL2=1,NLF-1,2
                        l2=(IL2+1)/2
                        if (IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                              & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND2
                           val=0.5*FACT*QFR_c(NUM2+2)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+3) /= 0.0).AND.(KN3 /= 0)) THEN
                  ! YINF SIDE.
                  do IL=1,NLF-1,2
                     l=(IL+1)/2
                     do IL2=1,NLF-1,2
                        l2=(IL2+1)/2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                                & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND3
                           val=0.5*FACT*QFR_c(NUM2+3)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+4) /= 0.0).AND.(KN4 /= 0)) THEN
                  ! YSUP SIDE.
                  do IL=1,NLF-1,2
                     l=(IL+1)/2
                     do IL2=1,NLF-1,2
                        l2=(IL2+1)/2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                               & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND4
                           val=0.5*FACT*QFR_c(NUM2+4)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+5) /= 0.0).AND.(KN5 /= 0)) THEN
                  ! ZINF SIDE.
                  do IL=1,NLF-1,2
                     l=(IL+1)/2
                     do IL2=1,NLF-1,2
                        l2=(IL2+1)/2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                               & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND5
                           val=0.5*FACT*QFR_c(NUM2+5)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+LL4Y+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+LL4Y+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
               IF((QFR_c(NUM2+6) /= 0.0).AND.(KN6 /= 0)) THEN
                  ! ZSUP SIDE.
                  do IL=1,NLF-1,2
                     l=(IL+1)/2
                     do IL2=1,NLF-1,2
                        l2=(IL2+1)/2
                        if(IL2 /= IL) then
                           ZMARS=REAL(IL2)*PNMARS(NZMAR,IL2-1,IL)/REAL(2*IL2-1)+ &
                               & REAL(IL2+1)*PNMARS(NZMAR,IL2+1,IL)/REAL(2*IL2+3)
                           FACT=REAL(2*IL+1)
                           row=IND6
                           val=0.5*FACT*QFR_c(NUM2+6)*ZMARS
                           row2=(gmat-g1)*LL4*NLF/2+LL4*(l-1)+LL4F+LL4W+LL4X+LL4Y+row 
                           col2=(gmat-g1)*LL4*NLF/2+LL4*(l2-1)+LL4F+LL4W+LL4X+LL4Y+row 
                           call MatSetValue(matA(gmat,gmat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        endif
                     enddo
                  enddo
               ENDIF
            enddo
         enddo
         NUM1=NUM1+1+6*IELEM**2
         NUM2=NUM2+6
      endif
   enddo
   DEALLOCATE(QFR_c,KN_c)

endif




DEALLOCATE(MAT)

end subroutine fill_matrix_Aii_raviart_thomas_spn
























! This subroutine fills matrix Aij (cartesian geometry) used in Raviart-Thomas for SPn, in SLEPc format 
! Matrixes Aij are those containing scattering terms
subroutine fill_matrix_Aij_cart_raviart_thomas_spn(IPTRK,IPSYS,g1,g2,g1mat,g2mat)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 
! g2: energy group g2 
! g1mat: first index of the energy group of the matrix matA: matA(g1mat,g2mat)
! g2mat: secondindex of the energy group of the matrix matA: matA(g1mat,g2mat)




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1, g2, g1mat, g2mat




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CYLIND,CHEX,DIAG 
CHARACTER TEXT12*12
INTEGER ISTATE(NSTATE), LL4,LL4F, LL4W, LL4X, LL4Y, LL4Z, ITY, ITYPE
INTEGER IHEX, IELEM, ICOL, ICHX, ISPLH, ISEG, LX, LY, LZ, IDIM 
INTEGER MAXKN,ITYLCM, NFL, NREG, NVD, NBMIX, NAN, LENGT
INTEGER, DIMENSION(:), ALLOCATABLE :: KN
REAL, DIMENSION(:), ALLOCATABLE :: ZZ, VOL
INTEGER, DIMENSION(:), ALLOCATABLE :: MAT
REAL, DIMENSION(:,:), ALLOCATABLE :: SIGT,R 
integer K,row,col,row2,col2,LC,il,IND1,IND2,IBM,IC,JC,IIC,JJC
integer  K1,K2, K3, NUM1, NUM2, KN1, KN2 
real     FACT, VOL0, GARS, S1, S2 

! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Recover information from TRIVAC
CALL LCMGET(IPSYS,'STATE-VECTOR',ISTATE)
NBMIX=ISTATE(7)
NAN=ISTATE(8)
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CYLIND=(ITYPE.EQ.3).OR.(ITYPE.EQ.6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
IDIM=1
IF((ITYPE.EQ.5).OR.(ITYPE.EQ.6).OR.(ITYPE.EQ.8)) IDIM=2
IF((ITYPE.EQ.7).OR.(ITYPE.EQ.9)) IDIM=3
IHEX=ISTATE(7)
DIAG=(ISTATE(8).EQ.1)
IELEM=ISTATE(9)
ICOL=ISTATE(10)
LL4=ISTATE(11)
ICHX=ISTATE(12)
ISPLH=ISTATE(13)
LX=ISTATE(14)
LY=ISTATE(15)
LZ=ISTATE(16)
ISEG=ISTATE(17)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
ENDIF
NLF=ISTATE(30)
NREG=ISTATE(1)
NVD=ISTATE(34)


! Get variables
ALLOCATE(MAT(NREG),VOL(NREG))
CALL LCMGET(IPTRK,'MATCOD',MAT)
CALL LCMGET(IPTRK,'VOLUME',VOL)
CALL LCMLEN(IPTRK,'T',LC,ITYLCM)
CALL LCMLEN(IPTRK,'KN',MAXKN,ITYLCM)
ALLOCATE(KN(MAXKN))
CALL LCMGET(IPTRK,'KN',KN)
ALLOCATE(ZZ(NREG),SIGT(NBMIX,NAN))
do il=1,NAN
   WRITE(TEXT12,'(4HSCAR,I2.2,2I3.3)') il-1,g1,g2
   CALL LCMLEN(IPSYS,TEXT12,LENGT,ITYLCM)
   IF(LENGT.EQ.0) THEN
      CALL XDRSET(SIGT(1,IL),NBMIX,0.0)
   else
      CALL LCMGET(IPSYS,TEXT12,SIGT(1,il))
   endif
enddo
CALL LCMSIX(IPTRK,'BIVCOL',1)
CALL LCMLEN(IPTRK,'T',LC,ITYLCM)
ALLOCATE(R(LC,LC))
CALL LCMGET(IPTRK,'R',R)
CALL LCMSIX(IPTRK,' ',2)






! Fill terms
do IL=0,NLF-1
   IF(IL < NAN) then
      FACT=REAL(2*IL+1)
      NUM1=0
      NUM2=0
      do K=1,NREG
         IBM=MAT(K)
         IF(IBM /= 0) then
            VOL0=VOL(K)
            GARS=SIGT(IBM,MIN(IL+1,NAN))
            IF(MOD(IL,2) == 0) THEN
               ! EVEN PARITY EQUATION
               do K3=0,IELEM-1
               do K2=0,IELEM-1
               do K1=0,IELEM-1
                  row=(IL/2)*LL4+KN(NUM1+1)+(K3*IELEM+K2)*IELEM+K1
                  col=row
                  val=FACT*VOL0*GARS
                  row2=row+(g1mat-g1)*LL4*NLF/2
                  col2=col+(g2mat-g2)*LL4*NLF/2
                  val=-1.0*val
                  call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
               enddo
               enddo
               enddo
            ELSE IF(MOD(IL,2) == 1) THEN
               ! ODD PARITY EQUATION.
               do K3=0,IELEM-1
               do K2=0,IELEM-1
                  do IC=1,2
                     IF(IC == 1) IIC=1
                     IF(IC == 2) IIC=IELEM+1
                     KN1=KN(NUM1+2+(IC-1)*IELEM**2+K3*IELEM+K2)
                     IND1=(IL/2)*LL4+ABS(KN1)
                     S1=REAL(SIGN(1,KN1))
                     do JC=1,2
                        KN2=KN(NUM1+2+(JC-1)*IELEM**2+K3*IELEM+K2)
                        IF((KN1 /= 0).AND.(KN2 /= 0)) THEN
                           JJC=1
                           IF(JC == 2) JJC=IELEM+1
                           IND2=(IL/2)*LL4+ABS(KN2)
                           S2=REAL(SIGN(1,KN2))
                           row=IND1
                           col=IND2
                           val=-S1*S2*FACT*R(IIC,JJC)*VOL0*GARS
                           row2=row+(g1mat-g1)*LL4*NLF/2
                           col2=col+(g2mat-g2)*LL4*NLF/2
                           val=-1.0*val
                           call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ENDIF
                     enddo
                  enddo
                  do IC=3,4
                     IF(IC == 3) IIC=1
                     IF(IC == 4) IIC=IELEM+1
                     KN1=KN(NUM1+2+(IC-1)*IELEM**2+K3*IELEM+K2)
                     IND1=(IL/2)*LL4+ABS(KN1)
                     S1=REAL(SIGN(1,KN1))
                     do JC=3,4
                        KN2=KN(NUM1+2+(JC-1)*IELEM**2+K3*IELEM+K2)
                        IF((KN1 /= 0).AND.(KN2 /= 0)) THEN
                           JJC=1
                           IF(JC == 4) JJC=IELEM+1
                           IND2=(IL/2)*LL4+ABS(KN2)
                           S2=REAL(SIGN(1,KN2))
                           row=IND1
                           col=IND2
                           val=-S1*S2*FACT*R(IIC,JJC)*VOL0*GARS
                           row2=row+(g1mat-g1)*LL4*NLF/2
                           col2=col+(g2mat-g2)*LL4*NLF/2
                           val=-1.0*val
                           call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ENDIF
                     enddo
                  enddo
                  do IC=5,6
                     IF(IC == 5) IIC=1
                     IF(IC == 6) IIC=IELEM+1
                     KN1=KN(NUM1+2+(IC-1)*IELEM**2+K3*IELEM+K2)
                     IND1=(IL/2)*LL4+ABS(KN1)
                     S1=REAL(SIGN(1,KN1))
                     do JC=5,6
                        KN2=KN(NUM1+2+(JC-1)*IELEM**2+K3*IELEM+K2)
                        IF((KN1 /= 0).AND.(KN2 /= 0)) THEN
                           JJC=1
                           IF(JC == 6) JJC=IELEM+1
                           IND2=(IL/2)*LL4+ABS(KN2)
                           S2=REAL(SIGN(1,KN2))
                           row=IND1
                           col=IND2
                           val=-S1*S2*FACT*R(IIC,JJC)*VOL0*GARS
                           row2=row+(g1mat-g1)*LL4*NLF/2
                           col2=col+(g2mat-g2)*LL4*NLF/2
                           val=-1.0*val
                           call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                        ENDIF
                     enddo
                  enddo
               enddo
               enddo
            ENDIF
            NUM1=NUM1+1+6*IELEM**2
            NUM2=NUM2+6
         endif
      enddo
   endif
enddo



end subroutine fill_matrix_Aij_cart_raviart_thomas_spn
























! This subroutine fills matrix Aij (hexagonal geometry) used in Raviart-Thomas for SPn, in SLEPc format 
! Matrixes Aij are those containing scattering terms
subroutine fill_matrix_Aij_hexa_raviart_thomas_spn(IPTRK,IPSYS,g1,g2,g1mat,g2mat)
! IPTRK: L_TRACK pointer to the tracking information.
! IPSYS: L_SYSTEM pointer to system matrices.
! g1: energy group g1 
! g2: energy group g2 
! g1mat: first index of the energy group of the matrix matA: matA(g1mat,g2mat)
! g2mat: secondindex of the energy group of the matrix matA: matA(g1mat,g2mat)




! Global variables
TYPE(C_PTR) IPTRK,IPSYS
integer g1, g2, g1mat, g2mat




! Local variables
PARAMETER (NSTATE=40)
LOGICAL CYLIND,CHEX,DIAG 
CHARACTER TEXT12*12
INTEGER ISTATE(NSTATE), LL4,LL4F, LL4W, LL4X, LL4Y, LL4Z, ITY, ITYPE
INTEGER IHEX, IELEM, ICOL, ICHX, ISPLH, ISEG, LX, LY, LZ, IDIM 
INTEGER MAXKN,ITYLCM, NFL, NREG, NVD, NBMIX, NAN, LENGT, NBLOS
DOUBLE PRECISION, DIMENSION(:,:), POINTER :: CTRAN
REAL, DIMENSION(:), POINTER :: FRZ
INTEGER, DIMENSION(:), POINTER :: IZGLOB
INTEGER, DIMENSION(:,:), POINTER :: KN
TYPE(C_PTR) KN_PTR,IZGLOB_PTR,CTRAN_PTR, FRZ_PTR
REAL, DIMENSION(:), ALLOCATABLE :: ZZ, DIFF 
INTEGER, DIMENSION(:), ALLOCATABLE :: MAT
REAL, DIMENSION(:,:), ALLOCATABLE :: SIGT,R 
integer K,row,col,row2,col2,LC,il,IND1,IND2,IBM,IC,JC,IIC,JJC
integer  K1,K2, K3, NUM1, NUM2, KN1, KN2 
real     FACT, VOL0, GARS, S1, S2, SIDE, TTTT,UUUU,VAR1 

! SLEPc variables
PetscErrorCode ierr
PetscScalar val



! Recover information from TRIVAC
CALL LCMGET(IPSYS,'STATE-VECTOR',ISTATE)
NBMIX=ISTATE(7)
NAN=ISTATE(8)
CALL LCMGET(IPTRK,'STATE-VECTOR',ISTATE)
ITYPE=ISTATE(6)
CYLIND=(ITYPE.EQ.3).OR.(ITYPE.EQ.6)
CHEX=(ITYPE.EQ.8).OR.(ITYPE.EQ.9)
IDIM=1
IF((ITYPE.EQ.5).OR.(ITYPE.EQ.6).OR.(ITYPE.EQ.8)) IDIM=2
IF((ITYPE.EQ.7).OR.(ITYPE.EQ.9)) IDIM=3
IHEX=ISTATE(7)
DIAG=(ISTATE(8).EQ.1)
IELEM=ISTATE(9)
ICOL=ISTATE(10)
LL4=ISTATE(11)
ICHX=ISTATE(12)
ISPLH=ISTATE(13)
LX=ISTATE(14)
LY=ISTATE(15)
LZ=ISTATE(16)
ISEG=ISTATE(17)
LL4F=ISTATE(25)
IF(ICHX.EQ.2) THEN
   ITY=3
   LL4W=ISTATE(26)
   LL4X=ISTATE(27)
   LL4Y=ISTATE(28)
   LL4Z=ISTATE(29)
ELSE
   ITY=2
   LL4W=LL4
   LL4X=LL4
   LL4Y=LL4
   LL4Z=LL4
ENDIF
NLF=ISTATE(30)
NREG=ISTATE(1)
NVD=ISTATE(34)


! Get variables
ALLOCATE(MAT(NREG))
CALL LCMGET(IPTRK,'MATCOD',MAT)
ALLOCATE(ZZ(NREG),SIGT(NBMIX,NAN))
CALL LCMGET(IPTRK,'ZZ',ZZ)
CALL LCMGET(IPTRK,'SIDE',SIDE)
do il=1,NAN
   WRITE(TEXT12,'(4HSCAR,I2.2,2I3.3)') il-1,g1,g2
   CALL LCMLEN(IPSYS,TEXT12,LENGT,ITYLCM)
   IF(LENGT.EQ.0) THEN
      CALL XDRSET(SIGT(1,IL),NBMIX,0.0)
   else
      CALL LCMGET(IPSYS,TEXT12,SIGT(1,il))
   endif
enddo
CALL LCMGPD(IPTRK,'CTRAN',CTRAN_PTR)
CALL LCMGPD(IPTRK,'KN',KN_PTR)
CALL LCMGPD(IPTRK,'ZGLOB',IZGLOB_PTR)
CALL LCMGPD(IPTRK,'FRZ',FRZ_PTR)
NBLOS=LX*LZ*ISPLH**2
CALL C_F_POINTER(CTRAN_PTR,CTRAN,(/ (IELEM+1)*IELEM,(IELEM+1)*IELEM /))
CALL C_F_POINTER(KN_PTR,KN,(/ NBLOS,3+6*(IELEM+2)*IELEM**2 /))
CALL C_F_POINTER(IZGLOB_PTR,IZGLOB,(/ NBLOS /))
CALL C_F_POINTER(FRZ_PTR,FRZ,(/ LX*LZ /))
CALL LCMSIX(IPTRK,'BIVCOL',1)
CALL LCMLEN(IPTRK,'T',LC,ITYLCM)
ALLOCATE(R(LC,LC))
CALL LCMGET(IPTRK,'R',R)
CALL LCMSIX(IPTRK,' ',2)






! Fill terms
ALLOCATE(DIFF(NBLOS))
!
!----
!  MAIN LOOP OVER LEGENDRE ORDERS FOR THE FLUX.
!----
do IL=0,NLF-1
   IF(IL < NAN) then
      FACT=REAL(2*IL+1)
      !----
      !  RECOVER CROSS SECTIONS FOR THE PIOLAT TERMS.
      !----
      IF(MOD(IL,2) == 1) THEN
         do I=1,NBLOS
            DIFF(I)=0.0
            IND=IZGLOB(I)
            IF(IND.GT.0) THEN
               IBM=MAT(IND)
               GARS=SIGT(IBM,MIN(IL+1,NAN))
               VAR1=FACT*ZZ(IND)*FRZ(IND)*GARS
               DIFF(I)=REAL(VAR1)
            ENDIF
         enddo
      ENDIF

      NELEH=(IELEM+1)*IELEM**2
      TTTT=0.5D0*SQRT(3.D00)*SIDE*SIDE
      JOFF=(IL/2)*LL4
      NUM=0
      do KEL=1,NBLOS
         IND=IZGLOB(KEL)
         IF(IND /= 0) then
            IBM=MAT(IND)
            NUM=NUM+1
            DZ=ZZ(IND)
            VOL0=TTTT*DZ*FRZ(IND)
            UUUU=SIDE*DZ*FRZ(IND)
            GARS=SIGT(IBM,MIN(IL+1,NAN))
            IF(MOD(IL,2) == 0) THEN
               ! EVEN PARITY EQUATION
               do K3=0,IELEM-1
               do K2=0,IELEM-1
               do K1=0,IELEM-1
                  row=JOFF+(((NUM-1)*IELEM+K3)*IELEM+K2)*IELEM+K1+1
                  col=row
                  val=REAL(FACT*VOL0*GARS)
                  val=-1.0*val
                  row2=row+(g1mat-g1)*LL4*NLF/2
                  col2=col+(g2mat-g2)*LL4*NLF/2
                  call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                  row=JOFF+(((KN(NUM,1)-1)*IELEM+K3)*IELEM+K2)*IELEM+K1+1
                  col=row
                  val=REAL(FACT*VOL0*GARS)
                  val=-1.0*val
                  row2=row+(g1mat-g1)*LL4*NLF/2
                  col2=col+(g2mat-g2)*LL4*NLF/2
                  call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                  row=JOFF+(((KN(NUM,2)-1)*IELEM+K3)*IELEM+K2)*IELEM+K1+1
                  col=row
                  val=REAL(FACT*VOL0*GARS)
                  val=-1.0*val
                  row2=row+(g1mat-g1)*LL4*NLF/2
                  col2=col+(g2mat-g2)*LL4*NLF/2
                  call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
               enddo
               enddo
               enddo
            ELSE IF(MOD(IL,2) == 1) THEN
               ! ODD PARITY EQUATION.
               do K5=0,1 ! TWO LOZENGES PER HEXAGON
               do K4=0,IELEM-1
               do K3=0,IELEM-1
               do K2=1,IELEM+1
                  KNW1=KN(NUM,3+K5*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
                  KNX1=KN(NUM,3+(K5+2)*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
                  KNY1=KN(NUM,3+(K5+4)*NELEH+(K4*IELEM+K3)*(IELEM+1)+K2)
                  INW1=JOFF+LL4F+ABS(KNW1)
                  INX1=JOFF+LL4F+ABS(KNX1)
                  INY1=JOFF+LL4F+ABS(KNY1)
                  do K1=1,IELEM+1
                     KNW2=KN(NUM,3+K5*NELEH+(K4*IELEM+K3)*(IELEM+1)+K1)
                     KNX2=KN(NUM,3+(K5+2)*NELEH+(K4*IELEM+K3)*(IELEM+1)+K1)
                     KNY2=KN(NUM,3+(K5+4)*NELEH+(K4*IELEM+K3)*(IELEM+1)+K1)
                     INW2=JOFF+LL4F+ABS(KNW2)
                     INX2=JOFF+LL4F+ABS(KNX2)
                     INY2=JOFF+LL4F+ABS(KNY2)
                     IF((KNW2 /= 0).AND.(KNW1 /= 0)) THEN
                        SG=REAL(SIGN(1,KNW1)*SIGN(1,KNW2))
                        VAR1=(4./3.)*SG*FACT*VOL0*GARS*R(K2,K1)
                        row=INW1
                        col=INW2
                        val=-REAL(VAR1)
                        val=-1.0*val
                        row2=row+(g1mat-g1)*LL4*NLF/2
                        col2=col+(g2mat-g2)*LL4*NLF/2
                        call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                     ENDIF
                     IF((KNX2 /= 0).AND.(KNX1 /= 0)) THEN
                        SG=REAL(SIGN(1,KNX1)*SIGN(1,KNX2))
                        VAR1=(4./3.)*SG*FACT*VOL0*GARS*R(K2,K1)
                        row=INX1
                        col=INX2
                        val=-REAL(VAR1)
                        val=-1.0*val
                        row2=row+(g1mat-g1)*LL4*NLF/2
                        col2=col+(g2mat-g2)*LL4*NLF/2
                        call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                     ENDIF
                     IF((KNY2 /= 0).AND.(KNY1 /= 0)) THEN
                        SG=REAL(SIGN(1,KNY1)*SIGN(1,KNY2))
                        VAR1=(4./3.)*SG*FACT*VOL0*GARS*R(K2,K1)
                        row=INY1
                        col=INY2
                        val=-REAL(VAR1)
                        val=-1.0*val
                        row2=row+(g1mat-g1)*LL4*NLF/2
                        col2=col+(g2mat-g2)*LL4*NLF/2
                        call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                     ENDIF
                  enddo
               enddo
               enddo
               enddo
               enddo
               do K5=0,2 ! THREE LOZENGES PER HEXAGON
               do K2=0,IELEM-1
               do K1=0,IELEM-1
               do IC=1,2
                  IF(IC == 1) IIC=1
                  IF(IC == 2) IIC=IELEM+1
                  KNZ1=KN(NUM,3+6*NELEH+((2*K5+IC-1)*IELEM+K2)*IELEM+K1+1)
                  INZ1=JOFF+LL4F+ABS(KNZ1)
                  do JC=1,2
                     IF(JC == 1) JJC=1
                     IF(JC == 2) JJC=IELEM+1
                     KNZ2=KN(NUM,3+6*NELEH+((2*K5+JC-1)*IELEM+K2)*IELEM+K1+1)
                     INZ2=JOFF+LL4F+ABS(KNZ2)
                     IF((KNZ1 /= 0).AND.(KNZ2 /= 0)) THEN
                        SG=REAL(SIGN(1,KNZ1)*SIGN(1,KNZ2))
                        VAR1=SG*FACT*VOL0*GARS*R(IIC,JJC)
                        row=INZ1
                        col=INZ2
                        val=-REAL(VAR1)
                        val=-1.0*val
                        row2=row+(g1mat-g1)*LL4*NLF/2
                        col2=col+(g2mat-g2)*LL4*NLF/2
                        call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                     ENDIF
                  enddo
               enddo
               enddo
               enddo
               enddo
            ENDIF
         endif
      enddo
      IF(MOD(IL,2) == 1) THEN ! ADD PIOLA TERMS (matC) to odd equations
         ! matrix matCwxl 
         NUM=0
         do k=1,NBLOS
            if (IZGLOB(k) /= 0) then
               NUM=NUM+1
               do i1=0,IELEM-1
                  do i0=1,NELEM
                     i=i1*NELEM+i0
                     KN_col=KN(NUM,3+2*NELEH+i)
                     if(KN_col /= 0) then
                        col=ABS(KN_col)-LL4W
                        do j0=1,NELEM
                           j=i1*NELEM+j0
                           KN_row=KN(NUM,3+NELEH+j)
                           if (KN_row /= 0) then
                              row=ABS(KN_row)
                              SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                              DINV=DIFF(k)
                              val=-SG*DINV*REAL(CTRAN(i0,j0))
                              val=-1.0*val
                              row2=(g1mat-g1)*LL4*NLF/2+JOFF+LL4F+row
                              col2=(g2mat-g2)*LL4*NLF/2+JOFF+LL4F+LL4W+col
                              call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                              col2=(g2mat-g2)*LL4*NLF/2+JOFF+LL4F+row
                              row2=(g1mat-g1)*LL4*NLF/2+JOFF+LL4F+LL4W+col
                              call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                           endif
                        enddo
                     endif
                  enddo
               enddo
            endif
         enddo
         ! matrix matCwyl 
         NUM=0
         do k=1,NBLOS
            if (IZGLOB(k) /= 0) then
               NUM=NUM+1
               do i1=0,IELEM-1
                  do i0=1,NELEM
                     i=i1*NELEM+i0
                     KN_row=KN(NUM,3+i)
                     if(KN_row /= 0) then
                        row=ABS(KN_row)
                        do j0=1,NELEM
                           j=i1*NELEM+j0
                           KN_col=KN(NUM,3+5*NELEH+j)
                           if (KN_col /= 0) then
                              col=ABS(KN_col)-LL4W-LL4X
                              SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                              DINV=DIFF(k)
                              val=-SG*DINV*REAL(CTRAN(i0,j0))
                              val=-1.0*val
                              row2=(g1mat-g1)*LL4*NLF/2+JOFF+LL4F+row
                              col2=(g2mat-g2)*LL4*NLF/2+JOFF+LL4F+LL4W+LL4X+col
                              call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                              col2=(g2mat-g2)*LL4*NLF/2+JOFF+LL4F+row
                              row2=(g1mat-g1)*LL4*NLF/2+JOFF+LL4F+LL4W+LL4X+col
                              call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                           endif
                        enddo
                     endif
                  enddo
               enddo
            endif
         enddo
         ! matrix matCxyl 
         NUM=0
         do k=1,NBLOS
            if (IZGLOB(k) /= 0) then
               NUM=NUM+1
               do i1=0,IELEM-1
                  do i0=1,NELEM
                     i=i1*NELEM+i0
                     KN_col=KN(NUM,3+4*NELEH+i)
                     if(KN_col /= 0) then
                        col=ABS(KN_col)-LL4W-LL4X
                        do j0=1,NELEM
                           j=i1*NELEM+j0
                           KN_row=KN(NUM,3+3*NELEH+j)
                           if (KN_row /= 0) then
                              row=ABS(KN_row)-LL4W
                              SG=REAL(SIGN(1,KN_col)*SIGN(1,KN_row))
                              DINV=DIFF(k)
                              val=-SG*DINV*REAL(CTRAN(i0,j0))
                              val=-1.0*val
                              row2=(g1mat-g1)*LL4*NLF/2+JOFF+LL4F+LL4W+row
                              col2=(g2mat-g2)*LL4*NLF/2+JOFF+LL4F+LL4W+LL4X+col
                              call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                              col2=(g2mat-g2)*LL4*NLF/2+JOFF+LL4F+LL4W+row
                              row2=(g1mat-g1)*LL4*NLF/2+JOFF+LL4F+LL4W+LL4X+col
                              call MatSetValue(matA(g1mat,g2mat),row2-1,col2-1,val,ADD_VALUES,ierr)
                           endif
                        enddo
                     endif
                  enddo
               enddo
            endif
         enddo
      ENDIF
   endif
enddo
!  SCRATCH STORAGE DEALLOCATION
DEALLOCATE(DIFF)



end subroutine fill_matrix_Aij_hexa_raviart_thomas_spn




















end module SLEPC
